/*
	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/


#include "bezier.h"

/*
    control points = p1,p2,p3,p4 whereof p1,p4 are endpoints and
    p1p2,p3p4 are the respective tangent vectors. If more than one 
    curve is concatenated make sure that p4 - p3 = p5 - p4 if you
    want C0 and C1 steadyness (sp?)
    
    for the math freaks: the path is given by
    w(t)  = (1-t)^3 * P1 + 3*t*(1-t)^2 * P2 + 3*t^2*(1-t)*P3 + t^3*P4; t in [0,1]
    
    and the length is 
    Lw = /int_{0}^{1} |w'(t)| dt
*/


CL_Bezier::CL_Bezier(const CL_Vector *cp, int cs, int steps, bool stepping)
{
	curve = new CL_Vector[cs*steps];
	this->cp = new CL_Vector[cs*4];
	this->steps = steps;
	this->stepping = stepping;
	this->cs = cs;
	
	for (int i=0;i<cs*4;i++)
		this->cp[i] = cp[i];

	make_curve();
}


CL_Bezier::~CL_Bezier()
{
	delete [] curve;
	delete [] cp;
}


// hey it is not the fastest code, but I'll change it to a faster algorithm
// somewhen in the future, though this is not a promise ;-)

void CL_Bezier::make_curve()
{
	// TODO: interpret stepping
	for (int n=0; n<cs; n++)
	{
		curve[n*steps] = cp[n*4];

		float t, delta;
		delta = 1.0/steps;
	
		for (int i=1; i<steps; i++)
		{
			t = i*delta;
			curve[n*steps+i].x = 
				cp[n*4+0].x * (1.0 - t) * (1.0 - t) * (1.0 - t) +
	        		cp[n*4+1].x * 3.0 * t * (1.0 - t) * (1.0 - t) +
		        	cp[n*4+2].x * 3.0 * t *  t  * (1.0 - t) + 
				cp[n*4+3].x * t * t * t;

			curve[n*steps+i].y = 
				cp[n*4+0].y * (1.0 - t) * (1.0 - t) * (1.0 - t) +
	        		cp[n*4+1].y * 3.0 * t * (1.0 - t) * (1.0 - t) +
	            		cp[n*4+2].y * 3.0 * t *  t  * (1.0 - t) + 
				cp[n*4+3].y * t * t * t;

			curve[n*steps+i].z =
				cp[n*4+0].z * (1.0 - t) * (1.0 - t) * (1.0 - t) +
				cp[n*4+1].z * 3.0 * t * (1.0 - t) * (1.0 - t) +
		    		cp[n*4+2].z * 3.0 * t *  t  * (1.0 - t) + 
				cp[n*4+3].z * t * t * t;
		}
	}
}


float CL_Bezier::get_length(int segment) const
{
	if (segment == -1)
	{
		float length = 0;
		for (int i=0;i<cs;i++)
			length += get_length(cs);
		return length;
	}
	// TODO: calculate Lw for segment
	return 0;	
}


void CL_Bezier::set_steps(int steps)
{
	this->steps = steps;
	make_curve();
}


void CL_Bezier::set_stepping(bool stepping)
{
	this->stepping = stepping;
	make_curve();
}
