/*
	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	1999/07/07	Daniel Vogel
	
	TODO: length and stepping per curve 
*/

//! common_component="Math"

#ifndef header_bezier
#define header_bezier

#include "vector.h"

class CL_Bezier
//: Class for Bezier curves
// This class generates points on a Bezier from given control points
{
protected:
	CL_Vector *curve;	// the generated curve
	CL_Vector *cp;		// control points

	int steps, cs;		// number of steps, curve segments 
	bool stepping;		// steps per segment or per curve
		
	virtual void make_curve();

public:
	CL_Bezier(const CL_Vector *cp, int cs, int steps, bool stepping = true);
	//: Constructor
	//!param: cp - an array of control points with size (4 * cs)
	//!param: cs - number of curve segments
	//!param: steps - number of points to generate
	//!param: stepping - whether steps is the number of points per segment or per curve. 

	virtual	~CL_Bezier();
	//: Destructor
	
	virtual int get_steps() const { return steps; }
	//: Returns the number of steps
	//!retval: number of steps
	
	virtual int get_cs() const { return cs; }
	//: Returns the number of curve segments
	//!retval: number of curve segments

	virtual bool get_stepping() const { return stepping; }
	//: Returns whether steps is the number of points per segment or per curve.
	//!retval: stepping

	virtual float get_length(int segment = -1) const;
	//: Returns the lenght of a segment (or the curve)
	//!param: segment - number of segment (-1 indicates whole curve)
	//!retval: length of segment (or curve)
	
	virtual void set_steps(int steps);
	//: Sets the number of steps
	//!param: steps - number of steps
	
	virtual void set_stepping(bool stepping);
	//: Sets the stepping
	//!param: stepping - whether steps is the number of points per segment or per curve.

	virtual CL_Vector* get_curve() const { return curve; }
	//: Returns an array of 4 * steps CL_Vectors -> the curve
	//!retval: the curve (4 * steps CL_Vectors)
	
	virtual CL_Vector* get_control_points() const { return cp; }
	//: Returns the 4 * cs control points
	//!retval: the 4 * cs control points

	CL_Vector operator [] (int n) { return curve[n]; }
	//: Returns the n-th vector of the curve (_doesn't_ return a reference)
	//!param: n - the vector to return (counting starts at 0)
	//!retval: the n-th vector of the curve (_doesn't_ return a reference)
};

#endif
