/*
	$Id: surface.h,v 1.8 1999/05/27 21:44:09 starch Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

//! l1_component="Display 2D"

#ifndef header_surface
#define header_surface

#include "displaycard.h"
#include "surfaceprovider.h"
#include "surfacepriority.h"

class CL_Surface
//: This class is used to draw (blit/put screen) images to the screen. The 
//: actual image is provided by a CL_SurfaceProvider.
// To create a surface, you can call the create() function:
// <code>
// CL_Surface *surf = CL_Surface::create(
// 	new CL_PCXProvider("image.pcx", NULL),
//      true);
// </code>
// However, in pratice you normally do not do it this way because the surface
// provider has a static create function as well. Its purpose is only to
// save you from some typing. The code below does exactly the same:
// <code>
// CL_Surface *surf = CL_PCXProvider::create("image.pcx", NULL);
// </code>
// Please note, that surfaces can also be loaded from resource files. If you
// want to do that, have a look on the CL_ResourceManager class.
//!also: Layer2 surface providers - All surface providers available.
//!also: CL_SurfaceProvider - The SurfaceProvider interface.
//!also: CL_ResourceManager - Resources in ClanLib.
{
public:

	static CL_Surface *create(CL_SurfaceProvider *provider, bool delete_provider=false);
//: This functions creates a CL_Surface instance.
//!param: provider - Image source.
//!param: delete_provider - If true, the surface will delete the provider when deleted itself.
//!retval: The surface created.

	static void set_priority_handler(CL_SurfacePriority *priority_handler);
//: Set the surface priority handler. This handler is contacted each time a 
//: surface is created.
//!param: priority_handler - new priority handler to use - or NULL if none.

	virtual ~CL_Surface() {;}

	virtual void reload()=0; 
//: Forces surface to reload surface data from provider.
	
	virtual CL_SurfaceProvider *get_provider() const =0;
//: Returns the surface provider used by this surface.
//!retval: Surface provider used by this surface.

//!info: Display functions.

	virtual void put_screen(int x, int y, int spr_no=0, CL_DisplayCard *card=NULL)=0;
//: Draws the surface onto the backbuffer.
//!param: x - x coordinate to draw surface.
//!param: y - y coordinate to draw surface.
//!param: spr_no - subsprite number to use.
//!param: card - Display card to be drawn onto. If NULL it will use the currently selected video card by CL_Display.

	virtual void put_screen(int x, int y, float scale_x, float scale_y, int spr_no=0, CL_DisplayCard *card=NULL)=0;
//: Draws the surface onto the backbuffer - scaling it to a specified size.
//!param: x - x coordinate to draw surface.
//!param: y - y coordinate to draw surface.
//!param: scale_x - scale x multiplyer.
//!param: scale_y - scale y multiplyer.
//!param: spr_no - subsprite number to use.
//!param: card - Display card to be drawn onto. If NULL it will use the currently selected video card by CL_Display.

	virtual void put_screen(int x, int y, int size_x, int size_y, int spr_no=0, CL_DisplayCard *card=NULL)=0;
//: Draws the surface onto the backbuffer - scaling it to a specified size.
//!param: x - x coordinate to draw surface.
//!param: y - y coordinate to draw surface.
//!param: size_x - width of the image when scaled.
//!param: size_y - height of the image when scaled.
//!param: spr_no - subsprite number to use.
//!param: card - Display card to be drawn onto. If NULL it will use the currently selected video card by CL_Display.

	virtual int get_width() const =0;
//: Returns the width of the surface
//!retval: Width of the surface.

	virtual int get_height() const =0;
//: Returns the height of the surface
//!retval: Height of the surface.

	virtual int get_no_sprs() const =0;
//: Returns the number of sprites in the surface
//!retval: Number of sprites in surface.

	virtual bool is_video(CL_DisplayCard *card = NULL) const =0;
//: returns true if in videomemory 
//: on the specified card (null = current dispcard)
//!retval: True if in video memory, false otherwise.
	
	virtual bool is_loaded(CL_DisplayCard *card = NULL) const =0;
//: returns true if loaded in either video- or system-memory
//!retval: True if in system memory, false otherwise.

	virtual bool convert_video(CL_DisplayCard *card = NULL)=0;
//: returns true if successfully loaded into videomemory, or already there
//!retval: True if successfully loaded into video memory.

	virtual bool convert_system(CL_DisplayCard *card = NULL)=0;
//: convert surface to system memory - never fails! (or serious heap usage!)
//!retval: True if successfully loaded into system memory.

	virtual void flush(CL_DisplayCard *card = NULL)=0;
//: completely flushes surface (removes from video/system-memory)
};

#endif
