/*
	$Id: dx_target.cpp,v 1.8 1999/09/04 00:40:55 starch Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#include "Layer1/System/API/cl_assert.h"

#include "dx_target.h"
#include "blit_dx.h"

CL_Target_DX::CL_Target_DX()
{
	m_surface = NULL;
	m_data = NULL;
	m_width = 0;
	m_height = 0;
	m_pitch = 0;
	m_depth = 0;
	m_red_mask = 0;
	m_green_mask = 0;
	m_blue_mask = 0;
	m_alpha_mask = 0;
}

CL_Target_DX::~CL_Target_DX()
{
	if (m_surface != NULL) m_surface->Release();
}

void CL_Target_DX::set_surface(LPDIRECTDRAWSURFACE surface)
{
	m_surface = surface;

	get_palette_from_surface();

	DDSURFACEDESC surface_desc;
	surface_desc.dwSize = sizeof(DDSURFACEDESC);

	m_surface->GetSurfaceDesc(&surface_desc);
	
	m_pitch = surface_desc.lPitch;
	m_width = surface_desc.dwWidth;
	m_height = surface_desc.dwHeight;
	m_depth = surface_desc.ddpfPixelFormat.dwRGBBitCount;
	m_red_mask = surface_desc.ddpfPixelFormat.dwRBitMask;
	m_green_mask = surface_desc.ddpfPixelFormat.dwGBitMask;
	m_blue_mask = surface_desc.ddpfPixelFormat.dwBBitMask;
	m_alpha_mask = surface_desc.ddpfPixelFormat.dwRGBAlphaBitMask;
}

bool CL_Target_DX::lock()
{
	cl_assert(m_surface != NULL);

	DDSURFACEDESC surface_desc;
	memset(&surface_desc, 0, sizeof(DDSURFACEDESC));
	surface_desc.dwSize = sizeof(DDSURFACEDESC);

	HRESULT err = m_surface->Lock(
		NULL,
		&surface_desc,
		DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,
		NULL);

	if (err != DD_OK) return false;

	m_data = (unsigned char *) surface_desc.lpSurface;
	m_pitch = surface_desc.lPitch;
	m_width = surface_desc.dwWidth;
	m_height = surface_desc.dwHeight;
	m_depth = surface_desc.ddpfPixelFormat.dwRGBBitCount;
	m_red_mask = surface_desc.ddpfPixelFormat.dwRBitMask;
	m_green_mask = surface_desc.ddpfPixelFormat.dwGBitMask;
	m_blue_mask = surface_desc.ddpfPixelFormat.dwBBitMask;
	m_alpha_mask = surface_desc.ddpfPixelFormat.dwRGBAlphaBitMask;

	return true;
}

void CL_Target_DX::unlock()
{
	HRESULT err = m_surface->Unlock(m_data);
	cl_assert(err == DD_OK);
}

void CL_Target_DX::set_palette(CL_Palette *pal)
{
	memcpy(m_palette.palette, pal->palette, m_palette.num_colors*3);
}

CL_Palette *CL_Target_DX::get_palette()
{
	return &m_palette;
}

void CL_Target_DX::get_palette_from_surface()
{
	cl_info(0, "not impl!");
}

/****************************************
        CL_OffscreenTarget_DX
****************************************/

CL_OffscreenTarget_DX::CL_OffscreenTarget_DX(
	LPDIRECTDRAW ddraw, 
	int width, 
	int height, 
	bool create_video)
{
	m_is_video = create_video;

	DDSURFACEDESC sur_desc;
	sur_desc.dwSize = sizeof(DDSURFACEDESC);
	sur_desc.dwFlags=DDSD_CAPS|DDSD_HEIGHT|DDSD_WIDTH;
	sur_desc.dwWidth=width;
	sur_desc.dwHeight=height;
	if (create_video)
	{
		sur_desc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_VIDEOMEMORY;
	}
	else
	{
		sur_desc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;
	}

	HRESULT err = ddraw->CreateSurface(&sur_desc, &m_surface, NULL);
	if (err != DD_OK && create_video)
	{
		m_is_video = false;

		cl_info(1, "Fallback to system");
		sur_desc.ddsCaps.dwCaps&=~DDSCAPS_VIDEOMEMORY;
		sur_desc.ddsCaps.dwCaps|=DDSCAPS_SYSTEMMEMORY;
		err = ddraw->CreateSurface(&sur_desc, &m_surface, NULL);
		cl_assert(err == DD_OK);
	}

	m_surface->AddRef();
	set_surface(m_surface);
}

CL_OffscreenTarget_DX::~CL_OffscreenTarget_DX()
{
	m_surface->Release();
}

/****************************************
        CL_Target_DX
****************************************/

CL_BackbufferTarget_DX::CL_BackbufferTarget_DX(
	LPDIRECTDRAW ddraw, 
	int backbuffer_count, 
	bool create_video_backbuffer)
{
	HRESULT err;

	m_ddraw = ddraw;

	DDSURFACEDESC surf_desc;
	memset(&surf_desc, 0, sizeof(surf_desc));
	surf_desc.dwSize=sizeof(DDSURFACEDESC);
	surf_desc.dwFlags=DDSD_CAPS;
	surf_desc.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE|DDSCAPS_3DDEVICE;

	if (create_video_backbuffer)
	{
		surf_desc.dwFlags|=DDSD_BACKBUFFERCOUNT;
		surf_desc.ddsCaps.dwCaps|=DDSCAPS_FLIP|DDSCAPS_COMPLEX|DDSCAPS_VIDEOMEMORY;
		surf_desc.dwBackBufferCount=backbuffer_count;
	}

	err = ddraw->CreateSurface(&surf_desc, &m_primary_surface, NULL);
	if (err != DD_OK)
	{
		create_video_backbuffer = false;

		memset(&surf_desc, 0, sizeof(surf_desc));
		surf_desc.dwSize=sizeof(DDSURFACEDESC);
		surf_desc.dwFlags=DDSD_CAPS;
		surf_desc.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE|DDSCAPS_3DDEVICE;

		err = ddraw->CreateSurface(&surf_desc, &m_primary_surface, NULL);
		cl_assert(err == DD_OK);
	}

	if (create_video_backbuffer)
	{
		m_system_backbuffer = NULL;

		LPDIRECTDRAWSURFACE backbuffer;
		DDSCAPS backbuffer_desc;
		backbuffer_desc.dwCaps=DDSCAPS_BACKBUFFER;
		m_primary_surface->GetAttachedSurface(&backbuffer_desc, &backbuffer);
		backbuffer->AddRef();
		set_surface(backbuffer);
	}
	else
	{
		memset(&surf_desc, 0, sizeof(surf_desc));
		surf_desc.dwSize=sizeof(DDSURFACEDESC);
		m_primary_surface->GetSurfaceDesc(&surf_desc);

		m_system_backbuffer = new CL_OffscreenTarget_DX(
			ddraw,
			surf_desc.dwWidth,
			surf_desc.dwHeight,
			false);

		m_system_backbuffer->m_surface->AddRef();
		set_surface(m_system_backbuffer->m_surface);
	}

}

CL_BackbufferTarget_DX::~CL_BackbufferTarget_DX()
{
	m_primary_surface->Release();
	delete m_system_backbuffer;
}

void CL_BackbufferTarget_DX::flip()
{
	if (m_primary_surface->IsLost())
	{
		if (m_primary_surface->Restore() == DD_OK)
		{
			CL_Blit_DX::reload_all_surfaces();
		}
		else
		{
			return;
		}
	}

	if (m_system_backbuffer)
	{
		DDBLTFX fx;
		fx.dwSize=sizeof(DDBLTFX);

//		m_ddraw->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);

		DDSURFACEDESC surface_desc;
		memset(&surface_desc, 0, sizeof(DDSURFACEDESC));
		surface_desc.dwSize = sizeof(DDSURFACEDESC);

		DDSURFACEDESC surface_desc2;
		memset(&surface_desc2, 0, sizeof(DDSURFACEDESC));
		surface_desc2.dwSize = sizeof(DDSURFACEDESC);

		HRESULT err = m_primary_surface->Lock(
			NULL,
			&surface_desc,
			DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT | DDLOCK_WRITEONLY,
			NULL);
		if (err != DD_OK) return;

		err = m_system_backbuffer->m_surface->Lock(
			NULL,
			&surface_desc2,
			DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT | DDLOCK_READONLY,
			NULL);
		if (err != DD_OK) 
		{
			m_primary_surface->Unlock(surface_desc.lpSurface);
			return;
		}

		unsigned char *data = (unsigned char *) surface_desc.lpSurface;
		unsigned char *src_data = (unsigned char *) surface_desc2.lpSurface;
		unsigned int pitch = surface_desc.lPitch;
		cl_assert(surface_desc.lPitch == surface_desc2.lPitch);

		unsigned int height = surface_desc.dwHeight;
		for (unsigned int y=0;y<height;y++)
		{
			memcpy(data, src_data, pitch);
			data += pitch;
			src_data += pitch;
		}

		err = m_primary_surface->Unlock(data);
		cl_assert(err == DD_OK);
		err = m_system_backbuffer->m_surface->Unlock(src_data);
		cl_assert(err == DD_OK);
	}
	else
	{
		m_primary_surface->Flip(NULL, DDFLIP_WAIT);
	}
}
