/*
	$Id: blitters.h,v 1.3 1999/06/22 02:19:17 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
	
	This file contains all the blitter interfaces available. Each interface
	is used to blit a special put_screen case. Eg. the Blit_NoClip is used
	when the surface doesn't need any clipping.
	
	The general idea is, that these blitters should be inheriated to
	software and hardware versions. These are the main benifits of this
	system:
	
	1) It is possible to combine hardware and software blitting. And the
	   software blitters share code on all platforms and implementations.

	2) If an implementation doesn't support some kind of hardware
	   acceleration, it is easy to use the generic software blitters as
	   fallbacks.

	3) All blitters can rearrange the input data, so it specially adjusted
	   for their type of blitting. This can make software blitters A LOT
	   faster than otherwise.

	4) Extendable. It is very easy to add new put_screens/blitters to the
	   CL_Surface class. Just add a new blit interface and a software
	   version that all implementations can fall back to.
	   
	5) Flexible. Blitters can use other blitters to divide a larger blit
	   into easiliy solveable "sub" blits. A general Blit_NoClip blitter
	   could for instance use other blitters to divide eg. the alpha part of
	   a surface from the rest.
	   

	TODO:
	
	 - Add blitters for primitive types: lines, triangles and polygons.
*/

#ifndef header_blitters
#define header_blitters

#include "../API/cliprect.h"
#include "../API/surfaceprovider.h"
#include "../API/target.h"
#include "pixeldata.h"

class CL_RefCountable
{
public:
	void add_ref() { ref++; }
	int release_ref() { ref--; return ref; }

	virtual ~CL_RefCountable() { return; }
protected:
	CL_RefCountable() { ref = 0; }
private:
	int ref;
};

class CL_Blit_NoClip : virtual public CL_RefCountable
{
public:
	virtual ~CL_Blit_NoClip() { return; }

	virtual void blt_noclip(
		CL_Target *target,
		int x,
		int y,
		int spr_no)=0;
};

class CL_Blit_Clip : virtual public CL_RefCountable
{
public:
	virtual ~CL_Blit_Clip() { return; }

	virtual void blt_clip(
		CL_Target *target,
		int x,
		int y,
		int spr_no,
		const CL_ClipRect &clip)=0;
};

class CL_Blit_Scale_NoClip : virtual public CL_RefCountable
{
public:
	virtual ~CL_Blit_Scale_NoClip() { return; }

	virtual void blt_scale_noclip(
		CL_Target *target,
		int x,
		int y,
		int dest_width,
		int dest_height,
		int spr_no)=0;
};

class CL_Blit_Scale_Clip : virtual public CL_RefCountable
{
public:
	virtual ~CL_Blit_Scale_Clip() { return; }

	virtual void blt_scale_clip(
		CL_Target *target,
		int x,
		int y,
		int dest_width,
		int dest_height,
		int spr_no,
		const CL_ClipRect &clip)=0;
};

class CL_Blitters
{
public:
	CL_Blitters()
	{
		noclip = NULL;
		clip = NULL;
		scale_noclip = NULL;
		scale_clip = NULL;
	}
	
	void delete_all()
	{
		delete_noclip();
		delete_clip();
		delete_scale_noclip();
		delete_scale_clip();
	}
	
	void set_noclip(CL_Blit_NoClip *ptr)
	{
		delete_noclip();
		noclip = ptr;
		ptr->add_ref();
	}
	
	void set_clip(CL_Blit_Clip *ptr)
	{
		delete_clip();
		clip = ptr;
		ptr->add_ref();
	}
	
	void set_scale_noclip(CL_Blit_Scale_NoClip *ptr)
	{
		delete_scale_noclip();
		scale_noclip = ptr;
		ptr->add_ref();
	}
	
	void set_scale_clip(CL_Blit_Scale_Clip *ptr)
	{
		delete_scale_clip();
		scale_clip = ptr;
		ptr->add_ref();
	}
	
	void delete_noclip() { remove(noclip); noclip = NULL; }
	void delete_clip() { remove(clip); clip = NULL; }
	void delete_scale_noclip() { remove(scale_noclip); scale_noclip = NULL; }
	void delete_scale_clip() { remove(scale_clip); scale_clip = NULL; }
	
	bool test_noclip() { return (noclip == NULL) ? false : true; }
	bool test_clip() { return (clip == NULL) ? false : true; }
	bool test_scale_noclip() { return (scale_noclip == NULL) ? false : true; }
	bool test_scale_clip() { return (scale_clip == NULL) ? false : true; }
	
	bool test_complete()
	{
		return
			test_noclip() &&
			test_clip() &&
			test_scale_noclip() &&
			test_scale_clip();
	}
	
	CL_Blit_NoClip *get_noclip() { return noclip; }
	CL_Blit_Clip *get_clip() { return clip; }
	CL_Blit_Scale_NoClip *get_scale_noclip() { return scale_noclip; }
	CL_Blit_Scale_Clip *get_scale_clip() { return scale_clip; }

private:
	CL_Blit_NoClip *noclip;
	CL_Blit_Clip *clip;
	CL_Blit_Scale_NoClip *scale_noclip;
	CL_Blit_Scale_Clip *scale_clip;

	void remove(CL_RefCountable *ptr)
	{
		if (ptr != NULL && ptr->release_ref() == 0) delete ptr;
	}
};

#endif
