/*
	$Id: displaycard_generic.cpp,v 1.19 1999/07/13 21:55:11 fischlustig Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#include <stdlib.h>

#include "Layer1/Display/API/display.h"
#include "Layer1/Display/API/surface.h"
#include "Layer1/System/API/cl_assert.h"
#include "displaycard_generic.h"
#include "surface_generic.h"
#include "cardsurface.h"
#include "colormap.h"
#include "blitters.h"
#include "cardsurface_blitter.h"
#include "lockable_surface_software.h"

void CL_DisplayCard::set_videomode(class CL_VidMode *vidmode)
{
	set_videomode(
		vidmode->width,
		vidmode->height,
		vidmode->bpp);
}

CL_CardSurface *CL_DisplayCard_Generic::create_cardsurface_video(
	CL_SurfaceProvider *provider)
{
	return new CL_CardSurface_Blitter(this, provider);
}

CL_CardSurface *CL_DisplayCard_Generic::create_cardsurface_system(
	CL_SurfaceProvider *provider)
{
	return new CL_CardSurface_Blitter(this, provider);
}

CL_Blitters CL_DisplayCard_Generic::create_hw_blitters(CL_SurfaceProvider *provider)
{
	CL_Blitters ret;
	return ret;
}

CL_LockableSurface *CL_DisplayCard_Generic::create_lockable_surface(
	int width,
	int height,
	int no_sprs,
	int minimum_alpha_mask_size)
{
	return new CL_LockableSurface_Software(
		width,
		height,
		no_sprs,
		minimum_alpha_mask_size,
		this);
}

void CL_DisplayCard_Generic::clear_display(
	float red,
	float green,
	float blue,
	float alpha)
{
	fill_rect(
		0,
		0,
		get_width(),
		get_height(),
		red,
		green,
		blue,
		alpha);
}

void CL_DisplayCard_Generic::fill_rect(
	int x1,
	int y1,
	int x2,
	int y2,
	float r,
	float g,
	float b,
	float a)
{
	if (a <= 0.01) return;
	if (a >= 0.99) 
	{
		fill_rect_opaque(x1, y1, x2, y2, r, g, b, a);
		return;
	}

	CL_ClipRect rect(x1, y1, x2, y2);

	if (m_cur_clip.test_all_clipped(rect)) 
	{
		return;
	}

	CL_ClipRect crect = m_cur_clip.clip(rect);

	CL_Target *target = get_target();
	if (!target->lock()) return;

	CL_ColorMap cmap(target);

	unsigned int fill_color = cmap.calc_color(r, g, b, a);

	unsigned int red_mask = target->get_red_mask();
	unsigned int green_mask = target->get_green_mask();
	unsigned int blue_mask = target->get_blue_mask();
	
	{
		unsigned int rr = fill_color & red_mask;
		unsigned int gg = fill_color & green_mask;
		unsigned int bb = fill_color & blue_mask;

		rr >>= cmap.m_red_start;
		gg >>= cmap.m_green_start;
		bb >>= cmap.m_blue_start;
	
		rr = (int) (rr*a);
		gg = (int) (gg*a);
		bb = (int) (bb*a);

		fill_color =
			(rr<<cmap.m_red_start) +
			(gg<<cmap.m_green_start) +
			(bb<<cmap.m_blue_start);
	}

	unsigned int dest_bytes_per_pixel = (get_target()->get_depth()+7)/8;
	unsigned int dest_pitch = target->get_pitch();

	unsigned char *dest = (unsigned char*) target->get_data();
	dest += crect.m_x1*dest_bytes_per_pixel + crect.m_y1*dest_pitch;

	unsigned int delta_x = crect.m_x2-crect.m_x1;
	unsigned int delta_y = crect.m_y2-crect.m_y1;

	int rr_start = cmap.m_red_start;
	int gg_start = cmap.m_green_start;
	int bb_start = cmap.m_blue_start;

	switch (dest_bytes_per_pixel)
	{
	case 2:
		{
			unsigned short rr_lookup[256];
			unsigned short gg_lookup[256];
			unsigned short bb_lookup[256];

			int i;
			for (i=0;i<(1<<cmap.m_red_length);i++)
			{
				rr_lookup[i] = int(i*(1-a)) << rr_start;
			}
			for (i=0;i<(1<<cmap.m_green_length);i++)
			{
				gg_lookup[i] = int(i*(1-a)) << gg_start;
			}
			for (i=0;i<(1<<cmap.m_blue_length);i++)
			{
				bb_lookup[i] = int(i*(1-a)) << bb_start;
			}

			int skip = dest_pitch - delta_x*2;
			unsigned short clear_value = (unsigned short) fill_color;
			
			if (target->is_video())
			{
				unsigned short data[4096];
				for (unsigned int yy=0; yy<delta_y; yy++)
				{
					memcpy(data, dest, delta_x*2);
					for (unsigned int xx=0;xx<delta_x;xx++)
					{
						unsigned short dval = data[xx];
						unsigned int rr = dval & red_mask;
						unsigned int gg = dval & green_mask;
						unsigned int bb = dval & blue_mask;
						rr >>= rr_start;
						gg >>= gg_start;
						bb >>= bb_start;

						dval =
							rr_lookup[rr] + 
							gg_lookup[gg] + 
							bb_lookup[bb] + 
							clear_value;

						data[xx] = dval;
					}
					memcpy(dest, data, delta_x*2);
					dest += delta_x*2;
					dest += skip;
				}
			}
			else
			{
				for (unsigned int yy=0; yy<delta_y; yy++)
				{
					for (unsigned int xx=0;xx<delta_x;xx++)
					{
						unsigned short dval = *((unsigned short *)dest+xx);
						unsigned int rr = dval & red_mask;
						unsigned int gg = dval & green_mask;
						unsigned int bb = dval & blue_mask;
						rr >>= rr_start;
						gg >>= gg_start;
						bb >>= bb_start;

						dval =
							rr_lookup[rr] + 
							gg_lookup[gg] + 
							bb_lookup[bb] + 
							clear_value;

						*((unsigned short *)dest+xx) = dval;
					}
					dest += delta_x*2;
					dest += skip;
				}
			}
		}
		break;

	case 4:
		{
			unsigned int rr_lookup[256];
			unsigned int gg_lookup[256];
			unsigned int bb_lookup[256];

			int i;
			for (i=0;i<(1<<cmap.m_red_length);i++)
			{
				rr_lookup[i] = int(i*(1-a)) << rr_start;
			}
			for (i=0;i<(1<<cmap.m_green_length);i++)
			{
				gg_lookup[i] = int(i*(1-a)) << gg_start;
			}
			for (i=0;i<(1<<cmap.m_blue_length);i++)
			{
				bb_lookup[i] = int(i*(1-a)) << bb_start;
			}

			int skip = dest_pitch - delta_x*4;

			if (target->is_video())
			{
				unsigned int data[1280];
				for (unsigned int yy=0; yy<delta_y; yy++)
				{
					memcpy(data, dest, delta_x*4);
					for (unsigned int xx=0;xx<delta_x;xx++)
					{
						unsigned int dval = data[xx];
						unsigned int rr = dval & red_mask;
						unsigned int gg = dval & green_mask;
						unsigned int bb = dval & blue_mask;
						rr >>= rr_start;
						gg >>= gg_start;
						bb >>= bb_start;

						dval =
							rr_lookup[rr] + 
							gg_lookup[gg] + 
							bb_lookup[bb] + 
							fill_color;

						data[xx] = dval;
					}
					memcpy(dest, data, delta_x*4);
					dest += delta_x*4;
					dest += skip;
				}
			}
			else
			{
				for (unsigned int yy=0; yy<delta_y; yy++)
				{
					for (unsigned int xx=0;xx<delta_x;xx++)
					{
						unsigned int dval = *((unsigned int *) dest);
						unsigned int rr = dval & red_mask;
						unsigned int gg = dval & green_mask;
						unsigned int bb = dval & blue_mask;
						rr >>= rr_start;
						gg >>= gg_start;
						bb >>= bb_start;

						dval =
							rr_lookup[rr] + 
							gg_lookup[gg] + 
							bb_lookup[bb] + 
							fill_color;

						*((unsigned int *) dest) = dval;

						dest += 4;
					}
					dest += skip;
				}
			}
		}
		break;

	default:
		cl_assert(false); // unsupported pixel depth!
		break;
	}

	target->unlock();
}

void CL_DisplayCard_Generic::fill_rect_opaque(
	int x1,
	int y1,
	int x2,
	int y2,
	float r,
	float g,
	float b,
	float a)
{
	CL_ClipRect rect(x1, y1, x2, y2);

	if (m_cur_clip.test_all_clipped(rect)) 
	{
		return;
	}

	CL_ClipRect crect = m_cur_clip.clip(rect);

	CL_Target *target = get_target();
	if (!target->lock()) return;

	CL_ColorMap cmap(target);

	unsigned int fill_color = cmap.calc_color(r, g, b, a);

	unsigned int dest_bytes_per_pixel = (get_target()->get_depth()+7)/8;
	unsigned int dest_pitch = target->get_pitch();

	unsigned char *dest = (unsigned char*) target->get_data();
	dest += crect.m_x1*dest_bytes_per_pixel + crect.m_y1*dest_pitch;

	unsigned int delta_x = crect.m_x2-crect.m_x1;
	unsigned int delta_y = crect.m_y2-crect.m_y1;

	switch (dest_bytes_per_pixel)
	{
	case 2:
		{
			unsigned short clear_color = (unsigned short) fill_color;
			int skip = dest_pitch - delta_x*2;
			// the blitter probes if he can perform a memset (when hi and lo byte of clear_color are equal)
			if (clear_color>>8 == (unsigned short)(clear_color<<8)>>8)
			{
				char clear_color8 = clear_color>>8;
				if (skip==0) // everything can be done with one single memset :-)
				{
					memset(dest, clear_color8, 2*delta_x*delta_y);
				} else
				{
					for (unsigned int yy=0; yy<delta_y; yy++)
					{
						memset(dest, clear_color8, 2*delta_x);
						dest += skip + 2*delta_x;
					}
				}
			} else
			{
				for (unsigned int yy=0; yy<delta_y; yy++)
				{
					for (unsigned int xx=0;xx<delta_x;xx++)
					{
						*((unsigned short *) dest) = clear_color;
						dest += 2;
					}
					dest += skip;
				}
			}
		}
		break;

	case 4:
		{
			int skip = dest_pitch - delta_x*4;
			for (unsigned int yy=0; yy<delta_y; yy++)
			{
				for (unsigned int xx=0;xx<delta_x;xx++)
				{
					*((unsigned int *) dest) = fill_color;
					dest += 4;
				}
				dest += skip;
			}
		}
		break;

	default:
		cl_assert(false); // unsupported pixel depth!
		break;
	}

	target->unlock();
}

void CL_DisplayCard_Generic::sync_buffers()
{
	CL_Target *front = get_frontbuffer();
	CL_Target *back = get_target();
	
	if (front == back) return; // single buffered system.
	
	back->lock();
	front->lock();

	memcpy(
		back->get_data(),
		front->get_data(),
		back->get_height()*back->get_pitch());

	front->unlock();
	back->unlock();
}

void CL_DisplayCard_Generic::draw_line(
	int x1,
	int y1,
	int x2,
	int y2,
	float r,
	float g,
	float b,
	float a)
{
	CL_ClipRect rect(x1, y1, x2, y2);

	if (m_cur_clip.test_all_clipped(rect))
	{
		return;
	}

	CL_ClipRect crect = m_cur_clip.clip(rect);

	CL_Target *target = get_target();
	if (!target->lock()) return;

	CL_ColorMap cmap(target);

	unsigned int fill_color = cmap.calc_color(r, g, b, a);

//FIXME
#define FBTYPE short
	FBTYPE color = (FBTYPE)fill_color;
	
	int orig_x1 = x1;
	int orig_x2 = x2;
	int orig_y1 = y1;
	int orig_y2 = y2;
	
	int clipped_x1 = crect.m_x1;
	int clipped_x2 = crect.m_x2;
	int clipped_y1 = crect.m_y1;
	int clipped_y2 = crect.m_y2;
	
	int orig_dx, orig_dy, dx, dy, sx, sy, w, i;
	
	void *clip_first = NULL;//FIXME
	void *clip_last = NULL; //HACK
	
	dy = clipped_y2 - clipped_y1;
	orig_dy = orig_y2 - orig_y1;
	sy=1;
	if (orig_dy<0) {
		orig_dy = -orig_dy;
		dy = -dy;
		sy = -1;
	}

	dx = clipped_x2-clipped_x1;
	orig_dx = orig_x2 - orig_x1;
	sx=1;
	if (orig_dx<0) {
		sx=-1;
		orig_dx = -orig_dx;
		dx = -dx;
	}

	w = target->get_pitch();

	FBTYPE *fb = (FBTYPE *) (((char*)target->get_data()) + clipped_y1*w+clipped_x1*sizeof(FBTYPE));

	if (dx==0) {
		if (sy<0)
			w = -w;

		for (i=dy;i>=0; i--) {
			*fb = color;
			fb = (FBTYPE *) (((char*)fb) + w);
		}
		return;
	}

	if (dy==0) {
		for (i=dx;i>=0; i--) {
			*fb = color;
			fb += sx;
		}
		return;
	}

	if (orig_dx==orig_dy) {
		if (sy<0)
			w = -w;
		w += sx*sizeof(FBTYPE);
		for (i=dx;i>=0; i--) {
			*fb = color;
			fb = (FBTYPE *) (((char*)fb) + w);
		}
		return;
	}

	if (orig_dx >= orig_dy) { /* x major */
		int runlen,adjup,adjdown,e,len;
		int firstlen,lastlen;

		runlen = orig_dx/orig_dy;
		adjup = orig_dx%orig_dy;
		lastlen = firstlen = (runlen>>1) + 1;
		if (clip_first) { /* clipped, Adjust firstlen */
			int clip_dx = abs(clipped_x1 - orig_x1);
			int clip_dy = abs(clipped_y1 - orig_y1);
			int d = (2*clip_dy+1)*orig_dx;
			firstlen = d/(2*orig_dy) - clip_dx + 1;
			e = d%(2*orig_dy);
			if ((e==0) && (sy>0)) { /* Special case, arbitrary choise. Select lower pixel.(?) */
				firstlen--;
				e += 2*orig_dy;
			}
			e -= (orig_dy*2);
		} else { /* Not clipped, calculate start error term */
			e = adjup - (orig_dy<<1); /* initial errorterm == half a step */
			if ((runlen&1) != 0) {
				e += orig_dy;
			}
		}
		if (clip_last) { /* Last endpoint clipped */
			int clip_dx = abs(clipped_x2 - orig_x2);
			int clip_dy = abs(clipped_y2 - orig_y2);
			int d = (1+2*clip_dy)*orig_dx;
			lastlen = d/(2*orig_dy) - clip_dx + 1;
			if ((sy<0)  && ((d%(2*orig_dy))==0) ) /* special arbitrary case */
				lastlen--;
		}
		adjup <<= 1;
		adjdown = orig_dy<<1;

		if (sy>0) {  /* line goes down */
			if ((adjup==0) && ((runlen&1)==0) && (!clip_first)) {
				firstlen--;
			}
			if (sx>0) { /* line goes right */
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb++;
				}
				fb = (FBTYPE *) (((char*)fb) + w);
				for (i=dy-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb++;
					}
					fb = (FBTYPE *) (((char*)fb) + w);
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb++;
				}
				return;
			} else { /* line goes left */
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb--;
				}
				fb = (FBTYPE *) (((char*)fb) + w);
				for (i=dy-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb--;
					}
					fb = (FBTYPE *) (((char*)fb) + w);
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb--;
				}
				return;
			}
		} else { /* line goes up */
			if ((adjup==0) && ((runlen&1)==0) && (!clip_last)) {
				lastlen--;
			}
			if (sx>0) { /* line goes right */
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb++;
				}
				fb = (FBTYPE *) (((char*)fb) - w);
				for (i=dy-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>=0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb++;
					}
					fb = (FBTYPE *) (((char*)fb) - w);
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb++;
				}
				return;
			} else { /* line goes left */
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb--;
				}
				fb = (FBTYPE *) (((char*)fb) - w);
				for (i=dy-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>=0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb--;
					}
					fb = (FBTYPE *) (((char*)fb) - w);
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb--;
				}
				return;
			}
		}
	} else { /* y major */
		int runlen,adjup,adjdown,e,len;
		int firstlen,lastlen;

		runlen = orig_dy/orig_dx;
		adjup = orig_dy%orig_dx;

		lastlen = firstlen = (runlen>>1) + 1;
		if (clip_first) { /* clipped, Adjust firstlen */
			int clip_dx = abs(clipped_x1 - orig_x1);
			int clip_dy = abs(clipped_y1 - orig_y1);
			int d = (2*clip_dx+1)*orig_dy;
			firstlen = d/(2*orig_dx) - clip_dy + 1;
			e = d%(2*orig_dx);
			if ((e==0) && (sx>0)) { /* Special case, arbitrary choise. Select lower pixel.(?) */
				firstlen--;
				e += 2*orig_dx;
			}
			e -= (orig_dx*2);
		} else { /* Not clipped, calculate start error term */
			e = adjup - (orig_dx<<1); /* initial errorterm == half a step */
			if ((runlen&1) != 0) {
				e += orig_dx;
			}
		}
		if (clip_last) { /* Last endpoint clipped */
			int clip_dx = abs(clipped_x2 - orig_x2);
			int clip_dy = abs(clipped_y2 - orig_y2);
			int d = (1+2*clip_dx)*orig_dy;
			lastlen = d/(2*orig_dx) - clip_dy + 1;
			if ((sx<0)  && ((d%(2*orig_dx))==0) ) /* special arbitrary case */
				lastlen--;
		}
		adjup <<= 1;
		adjdown = orig_dx<<1;
		if (sy>0) { /* Line goes DOWN */
			if (sx>0) { /* line goes RIGHT */
				if ((adjup==0) && ((runlen&1)==0) && (!clip_first)) {
					firstlen--;
				}
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) + w);
				}
				fb++;
				for (i=dx-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb = (FBTYPE *) (((char*)fb) + w);
					}
					fb ++;
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) + w);
				}
				return;
			} else { /* line goes LEFT */
				if ((adjup==0) && ((runlen&1)==0) && (!clip_last)) {
					lastlen--;
				}
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) + w);
				}
				fb--;
				for (i=dx-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>=0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb = (FBTYPE *) (((char*)fb) + w);
					}
					fb --;
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) + w);
				}
				return;
			}
		} else { /* Line goes UP */
			if (sx>0) { /* line goes RIGHT */
				if ((adjup==0) && ((runlen&1)==0) && (!clip_first)) {
					firstlen--;
				}
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) - w);
				}
				fb++;
				for (i=dx-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb = (FBTYPE *) (((char*)fb) - w);
					}
					fb ++;
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) - w);
				}
				return;
			} else { /* line goes LEFT */
				if ((adjup==0) && ((runlen&1)==0) && (!clip_last)) {
					lastlen--;
				}
				for (;firstlen>0; firstlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) - w);
				}
				fb--;
				for (i=dx-1; i>0; i--) {
					len = runlen;
					e += adjup;
					if (e>=0) {
						len++;
						e -= adjdown;
					}
					for (;len>0; len--) {
						*fb = color;
						fb = (FBTYPE *) (((char*)fb) - w);
					}
					fb --;
				}
				for (;lastlen>0; lastlen--) {
					*fb = color;
					fb = (FBTYPE *) (((char*)fb) - w);
				}
				return;
			}
		}
	}

	target->unlock();
}
