/*
	The new pixelformat converters in ClanLib.
	
	Important:
	
	  Call Hermes_Init() before using these classes.
*/

#ifndef header_pixeldata
#define header_pixeldata

extern "C"
{
	#include <Hermes/H_Conv.h>
	#include <Hermes/H_Pal.h>
}

#include "surfaceprovider_32bpp.h"

class CL_PixelData
{
protected:
	HermesHandle m_handle;

	HermesFormat m_src_format;
	HermesHandle m_src_palette;

	HermesFormat m_dest_format;
	
	CL_SurfaceProvider *m_provider;
	bool m_delete_provider;

	unsigned char *scanline;

public:
	CL_PixelData(
		unsigned int red_mask,
		unsigned int green_mask,
		unsigned int blue_mask,
		unsigned int alpha_mask,
		CL_SurfaceProvider *src,
		unsigned int bytes_per_pixel)
	{
		m_delete_provider = false;

		// if PAL8, convert to RGBA8888 because Hermes doesn't support
		// colorkeys.
		if (src->get_pixel_format() == PAL8)
		{
			src = new CL_SurfaceProvider_32bpp(src);
			m_delete_provider = true;
		}

		m_provider = src;
	
		// Access surface provider until we're destructed.
		m_provider->lock();
		
		m_handle = Hermes_ConverterInstance(HERMES_CONVERT_NORMAL);
		cl_assert(m_handle!=0);

		m_src_palette = Hermes_PaletteInstance();
		cl_assert(m_src_palette != 0);

		// Set dest pixel format:
		
		m_dest_format.bits = bytes_per_pixel*8;
		m_dest_format.r = red_mask;
		m_dest_format.g = green_mask;
		m_dest_format.b = blue_mask;
		m_dest_format.a = alpha_mask;
		
		// only PAL8 colorkey support (for now):
		cl_assert(m_provider->get_transcol() == -1);

		// Set source pixel format:
		
		switch (src->get_pixel_format())
		{
		// PAL8 is handled above...
		
		case RGB565:
			m_src_format.bits = 16;
			m_src_format.indexed = false;
			m_src_format.r = ( 31 << (6+5) );
			m_src_format.g = ( 63 << (5) );
			m_src_format.b = 31;
			m_src_format.a = 0;
			break;

		case RGBA4444:
			m_src_format.bits = 16;
			m_src_format.indexed = false;
			m_src_format.r = ( 15 << (4+4+4) );
			m_src_format.g = ( 15 << (4+4) );
			m_src_format.b = ( 15 << (4) );
			m_src_format.a = 15;
			break;

		case RGBA5551:
			m_src_format.bits = 16;
			m_src_format.indexed = false;
			m_src_format.r = ( 31 << (5+5+1) );
			m_src_format.g = ( 31 << (5+1) );
			m_src_format.b = ( 31 << (1) );
			m_src_format.a = 1;
			break;

		case RGB888:
			m_src_format.bits = 24;
			m_src_format.indexed = false;
			m_src_format.r = ( 255 << (8+8) );
			m_src_format.g = ( 255 << (8) );
			m_src_format.b = 255;
			m_src_format.a = 0;
			break;

		case RGBA8888:
			m_src_format.bits = 32;
			m_src_format.indexed = false;
			m_src_format.r = ( 255 << (8+8+8) );
			m_src_format.g = ( 255 << (8+8) );
			m_src_format.b = ( 255 << (8) );
			m_src_format.a = 255;
			break;

		default:
			cl_assert(false); // unknown pixelformat.
		}
		
		scanline = new unsigned char[get_bytes_per_pixel()*get_width()+1];
	}

	virtual ~CL_PixelData()
	{
		Hermes_ConverterReturn(m_handle);
		Hermes_PaletteReturn(m_src_palette);

		delete[] scanline;

		// Release surface provider.
		m_provider->unlock();
		if (m_delete_provider) delete m_provider;
	}

	unsigned int get_width()
	{
		return m_provider->get_width();
	}

	unsigned int get_height()
	{
		return m_provider->get_height();
	}
	
	unsigned int get_bytes_per_pixel()
	{
		return (m_dest_format.bits+7)/8;
	}

	int get_bytes_pr_line() { return get_bytes_per_pixel()*get_width(); }
	
	void get_line_pixel_to_dest(int y, unsigned char *dest)
	{
		// Hack to support alpha when Hermes doesn't:

		if (y >=  m_provider->get_height()*m_provider->get_no_sprs())
		{
			cl_info(1, CL_String(CL_String("Tried to obtain line ")+y).get_string());
			cl_assert(false);
		}

		int width = get_width();

		if (m_dest_format.a == 255 && get_bytes_per_pixel()==1)
		{
			switch (m_src_format.bits)
			{
			case 15:
			case 16:
				{
					unsigned short *ptr =
						(unsigned short *) m_provider->get_surface_data();

					for (int x=0; x<width; x++)
					{
						dest[x] = (unsigned char) (ptr[x+y*width]&255);
					}
				}
				return;

			case 32:
				{
					unsigned int *ptr =
						(unsigned int *) m_provider->get_surface_data();

					for (int x=0; x<width; x++)
					{
						dest[x] = (unsigned char) (ptr[x+y*width]&255);
					}
				}
				return;
			}
		}
	
	
		int res = Hermes_ConverterRequest(
			m_handle,
			&m_src_format,
			&m_dest_format);

		if (res == 0)
		{
			cout << "Hermes convert FAILED:" << endl;
			cout << "  m_src_format.r = " << (unsigned int) m_src_format.r << endl;
			cout << "  m_src_format.g = " << (unsigned int) m_src_format.g << endl;
			cout << "  m_src_format.b = " << (unsigned int) m_src_format.b << endl;
			cout << "  m_src_format.a = " << (unsigned int) m_src_format.a << endl;
			cout << "  m_dest_format.r = " << (unsigned int) m_dest_format.r << endl;
			cout << "  m_dest_format.g = " << (unsigned int) m_dest_format.g << endl;
			cout << "  m_dest_format.b = " << (unsigned int) m_dest_format.b << endl;
			cout << "  m_dest_format.a = " << (unsigned int) m_dest_format.a << endl;
			cl_assert(res != 0);
		}

		res = Hermes_ConverterPalette(m_handle, m_src_palette, 0);
		if (res == 0)
		{
			cout << "Hermes convert FAILED:" << endl;
			cout << "  m_src_format.r = " << (unsigned int) m_src_format.r << endl;
			cout << "  m_src_format.g = " << (unsigned int) m_src_format.g << endl;
			cout << "  m_src_format.b = " << (unsigned int) m_src_format.b << endl;
			cout << "  m_src_format.a = " << (unsigned int) m_src_format.a << endl;
			cout << "  m_dest_format.r = " << (unsigned int) m_dest_format.r << endl;
			cout << "  m_dest_format.g = " << (unsigned int) m_dest_format.g << endl;
			cout << "  m_dest_format.b = " << (unsigned int) m_dest_format.b << endl;
			cout << "  m_dest_format.a = " << (unsigned int) m_dest_format.a << endl;
			cl_assert(res != 0);
		}
		
		if (width % 2 == 1) // Hermes bug - workaround.
		{
			void *bug_buffer = new char[m_provider->get_pitch()+1];
			memcpy(
				bug_buffer,
				((unsigned char *) m_provider->get_surface_data())+m_provider->get_pitch()*y,
				m_provider->get_pitch());

			res = Hermes_ConverterCopy(
				m_handle,
				bug_buffer,
				0,
				0,
				width+1,
				1, // m_provider->get_height(),
				m_provider->get_pitch(),
				dest,
				0,
				0,
				width,
				1, // m_provider->get_height(),
				get_bytes_per_pixel()*width
				);

			delete[] bug_buffer;
		}
		else
		{
			res = Hermes_ConverterCopy(
				m_handle,
				((unsigned char *) m_provider->get_surface_data())+m_provider->get_pitch()*y,
				0,
				0,
				width,
				1, // m_provider->get_height(),
				m_provider->get_pitch(),
				dest,
				0,
				0,
				width,
				1, // m_provider->get_height(),
				get_bytes_per_pixel()*width
				);
		}
			
		if (res == 0)
		{
			cout << "Hermes convert FAILED:" << endl;
			cout << "  m_src_format.r = " << (unsigned int) m_src_format.r << endl;
			cout << "  m_src_format.g = " << (unsigned int) m_src_format.g << endl;
			cout << "  m_src_format.b = " << (unsigned int) m_src_format.b << endl;
			cout << "  m_src_format.a = " << (unsigned int) m_src_format.a << endl;
			cout << "  m_dest_format.r = " << (unsigned int) m_dest_format.r << endl;
			cout << "  m_dest_format.g = " << (unsigned int) m_dest_format.g << endl;
			cout << "  m_dest_format.b = " << (unsigned int) m_dest_format.b << endl;
			cout << "  m_dest_format.a = " << (unsigned int) m_dest_format.a << endl;
			cl_assert(res != 0);
		}
	}

	unsigned char *get_line_pixel(int y)
	{
		get_line_pixel_to_dest(y, scanline);
		return scanline;
	}
	
	unsigned short *get_line_pixel16(int y)
	{
		return (unsigned short *) get_line_pixel(y);
	}
	
	unsigned int *get_line_pixel32(int y)
	{
		return (unsigned int *) get_line_pixel(y);
	}
};

class CL_PixelData_Palette
{
public:
	CL_PixelData_Palette(
		CL_Palette * /*dest_pal*/,
		CL_SurfaceProvider * /*src*/)
	{
		cl_assert(false); // not implemented yet.
	}
	
	virtual ~CL_PixelData_Palette()
	{
	}
	
	unsigned int get_width()
	{
		return 0;
	}

	unsigned int get_height()
	{
		return 0;
	}
	
	bool has_colorkey()
	{
		return false;
	}

	unsigned int get_colorkey()
	{
		return 0;
	}
	
	unsigned char *get_line_pixel(int /*y*/)
	{
		return NULL;
	}
};

#endif
