/*
	$Id: surface_generic.cpp,v 1.3 1999/06/04 08:35:15 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#include <stdlib.h>

#include "Layer1/Display/API/display.h"
#include "Layer1/Display/API/surface.h"
#include "surface_generic.h"
#include "cardsurface.h"
#include "displaycard_generic.h"

CL_SurfacePriority *CL_Surface_Generic::priority_handler = NULL;

CL_Surface *CL_Surface::create(
	CL_SurfaceProvider *provider,
	bool delete_provider)
{
	CL_Surface *surf = new CL_Surface_Generic(provider, delete_provider);
	if (CL_Surface_Generic::priority_handler != NULL) 
	{
		return CL_Surface_Generic::priority_handler->manage_surface(surf);
	}
	else
	{
		provider->lock();
		if (!surf->convert_video(CL_Display::get_current_card()))
			surf->convert_system(CL_Display::get_current_card());
		provider->unlock();
	}
	return surf;
}

void CL_Surface::set_priority_handler(
	CL_SurfacePriority *_priority_handler)
{
	CL_Surface_Generic::priority_handler = _priority_handler;
}

CL_Surface_Generic::CL_Surface_Generic(
	CL_SurfaceProvider *_provider,
	bool _delete_provider)
{
	provider = _provider;
	delete_provider = _delete_provider;
	translate_x = 0;
	translate_y = 0;

	card_surfaces = new CL_CardSurface*[CL_Display::cards.get_num_items()];
	for (int i=0; i<CL_Display::num_cards; i++)
	{
		card_surfaces[i] = NULL;
	}
}

CL_Surface_Generic::~CL_Surface_Generic()
{
	for (int i=0; i<CL_Display::num_cards; i++) delete card_surfaces[i];
	delete[] card_surfaces;

	if (delete_provider == true) delete provider;
}

inline int CL_Surface_Generic::ensure_surface_prepared(CL_DisplayCard *card)
{
	int card_no = card == NULL ? CL_Display::get_current_card()->get_card_no() : card->get_card_no();
	if (card_surfaces[card_no] == NULL)
	{
		provider->lock();

		translate_x = provider->get_translate_x();
		translate_y = provider->get_translate_y();

		width = provider->get_width();
		height = provider->get_height();
		no_sprs = provider->get_no_sprs();

		CL_DisplayCard_Generic *gen = ((CL_DisplayCard_Generic *) CL_Display::cards[card_no]);
		card_surfaces[card_no] = gen->create_cardsurface_video(provider);
		if (card_surfaces[card_no] == NULL) card_surfaces[card_no] = gen->create_cardsurface_system(provider);

		provider->unlock();
	}

	return card_no;
}

CL_CardSurface *CL_Surface_Generic::get_cardsurface(int surface_num) 
{ 
	return card_surfaces[ensure_surface_prepared(CL_Display::cards[surface_num])]; 
}

CL_CardSurface *CL_Surface_Generic::get_cardsurface(CL_DisplayCard *card) 
{ 
	return card_surfaces[ensure_surface_prepared(card)];
}

void CL_Surface_Generic::flush(CL_DisplayCard *card)
{
	if (card == NULL)
	{
		delete card_surfaces[CL_Display::get_current_card()->get_card_no()];
		card_surfaces[CL_Display::get_current_card()->get_card_no()] = NULL;
	}
	else
	{
		delete card_surfaces[card->get_card_no()];
		card_surfaces[card->get_card_no()] = NULL;
	}
}

void CL_Surface_Generic::reload()
{
	bool provider_locked = false;

	for (int i=0; i<CL_Display::num_cards; i++)
	{
		if (card_surfaces[i] != NULL)
		{
			if (!provider_locked)
			{
				provider->lock();
				translate_x = provider->get_translate_x();
				translate_y = provider->get_translate_y();

				width = provider->get_width();
				height = provider->get_height();
				no_sprs = provider->get_no_sprs();

				provider_locked = true;
			}
			card_surfaces[i]->reload();
		}
	}

	if (provider_locked) provider->unlock();
}

bool CL_Surface_Generic::is_video(CL_DisplayCard *card) const
{
	int card_no = card == NULL ? CL_Display::get_current_card()->get_card_no() : card->get_card_no();

	return card_surfaces[card_no] != NULL && card_surfaces[card_no]->is_video();
}

bool CL_Surface_Generic::is_loaded(CL_DisplayCard *card) const
{
	int card_no = card == NULL ? CL_Display::get_current_card()->get_card_no() : card->get_card_no();

	return card_surfaces[card_no] != NULL;
}

bool CL_Surface_Generic::convert_video(CL_DisplayCard *card)
{
	int card_no = card == NULL ? CL_Display::get_current_card()->get_card_no() : card->get_card_no();

	translate_x = provider->get_translate_x();
	translate_y = provider->get_translate_y();

	width = provider->get_width();
	height = provider->get_height();
	no_sprs = provider->get_no_sprs();

	if (card_surfaces[card_no] == NULL)
	{
		CL_DisplayCard_Generic *gen = ((CL_DisplayCard_Generic *) CL_Display::cards[card_no]);
		card_surfaces[card_no] = gen->create_cardsurface_video(provider);

		if (card_surfaces[card_no] == NULL) return false;
	}
	else
	{
		if (card_surfaces[card_no]->is_video()) return true;

		if (card_surfaces[card_no]->can_convert_video())
		{
			return card_surfaces[card_no]->convert_video();
		}
		else
		{
			return false;
		}
	}
	return true;
}

bool CL_Surface_Generic::convert_system(CL_DisplayCard *card)
{
	int card_no = card == NULL ? CL_Display::get_current_card()->get_card_no() : card->get_card_no();

	if (card_surfaces[card_no] == NULL)
	{
		CL_DisplayCard_Generic *gen = ((CL_DisplayCard_Generic *) CL_Display::cards[card_no]);
		card_surfaces[card_no] = gen->create_cardsurface_system(provider);

		if (card_surfaces[card_no] == NULL) return false;
	}
	else
	{
		if (!card_surfaces[card_no]->is_video()) return true;

		if (card_surfaces[card_no]->can_convert_system())
		{
			return card_surfaces[card_no]->convert_system();
		}
		else
		{
			return false;
		}
	}
	return true;
}

CL_SurfaceProvider *CL_Surface_Generic::get_provider() const
{
	return provider;
}

void CL_Surface_Generic::put_screen(
	int x,
	int y,
	int spr_no,
	CL_DisplayCard *card)
{
	card_surfaces[ensure_surface_prepared(card)]->put_screen(
		x+translate_x,
		y+translate_y,
		spr_no);
}

void CL_Surface_Generic::put_screen(
	int x,
	int y,
	float scale_x,
	float scale_y,
	int spr_no,
	CL_DisplayCard *card)
{
	card_surfaces[ensure_surface_prepared(card)]->put_screen(
		x+translate_x,
		y+translate_y,
		scale_x,
		scale_y,
		spr_no);
}

void CL_Surface_Generic::put_screen(
	int x,
	int y,
	int size_x,
	int size_y,
	int spr_no,
	CL_DisplayCard *card)
{
	card_surfaces[ensure_surface_prepared(card)]->put_screen(
		x+translate_x,
		y+translate_y,
		size_x,
		size_y,
		spr_no);
}
