/*
	$Id: keyboard_fbdev.cpp,v 1.11 1999/08/24 21:24:25 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#ifndef USE_TTY_INPUT
#ifdef USE_SVGALIB
#define USE_TTY_INPUT
#else
#ifdef USE_FBDEV
#define USE_TTY_INPUT
#endif
#endif
#endif

#ifdef USE_TTY_INPUT

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <unistd.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#ifdef HAVE_SYS_KD_H
#include <sys/kd.h>
#else
#include <linux/kd.h>
#endif
#ifdef HAVE_SYS_VT_H
#include <sys/vt.h>
#else
#include <linux/vt.h>
#endif
#include <linux/tty.h>
#include <linux/keyboard.h>

#include "keyboard_fbdev.h"

#include "Display/API/display.h"
#include "System/Unix/implementation_fbdev.h"
#include "Display/FBDev/display_fbdev.h"
#include "System/API/cl_assert.h"
#include "System/API/error.h"

#include "Layer1/System/Generic/string_asm.h"

/******************
  Keyboard
******************/

CL_FBDevKeyboard::CL_FBDevKeyboard()
{
	memset( &keymap, 0, sizeof(keymap) );

	CL_System_Generic::keep_alives.add(this);

	/* open the tty */
	fd = open("/dev/tty", O_RDWR | O_NONBLOCK);

	if (fd < 0)
	{
		throw CL_Error("Couldn't open /dev/tty.");
	}

	/* put tty into "straight through" mode. */
	struct termios newterm;

	if (tcgetattr(fd, &old_termios) < 0)
	{
		perror("tcgetattr failed");
	}

	memcpy ( &newterm, &old_termios, sizeof(termios) );

	newterm.c_lflag &= ~(ICANON | ECHO  | ISIG);
	newterm.c_iflag &= ~(ISTRIP | IGNCR | ICRNL | INLCR | IXOFF | IXON);
	newterm.c_iflag |= IGNBRK;
	newterm.c_cc[VMIN]  = 0;
	newterm.c_cc[VTIME] = 0;

	if (tcsetattr(fd, TCSANOW, &newterm) < 0)
	{
		perror("tcsetattr failed");
	}

	/* save old mode and set to mediumraw */
	if (ioctl(fd, KDGKBMODE, &old_mode) < 0)
	{
		perror("Couldn't get keyboard mode");
		old_mode = K_XLATE;
	}
	if (ioctl(fd, KDSKBMODE, K_MEDIUMRAW) < 0)
	{
		perror("Couldn't set keyboard mode to K_MEDIUMRAW");
	}
	
	/* save old kd mode and set to graphics */
	if (ioctl(fd, KDGETMODE, &old_kd) < 0)
	{
		perror("Couldn't get kd mode");
		old_kd = KD_TEXT;
	}
	if (ioctl(fd, KDSETMODE, KD_GRAPHICS) < 0)
	{
		perror("Couldn't set kd mode to KD_GRAPHICS");
	}
}

CL_FBDevKeyboard::~CL_FBDevKeyboard()
{
	CL_System_Generic::keep_alives.del(this);
	restore();
#ifdef USE_FBDEV
	CL_Implementation_FBDev::clean_up();
#endif

	int num = buttons.get_num_items();
	for (int i=0; i<num; i++) delete buttons[i];
}

void CL_FBDevKeyboard::restore()
{
	if (fd>=0)
	{
		ioctl(fd, KDSKBMODE, old_mode);
		ioctl(fd, KDSETMODE, old_kd);
		if (tcsetattr(fd, TCSANOW, &old_termios) < 0)
		{
			cerr << "Could not restore old terminal input settings! Please run 'reset'!" << endl;
			perror( "System error message" );
		}
		close(fd);
	}
}

bool CL_FBDevKeyboard::keep_alive()
{
	int readlen;
	unsigned char buf[256];
	
	// Read keyboard data
	while ((readlen = read(fd, buf, 256)) > 0) {
		for (int i = 0; i < readlen; i++)
		{
			handle_code( buf[i] );
		}
	}

	return false;
}

void CL_FBDevKeyboard::handle_code(char code)
{
	bool keydown;
	kbentry entry;
	
	if (code & 0x80)
	{
		code &= 0x7f;
		keydown = false;
	} else
	{
		keydown = true;
	}
			
	//fetch the keycode
	entry.kb_table = 0;
	entry.kb_index = code;
	ioctl(fd,KDGKBENT,&entry);
	
	CL_Key key;
	key.id = translate(entry.kb_value);
	key.ascii = -1;
	key.state = keydown ? CL_Key::Pressed : CL_Key::Released;

	buffer.add_key( key );
	keymap[translate(entry.kb_value)] = keydown;
}

char CL_FBDevKeyboard::translate(int kb_value)
{
	switch (kb_value)
	{
	case K_F1:	return KEY_F1;
	case K_F2:	return KEY_F2;
	case K_F3:	return KEY_F3;
	case K_F4:	return KEY_F4;
	case K_F5:	return KEY_F5;
	case K_F6:	return KEY_F6;
	case K_F7:	return KEY_F7;
	case K_F8:	return KEY_F8;
	case K_F9:	return KEY_F9;
	case K_F10:	return KEY_F10;
	case K_F11:	return KEY_F11;
	case K_F12:	return KEY_F12;
	
	case 2816+'a':	return KEY_A;
	case 2816+'b':	return KEY_B;
	case 2816+'c':	return KEY_C;
	case 2816+'d':	return KEY_D;
	case 2816+'e':	return KEY_E;
	case 2816+'f':	return KEY_F;
	case 2816+'g':	return KEY_G;
	case 2816+'h':	return KEY_H;
	case 2816+'i':	return KEY_I;
	case 2816+'j':	return KEY_J;
	case 2816+'k':	return KEY_K;
	case 2816+'l':	return KEY_L;
	case 2816+'m':	return KEY_M;
	case 2816+'n':	return KEY_N;
	case 2816+'o':	return KEY_O;
	case 2816+'p':	return KEY_P;
	case 2816+'q':	return KEY_Q;
	case 2816+'r':	return KEY_R;
	case 2816+'s':	return KEY_S;
	case 2816+'t':	return KEY_T;
	case 2816+'u':	return KEY_U;
	case 2816+'v':	return KEY_V;
	case 2816+'w':	return KEY_W;
	case 2816+'x':	return KEY_X;
	case 2816+'y':	return KEY_Y;
	case 2816+'z':	return KEY_Z;
	
	case 48:		return KEY_0;
	case 49:		return KEY_1;
	case 50:		return KEY_2;
	case 51:		return KEY_3;
	case 52:		return KEY_4;
	case 53:		return KEY_5;
	case 54:		return KEY_6;
	case 55:		return KEY_7;
	case 56:		return KEY_8;
	case 57:		return KEY_9;
	
	case 27:		return KEY_ESCAPE;
	case K_LEFT:	return KEY_LEFT;
	case K_RIGHT:	return KEY_RIGHT;
	case K_UP:	return KEY_UP;
	case K_DOWN:	return KEY_DOWN;
	case K_ENTER:	return KEY_ENTER;

	case K_CTRLL:	return KEY_LCTRL;
	case K_CTRLR:	return KEY_RCTRL;
	case K_SHIFTL:	return KEY_LSHIFT;
	case K_SHIFTR:	return KEY_RSHIFT;
	case K_ALT:	return KEY_ALT;
	case K_ALTGR:	return KEY_ALTGR;
	case 9:		return KEY_TAB;
	case 32:		return KEY_SPACE;
	case 127:		return KEY_BACKSPACE;
	case K_INSERT:	return KEY_INSERT;
	case K_REMOVE:	return KEY_DELETE;
	case K_FIND:	return KEY_HOME;
	case K_SELECT:	return KEY_END;
	case K_PGUP:	return KEY_PAGEUP;
	case K_PGDN:	return KEY_PAGEDOWN;
	case K_CAPS:	return KEY_CAPSLOCK;
	case K_NUM:	return KEY_NUMLOCK;
	case K_HOLD:	return KEY_SCRLOCK;
	case 28:		return KEY_PRINT;
	case K(1,29):	return KEY_PAUSE;
	case K_PSLASH:	return KEY_KP_DIV;
	case K_PSTAR:	return KEY_KP_MULT;
	case K_PMINUS:	return KEY_KP_MINUS;
	case K_PPLUS:	return KEY_KP_PLUS;
	case K_PENTER:	return KEY_KP_ENTER;
	}
	
	return KEY_NONE_OF_THE_ABOVE;
}

char *CL_FBDevKeyboard::get_name() const
{
	return "FBDev keyboard";
}

int CL_FBDevKeyboard::get_num_buttons() const
{
	return NUM_KEYS;
}

CL_InputButton *CL_FBDevKeyboard::get_button(int button_num)
{
	if (buttons[button_num] != NULL) return buttons[button_num];
	
	buttons.add( new CL_InputButton_FBDevKeyboard(button_num, keymap), button_num );

	return buttons[button_num];
}

int CL_FBDevKeyboard::get_num_axes() const
{
	return 0;
}

CL_InputAxis *CL_FBDevKeyboard::get_axis(int /*axis_num*/)
{
	return NULL;
}

int CL_FBDevKeyboard::get_num_hats() const
{
	return 0;
}

CL_InputHat *CL_FBDevKeyboard::get_hat(int /*hat_num*/)
{
	return NULL;
}

int CL_FBDevKeyboard::get_num_buffers() const
{
	return 1;
}

CL_InputBuffer *CL_FBDevKeyboard::get_buffer(int /*buffer_num*/)
{
	return &buffer;
}

int CL_FBDevKeyboard::get_num_cursors() const
{
	return 0;
}

CL_InputCursor *CL_FBDevKeyboard::get_cursor(int /*cursor_num*/)
{
	return NULL;
}

/***************************
  CL_InputBuffer_FBDevKeyboard
***************************/

CL_InputBuffer_FBDevKeyboard::CL_InputBuffer_FBDevKeyboard()
{
	m_begin = 0;
	m_end = 0;
	m_buffer_size = 1024;
	m_buffer = new CL_Key[m_buffer_size];
}

CL_InputBuffer_FBDevKeyboard::~CL_InputBuffer_FBDevKeyboard()
{
	delete[] m_buffer;
}

CL_Key CL_InputBuffer_FBDevKeyboard::peek_key() const
{
	if (m_begin == m_end)
	{
		CL_Key k;
		k.state = CL_Key::NoKey;
		return k;
	}

	return m_buffer[m_begin];
}

CL_Key CL_InputBuffer_FBDevKeyboard::get_key()
{
	if (m_begin == m_end)
	{
		CL_Key k;
		k.state = CL_Key::NoKey;
		return k;
	}

	return (m_buffer[m_begin++]);
}

int CL_InputBuffer_FBDevKeyboard::keys_left()
{
	return m_end - m_begin;
}

void CL_InputBuffer_FBDevKeyboard::clear()
{
	m_begin = 0;
	m_end = 0;
}

void CL_InputBuffer_FBDevKeyboard::add_key(CL_Key key)
{
	if (m_end == m_buffer_size) // reached end of buffer
	{
		if (m_begin == 0) return; // totally full! - discard

		// Move contents to beginning of buffer...
		asm_memmove(
			(char *) m_buffer,
			(char *) (m_buffer+m_begin),
			(m_end-m_begin)*sizeof(CL_Key));

		m_end = m_end-m_begin;
		m_begin = 0;
	}
	m_buffer[m_end++] = key;
}

/***************************
  CL_InputButton_FBDevKeyboard
***************************/

CL_InputButton_FBDevKeyboard::CL_InputButton_FBDevKeyboard(
	int _key, char *_keymap)
{
	key = _key;
	keymap = _keymap;
}

bool CL_InputButton_FBDevKeyboard::is_pressed()
{
	return keymap[key];
}

#endif
