/*
	$Id: mouse_x11.cpp,v 1.10 1999/08/10 23:29:03 grumbel Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#ifdef USE_X11

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iostream.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "System/API/cl_assert.h"

#include "mouse_x11.h"

#include "Display/API/display.h"
#include "System/Unix/implementation_xwindow.h"
#include "Display/X11/display_xwindow.h"



/*************************************
  CL_Mouse_XWin
*************************************/

CL_Mouse_XWin::CL_Mouse_XWin(CL_XWindow_CompatibleCard *_card)
{
	card = _card;

	cursor = new CL_InputCursor_Mouse_XWin(card);
	axes = new CL_InputAxis_Mouse_XWin[2]; 

	buttons.add(new CL_InputButton_Mouse_XWin(),0);
	buttons.add(new CL_InputButton_Mouse_XWin(),1);
	buttons.add(new CL_InputButton_Mouse_XWin(),2);  

	CL_System_Generic::keep_alives.add(this);
}

CL_Mouse_XWin::~CL_Mouse_XWin()
{
	delete cursor;

	int num_buttons = buttons.get_num_items();
	for (int i=0; i<num_buttons; i++) delete buttons[i];

	CL_System_Generic::keep_alives.del(this);
}


int CL_Mouse_XWin::get_num_buttons() const
{
	return 3;
}

CL_InputButton *CL_Mouse_XWin::get_button(int button_num)
{
	if (buttons[button_num] == NULL)
		buttons.add(new CL_InputButton_Mouse_XWin(),button_num);
	return buttons[button_num];
}

int CL_Mouse_XWin::get_num_axes() const
{
	return 2;
}

CL_InputAxis *CL_Mouse_XWin::get_axis(int axis_num)
{
	cl_assert(((unsigned int) axis_num)<2);
	return &axes[axis_num];
}

int CL_Mouse_XWin::get_num_hats() const
{
	return 0;
}

CL_InputHat *CL_Mouse_XWin::get_hat(int /*hat_num*/)
{
	return NULL;
}

int CL_Mouse_XWin::get_num_buffers() const
{
	return 0;
}

CL_InputBuffer *CL_Mouse_XWin::get_buffer(int /*buffer_num*/)
{
	return NULL;
}

int CL_Mouse_XWin::get_num_cursors() const
{
	return 1;
}

bool CL_Mouse_XWin::keep_alive()
{
	if (card->is_initialized() == false) return false;

	Window child;
	Window root;
	int root_x;
	int root_y;
	int win_x;
	int win_y;
	unsigned int mask;

	axes[0].center = (card->get_width())/2.0;
	axes[1].center = (card->get_height())/2.0;

	// Check if window has focus right now:
	Window focus_win;
	int focus_state;
	XGetInputFocus(
		card->get_dpy(),
		&focus_win,
		&focus_state);

	XQueryPointer(card->get_dpy(),
		      card->get_window(),
		      &root,
		      &child,
		      &root_x,
		      &root_y,
		      &win_x,
		      &win_y,
		      &mask);

	// get_window() returns NULL if in DGA mode.	
	if (card->get_window() == 0 || card->get_window() == focus_win) {
	
	  //the mouse is on our display:
	  cursor->x = win_x ;
	  cursor->y = win_y ;
	
	  axes[0].pos = cursor->x;
	  axes[1].pos = cursor->y;
		
	  if(mask & 0x100) buttons[0]->button_state = true;
	  else buttons[0]->button_state = false;
		
	  if(mask & 0x200) buttons[1]->button_state = true;
	  else buttons[1]->button_state = false;

	  if(mask & 0x400) buttons[2]->button_state = true;
	  else buttons[2]->button_state = false;
	}
	else // we don't have window focus. keep last coordinates and release buttons
	{
	  XQueryPointer(card->get_dpy(),
			card->get_window(),
			&root,
			&child,
			&root_x,
			&root_y,
			&win_x,
			&win_y,
			&mask);

	  cursor->x = win_x ;
	  cursor->y = win_y ;
	
	  if (cursor->x<0) cursor->x = 0;
	  if (cursor->y<0) cursor->y = 0;
	
	  if (cursor->x>cursor->get_max_x()) cursor->x = cursor->get_max_x();
	  if (cursor->y>cursor->get_max_y()) cursor->y = cursor->get_max_y();
	
	  buttons[0]->button_state = false;
	  buttons[1]->button_state = false;
	  buttons[2]->button_state = false;
	}

	return false;
}

CL_InputCursor *CL_Mouse_XWin::get_cursor(int /*cursor_num*/)
{
	return cursor;
}

/*******************************
  CL_InputButton_Mouse_XWin
*******************************/

CL_InputButton_Mouse_XWin::CL_InputButton_Mouse_XWin()
{
	button_state = false;
}

CL_InputButton_Mouse_XWin::~CL_InputButton_Mouse_XWin()
{
}

bool CL_InputButton_Mouse_XWin::is_pressed()
{
	return button_state;
}

/*******************************
  CL_InputCursor_Mouse_XWin
*******************************/

CL_InputCursor_Mouse_XWin::CL_InputCursor_Mouse_XWin(
	CL_XWindow_CompatibleCard *_card)
{
	card = _card;
	x = 0;
	y = 0;
}

CL_InputCursor_Mouse_XWin::~CL_InputCursor_Mouse_XWin()
{
}


float CL_InputCursor_Mouse_XWin::get_x()
{
	return x;
}

float CL_InputCursor_Mouse_XWin::get_y()
{
	return y;
}

float CL_InputCursor_Mouse_XWin::get_max_x()
{
	return card->get_width();
}

float CL_InputCursor_Mouse_XWin::get_max_y()
{
	return card->get_height();
}

/*******************************
  CL_InputAxis_Mouse_XWin
*******************************/

CL_InputAxis_Mouse_XWin::CL_InputAxis_Mouse_XWin()
{
	center = 1;
	pos = 0;
}

CL_InputAxis_Mouse_XWin::~CL_InputAxis_Mouse_XWin()
{
}


float CL_InputAxis_Mouse_XWin::get_pos()
{
	return (pos - center) / center;
}

#endif
