
#ifdef USE_NETWORK

#include "netgame_client.h"
#include "Layer1/System/API/cl_assert.h"
#include "Layer1/System/API/error.h"

CL_NetGame_Client::CL_NetGame_Client(
	int ip_addr,
	int port,
	CL_String &game_id,
	CL_Network_Generic *network)
: CL_NetGame_Generic(network)
{
	mutex = CL_Mutex::create();

	tcp_connection = network->provider->create_tcp_connection(
		ip_addr,
		port);

	if (tcp_connection == NULL)
	{
		throw CL_Error("Could not connect to host.");
	}

	udp_connection = NULL; // still no support for udp in client.

	// client code currently only knows about the server.
	all.computers.add(&server);

	// do "hello to you too" handshake.
	CL_ConnectionPacket p = tcp_connection->receive();
	while (p.size == 0) // wait until we get a hello packet.
	{
		network->provider->wait_for_connection_data(mutex);
		p = tcp_connection->receive();
	}
	
	CL_InputSource_MemoryGeneric input(p.data, p.size, true);
	if (input.read_int32() != Packet_Hello)
	{
		throw CL_Error("Protocol error. Didn't get a Hello.");
	}
	our_id = input.read_int32();

	int size = input.read_int32();
	if (size > 1000)
		throw CL_Error("Protocol error. Game ID size above 1000.");

	char *str = new char[size+1];
	str[size] = 0; // null-terminate.
	input.read(str, size);
	
	if (game_id != str) // not same game_id. probably an other clanlib game.
	{
		delete[] str;
		throw CL_Error("Wrong netgame id.");
	}

	delete[] str;
	
	// say "HelloToYouToo" to officially join the netgame:
	CL_OutputSource_MemoryGeneric output;
	output.write_int32(Packet_Hello_ToYouToo);
	tcp_connection->send(
		CL_ConnectionPacket(
			output.get_data(),
			output.size()));

	thread = CL_Thread::create(run_thread);
	thread->start(this);
}

CL_NetGame_Client::~CL_NetGame_Client()
{
	mutex->enter();
	delete thread;
	mutex->leave();

	delete tcp_connection;
	delete udp_connection;

	// clean up connections to clients:
	{
		CL_Iterator<CL_NetChannelQueue_Client> it(netchannels);
		while (it.next() != NULL) delete it();
	}
/*
	should be enabled when the client knows about other clients...

	while (leave_queue.is_empty() == false)
	{
		delete leave_queue.get();
	}
*/

	delete mutex;
}

const CL_NetComputer *CL_NetGame_Client::get_server() const
{
	return &server;
}

const CL_NetGroup *CL_NetGame_Client::get_all() const
{
	return &all;
}

bool CL_NetGame_Client::peek(int channel) const
{
	CL_MutexSection mutex_section(mutex);

	CL_NetChannelQueue_Client *queue = find_queue(channel);
	if (queue != NULL)
	{
		return !queue->is_empty();
	}

	return false;
}

CL_NetMessage CL_NetGame_Client::receive(int channel, int timeout)
{
	CL_MutexSection mutex_section(mutex);

	static bool warning = true;
	if (warning)
	{
		cl_info(info_network, "timeout parameter ignored: not implemented yet!");
		warning = false;
	}

	CL_NetChannelQueue_Client *queue = find_queue(channel);
	if (queue != NULL)
	{
		if (queue->is_empty()) throw CL_Error("No message to receive!");
		CL_NetMessage msg = queue->get();
		if (msg.size <= 0) cout << "*** HEY, QUEUE IS BROKEN ***" << endl;
		return msg;
	}
	
	throw CL_Error("No message to receive!");
	return CL_NetMessage(NULL, 0);
}

void CL_NetGame_Client::send(
	const int dest_channel,
	const CL_NetComputer *dest,
	const CL_NetMessage &message,
	bool reliable /*= true*/)
{
	CL_NetGroup group;
	group.computers.add((CL_NetComputer *) dest);

	send(
		dest_channel,
		&group,
		message,
		reliable);
}

void CL_NetGame_Client::send(
	const int dest_channel,
	const CL_NetGroup *dest,
	const CL_NetMessage &message,
	bool reliable /*= true*/)
{
	CL_MutexSection mutex_section(mutex);

	static bool warning = true;
	if (warning && reliable == false)
	{
		cl_info(info_network, "cannot send data unreliable (udp): not implemented yet!");
		warning = false;
	}
	
	CL_NetChannelQueue_Client *queue = find_queue(dest_channel);

	if (queue == NULL)
		return; // nobody has read or write access. Don't send message.

	if ((queue->access & ACCESS_CHANNEL_WRITE) != ACCESS_CHANNEL_WRITE)
		throw CL_Error("No write access to netchannel.");

	// permissions ok, send message:
	CL_OutputSource_MemoryGeneric output;
	output.write_int32(Packet_NetChannel_Message_ToServer);
	output.write_int32(dest_channel);

	// todo: write the computer id's on destination computers here.

	output.write_int32(message.size);
	output.write(message.data, message.size);

	CL_ConnectionPacket packet;
	packet.size = output.size();
	packet.data = output.get_data();

	tcp_connection->send(packet);
}

CL_NetComputer *CL_NetGame_Client::receive_computer_leave()
{
	return NULL;
}

const CL_NetComputer *CL_NetGame_Client::receive_computer_join()
{
	return NULL;
}

const CL_NetComputer *CL_NetGame_Client::receive_computer_rejoin()
{
	return NULL;
}

bool CL_NetGame_Client::receive_game_closed()
{
	CL_MutexSection mutex_section(mutex);
	return tcp_connection->connection_lost();
}

int CL_NetGame_Client::access_status(int channel) const
{
	CL_MutexSection mutex_section(mutex);

	CL_NetChannelQueue_Client *queue = find_queue(channel);
	return (queue == NULL) ? 0 : queue->access;
}

bool CL_NetGame_Client::is_writable(int channel) const
{
	CL_MutexSection mutex_section(mutex);

	CL_NetChannelQueue_Client *queue = find_queue(channel);
	if (queue == NULL) return false;

	return (queue->access & ACCESS_CHANNEL_WRITE) == ACCESS_CHANNEL_WRITE;
}

bool CL_NetGame_Client::is_readable(int channel) const
{
	CL_MutexSection mutex_section(mutex);

	CL_NetChannelQueue_Client *queue = find_queue(channel);
	if (queue == NULL) return false;

	return (queue->access & ACCESS_CHANNEL_READ) == ACCESS_CHANNEL_READ;
}

// Client side only:
int CL_NetGame_Client::receive_access_changed()
{
	CL_MutexSection mutex_section(mutex);

	if (access_queue.is_empty()) return -1;
	return access_queue.get()->channel_id;
}

// Server side only:
void CL_NetGame_Client::set_access(
	int /*channel*/,
	const CL_NetComputer * /*computer*/,
	int access_rights /*=ACCESS_CHANNEL_READ|ACCESS_CHANNEL_WRITE*/)
{
	throw CL_Error("We are not the network server. Cannot change access.");
}

void CL_NetGame_Client::set_access(
	int /*channel*/,
	const CL_NetGroup * /*group*/,
	int access_rights /*=ACCESS_CHANNEL_READ|ACCESS_CHANNEL_WRITE*/)
{
	throw CL_Error("We are not the network server. Cannot change access.");
}

void CL_NetGame_Client::keep_alive()
{
	CL_MutexSection mutex_section(mutex);
	while (tcp_connection->peek())
	{
		CL_ConnectionPacket msg = tcp_connection->receive();
			
		CL_InputSource_MemoryGeneric input(msg.data, msg.size, true);
		switch (input.read_int32())
		{
		case Packet_NetChannel_AccessChange:
			{
				CL_NetChannelQueue_Client *queue = create_queue(input.read_int32());
				queue->access = input.read_int32();
			}
			break;
		
		case Packet_NetChannel_Message_ToClient:
			{
				CL_NetChannelQueue_Client *queue = create_queue(input.read_int32());
				CL_NetMessage game_msg;
				game_msg.size = input.read_int32();
				game_msg.data = new char[game_msg.size];
				game_msg.from = &server;
				if (game_msg.size <= 0) cout << "*** WHAT THE HECK!? ***" << endl;
				input.read(game_msg.data, game_msg.size);
				queue->put(game_msg);
			}
			break;

		default:
			cl_info(info_network, "Network Protocol error!");
			cl_assert(false);
		}
	}
}

CL_NetChannelQueue_Client *CL_NetGame_Client::find_queue(int netchannel) const
{
	CL_MutexSection mutex_section(mutex);

	CL_Iterator<CL_NetChannelQueue_Client> it(
		(CL_List<CL_NetChannelQueue_Client> *) &netchannels);
	while (it.next() != NULL)
	{
		if (it()->channel_id == netchannel) return it();
	}
	return NULL;
}

CL_NetChannelQueue_Client *CL_NetGame_Client::create_queue(int netchannel)
{
	CL_MutexSection mutex_section(mutex);

	CL_Iterator<CL_NetChannelQueue_Client> it(
		(CL_List<CL_NetChannelQueue_Client> *) &netchannels);

	while (it.next() != NULL)
	{
		if (it()->channel_id == netchannel) return it();
	}
	
	CL_NetChannelQueue_Client *c = new CL_NetChannelQueue_Client;
	c->channel_id = netchannel;
	c->access = 0;

	netchannels.add(c);

	return c;
}

void CL_NetGame_Client::run_thread(void *_self)
{
	CL_NetGame_Client *self = (CL_NetGame_Client *) _self;
	cl_assert(self != NULL); // totally lame assert!
	while (true)
	{
		self->keep_alive();
		self->network->provider->wait_for_connection_data(self->mutex);
	}
}

#endif
