
#ifdef USE_NETWORK

#include "netgame_server.h"
#include "Layer1/System/API/cl_assert.h"
#include "Layer1/System/API/error.h"

CL_NetGame_Server::CL_NetGame_Server(
	CL_Network_Generic *network,
	const char *game_id,
	int port)
: CL_NetGame_Generic(network)
{
	mutex = CL_Mutex::create();

	this->game_id = game_id;
	this->port = port;
	
	id_counter = 1; // server is zero, that's why we start at 1.
	
	udp_connection = NULL; // still no support for UDP.
	
	network->provider->start_accept_on_port(port);

	thread = CL_Thread::create(run_thread);
	thread->start(this);
}

CL_NetGame_Server::~CL_NetGame_Server()
{
	mutex->enter();
	delete thread;
	mutex->leave();

	network->provider->stop_accept_on_port(port);
	
	// clean up connections to clients:
	{
		CL_Iterator<CL_NetComputer_Host> it(prejoin);
		while (it.next() != NULL) delete it();
	}
	{
		CL_Iterator<CL_NetComputer_Host> it(computers);
		while (it.next() != NULL) delete it();
	}
	{
		CL_Iterator<CL_NetChannelQueue> it(netchannels);
		while (it.next() != NULL) delete it();
	}

	while (leave_queue.is_empty() == false)
	{
		delete leave_queue.get();
	}

	delete mutex;
}

const CL_NetComputer *CL_NetGame_Server::get_server() const
{
	// Should we return NULL if server? Or should we have a instance
	// representing ourself?

	return NULL;
}

const CL_NetGroup *CL_NetGame_Server::get_all() const
{
	return &all;
}

bool CL_NetGame_Server::peek(int channel) const
{
	CL_MutexSection mutex_section(mutex);

	CL_NetChannelQueue *queue = find_queue(channel);
	if (queue != NULL) return !queue->is_empty();
	
	return false;
}

CL_NetMessage CL_NetGame_Server::receive(int channel, int timeout)
{
	CL_MutexSection mutex_section(mutex);

	static bool warning = true;
	if (warning)
	{
		cl_info(info_network, "timeout parameter ignored: not implemented yet!");
		warning = false;
	}

	CL_NetChannelQueue *queue = find_queue(channel);
	if (queue != NULL)
	{
		if (queue->is_empty()) throw CL_Error("No message to receive.");
		return queue->get();
	}
	
	CL_NetMessage msg;
	msg.size = 0;
	msg.data = NULL;
	return msg;
}

void CL_NetGame_Server::send(
	const int dest_channel,
	const CL_NetComputer *dest,
	const CL_NetMessage &message,
	bool reliable /*= true*/)
{
	CL_MutexSection mutex_section(mutex);

	static bool warning = true;
	if (warning && reliable == false)
	{
		cl_info(info_network, "cannot send data unreliable (udp): not implemented yet!");
		warning = false;
	}

	CL_NetComputer_Host *dest_host = (CL_NetComputer_Host *) dest;
	CL_NetChannelQueue *queue = find_queue(dest_channel);
	
	if (queue == NULL)
		return; // nobody has read or write access. Don't send message.

	if (has_read_access(queue, dest_host) == false)
		return; // dest_host is not allowed to read messages on channel.

	// permissions ok, send message:
	CL_OutputSource_MemoryGeneric output;
	output.write_int32(Packet_NetChannel_Message_ToClient);
	output.write_int32(dest_channel);
	output.write_int32(message.size);
	output.write(message.data, message.size);
	
	CL_ConnectionPacket packet;
	packet.size = output.size();
	packet.data = output.get_data();

	dest_host->connection->send(packet);
}

void CL_NetGame_Server::send(
	const int dest_channel,
	const CL_NetGroup *dest,
	const CL_NetMessage &message,
	bool reliable /*= true*/)
{
	CL_MutexSection mutex_section(mutex);

	CL_Iterator<CL_NetComputer> it_computers(((CL_NetGroup *) dest)->computers);
	while (it_computers.next() != NULL)
	{
		send(
			dest_channel,
			it_computers(),
			message,
			reliable);
	}
}

CL_NetComputer *CL_NetGame_Server::receive_computer_leave()
{
	CL_MutexSection mutex_section(mutex);

	if (leave_queue.is_empty()) return NULL;
	return leave_queue.get();
}

const CL_NetComputer *CL_NetGame_Server::receive_computer_join()
{
	CL_MutexSection mutex_section(mutex);

	if (join_queue.is_empty()) return NULL;
	return join_queue.get();
}

const CL_NetComputer *CL_NetGame_Server::receive_computer_rejoin()
{
	CL_MutexSection mutex_section(mutex);

	if (rejoin_queue.is_empty()) return NULL;
	return rejoin_queue.get();
}

bool CL_NetGame_Server::receive_game_closed()
{
	return false; // we are the server - only closed if deleted.
}

int CL_NetGame_Server::access_status(int /*channel*/) const
{
	// server has always full access.
	return ACCESS_CHANNEL_READ|ACCESS_CHANNEL_WRITE;
}

bool CL_NetGame_Server::is_writable(int /*channel*/) const
{
	return true;
}

bool CL_NetGame_Server::is_readable(int /*channel*/) const
{
	return true;
}

// Client side only:
int CL_NetGame_Server::receive_access_changed()
{
	return -1; // we are not a client!
}

// Server side only:
void CL_NetGame_Server::set_access(
	int channel,
	const CL_NetComputer *computer,
	int access_rights)
{
	CL_MutexSection mutex_section(mutex);

	CL_NetComputer_Host *computer_host = (CL_NetComputer_Host *) computer;
	CL_NetChannelQueue *queue = create_queue(channel);

	CL_Link *link_read = queue->access_read.find_link(computer_host);
	CL_Link *link_write = queue->access_write.find_link(computer_host);

	bool send_rights = false;
	if ((access_rights & ACCESS_CHANNEL_READ) == ACCESS_CHANNEL_READ)
	{
		if (link_read == NULL)
		{
			queue->access_read.add(computer_host);
			send_rights = true;
		}
	}
	else
	{
		if (link_read != NULL)
		{
			queue->access_read.del(link_read);
			send_rights = true;
		}
	}

	if ((access_rights & ACCESS_CHANNEL_WRITE) == ACCESS_CHANNEL_WRITE)
	{
		if (link_write == NULL)
		{
			queue->access_write.add(computer_host);
			send_rights = true;
		}
	}
	else
	{
		if (link_write != NULL)
		{
			queue->access_write.del(link_write);
			send_rights = true;
		}
	}

	if (send_rights)
	{
		CL_OutputSource_MemoryGeneric output;
		output.write_int32(Packet_NetChannel_AccessChange);
		output.write_int32(channel);
		output.write_int32(access_rights);
	
		CL_ConnectionPacket packet;
		packet.size = output.size();
		packet.data = output.get_data();

		computer_host->connection->send(packet);
	}
}

void CL_NetGame_Server::set_access(
	int channel,
	const CL_NetGroup *group,
	int access_rights)
{
	CL_MutexSection mutex_section(mutex);

	CL_Iterator<CL_NetComputer> it_computers(((CL_NetGroup *) group)->computers);
	while (it_computers.next() != NULL)
	{
		set_access(
			channel,
			it_computers(),
			access_rights);
	}
}

void CL_NetGame_Server::keep_alive()
{
	CL_MutexSection mutex_section(mutex);

	// check connections for data:
	CL_Iterator<CL_NetComputer_Host> it_computers(computers);
	while (it_computers.next() != NULL)
	{
		while (it_computers()->connection->peek())
		{
			CL_ConnectionPacket msg = it_computers()->connection->receive();
			
			CL_InputSource_MemoryGeneric input(msg.data, msg.size, true);
			switch (input.read_int32())
			{
			case Packet_NetChannel_AccessChange:
				cl_info(info_network, "Network Protocol error!");
				cl_assert(false);
				break;
			
			case Packet_NetChannel_Message_ToServer:
				{
					static bool warning = true;
					if (warning)
					{
						cl_info(info_network, "permissions not checked: Not implemented yet.");
						warning = false;
					}
					CL_NetChannelQueue *queue = create_queue(input.read_int32());
					CL_NetMessage game_msg;
					game_msg.size = input.read_int32();
					game_msg.data = new char[game_msg.size];
					game_msg.from = it_computers();
					input.read(game_msg.data, game_msg.size);
					queue->put(game_msg);
				}
				break;

			default:
				cl_info(info_network, "Network Protocol error!");
				cl_assert(false);
			}
		}

		if (it_computers()->connection->connection_lost())
		{
			leave_queue.put(it_computers());
			all.computers.del(it_computers());
			it_computers.remove();
		}
	}

	// check for "hello to you too" connections:
	CL_Iterator<CL_NetComputer_Host> it_prejoin(prejoin);
	while (it_prejoin.next() != NULL)
	{
		if (it_prejoin()->connection->connection_lost()) // how rude!
		{
			delete it_prejoin();
			it_prejoin.remove();
			
			continue;
		}

		if (it_prejoin()->connection->peek())
		{
			CL_ConnectionPacket msg = it_prejoin()->connection->receive();
			
			CL_InputSource_MemoryGeneric input(msg.data, msg.size);
			switch (input.read_int32())
			{
			case Packet_Hello_ToYouToo: // ahh - he wants to join.
				computers.add(it_prejoin());
				all.computers.add(it_prejoin());
				join_queue.put(it_prejoin());
				it_prejoin.remove();
				break;

			default:
				cl_info(info_network, "Network Protocol error!");
				cl_assert(false);
			}
			
			delete[] msg.data;
		}
	}
	
	// accept incoming clients:
	while (true)
	{
		CL_Connection *new_client = network->provider->accept();
		if (new_client == NULL) break;
		
		CL_OutputSource_MemoryGeneric msg;
		msg.write_int32(Packet_Hello);
		msg.write_int32(id_counter);
		msg.write_int32(game_id.get_length());
		msg.write(game_id.get_string(), game_id.get_length());
		
		CL_ConnectionPacket packet(msg.get_data(), msg.size());
		new_client->send(packet);
		
		prejoin.add(new CL_NetComputer_Host(new_client, id_counter++));
	}
}

void CL_NetGame_Server::run_thread(void *_self)
{
	CL_NetGame_Server *self = (CL_NetGame_Server *) _self;
	cl_assert(self != NULL); // totally lame assert!

	while (true)
	{
		self->keep_alive();
		self->network->provider->wait_for_connection_data(self->mutex);
	}
}

CL_NetChannelQueue *CL_NetGame_Server::find_queue(int netchannel) const
{
	CL_Iterator<CL_NetChannelQueue> it(
		(CL_List<CL_NetChannelQueue> *) &netchannels);
	while (it.next() != NULL)
	{
		if (it()->channel_id == netchannel) return it();
	}
	return NULL;
}

CL_NetChannelQueue *CL_NetGame_Server::create_queue(int netchannel)
{
	CL_Iterator<CL_NetChannelQueue> it(netchannels);

	while (it.next() != NULL)
	{
		if (it()->channel_id == netchannel) return it();
	}
	
	CL_NetChannelQueue *c = new CL_NetChannelQueue;
	c->channel_id = netchannel;

	netchannels.add(c);

	return c;
}

bool CL_NetGame_Server::has_read_access(
	CL_NetChannelQueue *channel,
	CL_NetComputer_Host *host)
{
	CL_Iterator<CL_NetComputer_Host> it(channel->access_read);
	while (it.next() != NULL)
	{
		if (it() == host) return true;
	}

	return false;	
}

// ---------

CL_NetComputer_Host::CL_NetComputer_Host(
	CL_Connection *connection,
	int id)
{
	this->connection = connection;
	this->id = id;
}

CL_NetComputer_Host::~CL_NetComputer_Host()
{
	delete connection;
}

#endif
