/*
	$Id: network_delivery_socket.cpp,v 1.8 1999/08/13 22:04:59 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#ifdef USE_NETWORK

#include "Layer1/System/API/cl_assert.h"
#include "Layer1/System/API/error.h"
#include "network_delivery_socket.h"
#include "network_delivery_unix.h"
#include <string.h>
#include <errno.h>

#define NETWORK_MAGIC 0x16042104

/******************************************************************************
                         Unix Socket implementation
******************************************************************************/

#define AVAIL_BUFFER_SIZE 16*1024

CL_UnixSocket::CL_UnixSocket(CL_Connections_Unix *_provider)
{
	provider = _provider;
	is_connection_lost = false;
	sock = -1;

	recv_state = expect_magic;
	cur_message = NULL;
	cur_message_size = -1;

	avail_buffers.add(new char[AVAIL_BUFFER_SIZE]);
	avail_buffer_pos = 0;
	avail_buffer_end = 0;
}

CL_UnixSocket::~CL_UnixSocket()
{
	if (sock != -1) close(sock);

	// <tm> code:
	CL_Iterator<char> counter(avail_buffers);
	while (counter.next() != NULL) delete[] counter();
	
	provider->remove_connection(this);
}

bool CL_UnixSocket::read_avail()
{
	fd_set rfds;

	FD_ZERO(&rfds);
	FD_SET(sock, &rfds);

	timeval timeout;
	memset(&timeout, 0, sizeof(timeval));

	int retval = select(sock+1, &rfds, NULL, NULL, &timeout);
	if (retval == -1 || retval == 0) 
	{
		return false;
	}

	int data_gotten = ::recv(
		sock,
		avail_buffers.get_last()+avail_buffer_end,
		AVAIL_BUFFER_SIZE-avail_buffer_end,
		0);

	if (data_gotten == 0 || data_gotten == -1)
	{
		is_connection_lost = true;
		return false;
	}

	avail_buffer_end += data_gotten;
	
	cl_assert(avail_buffer_end <= AVAIL_BUFFER_SIZE);

	if (avail_buffer_end == AVAIL_BUFFER_SIZE) // expand data avail buffer
	{
		avail_buffers.add(new char[AVAIL_BUFFER_SIZE]);
		avail_buffer_end = 0;
		read_avail(); // go get the rest!
	}

	return true;
}

bool CL_UnixSocket::require_avail(int size)
{
	if (avail_buffers.get_first() == avail_buffers.get_last())
	{
		if (avail_buffer_end-avail_buffer_pos >= size) 
		{
			return true;
		}
		return false;
	}
	
	CL_Iterator<char> counter(avail_buffers);

	int total_avail = 0;
	while (counter.next() != NULL)
	{
		int available = AVAIL_BUFFER_SIZE;
		if (counter() == avail_buffers.get_first()) available -= avail_buffer_pos;
		else if (counter() == avail_buffers.get_last()) available = avail_buffer_end;

		total_avail += available;
		if (total_avail >= size) return true;
	}

	return false;
}

void CL_UnixSocket::get_avail(void *buf, int size)
{
	char *buf2 = (char *) buf;

	while (true)
	{
		int available = AVAIL_BUFFER_SIZE - avail_buffer_pos;
		if (avail_buffers.get_first() == avail_buffers.get_last())
			available = avail_buffer_end-avail_buffer_pos;

		if (available > size) available = size;

		memcpy(buf2, avail_buffers.get_first()+avail_buffer_pos, available);

		avail_buffer_pos += available;
		size -= available;
		buf2 += available;

		if (avail_buffer_pos == AVAIL_BUFFER_SIZE)
		{
			delete[] avail_buffers.get_first();
			avail_buffers.del(avail_buffers.get_first());
			
//			cout << "*** NEXT AVAIL_BUFFER ***" << endl;

			cl_assert(avail_buffers.get_first() != NULL); // oh oh!
			avail_buffer_pos = 0;
		}
		else if (
			avail_buffers.get_first() == avail_buffers.get_last() &&
			avail_buffer_pos == avail_buffer_end)
		{
			avail_buffer_pos = 0;
			avail_buffer_end = 0;
		}

		cl_assert(size >= 0);
		if (size == 0) return;
	}
}

bool CL_UnixSocket::init_socket(int init_socket)
{
	if (init_socket != -1)
	{
		sock = init_socket;
		write_int(NETWORK_MAGIC);
		return true;
	}
	else
	{
		sock = socket(AF_INET, SOCK_STREAM, 0);
		return (sock != -1);
	}
}

bool CL_UnixSocket::peek()
{
	if (sock == -1) return false;

	read_avail();

	switch (recv_state)
	{
	case expect_magic:
		{
			if (!require_avail(4)) return false;
			int found_magic;
			get_avail(&found_magic, sizeof(int));
			found_magic = ntohl(found_magic);
			if (found_magic != NETWORK_MAGIC)
			{
				is_connection_lost=true;
			}
			else recv_state = expect_packet_size;
		}

	case expect_packet_size:
		{
			if (!require_avail(4)) return false;
			get_avail(&cur_message_size, sizeof(int));
			cur_message_size = ntohl(cur_message_size);
			recv_state = expect_packet_data;
		}

	case expect_packet_data:
		{
			if (!require_avail(cur_message_size)) return false;
			cur_message = new char[cur_message_size];
			get_avail(cur_message, cur_message_size);
			recv_state = packet_finished;
		}

	case packet_finished:
		return true;

	default:
		throw CL_Error("Network protocol error.");
		return false;
	};

	return false;
}

CL_ConnectionPacket CL_UnixSocket::receive()
{
	if (peek() == false)
	{
		CL_ConnectionPacket ret;
		ret.size = 0;
		ret.data = NULL;
		return ret;
	}

	CL_ConnectionPacket ret;
	ret.size = cur_message_size;
	ret.data = cur_message;

	recv_state = expect_packet_size;

	return ret;
}

void CL_UnixSocket::send(CL_ConnectionPacket message)
{
	write_int(message.size);
	write_data(message.data, message.size);
}

bool CL_UnixSocket::connection_lost()
{
	return is_connection_lost;
}

unsigned long CL_UnixSocket::read_int()
{
	unsigned char data[4];
	int read = 0;
	while (read < 4)
	{
		int val = recv(sock, (char *) data+read, 4-read, 0);
		if (val == 0 || val == -1) 
		{
			is_connection_lost = true;
			return 0;
		}

		read += val;
	}

	return ntohl(*((unsigned long *) data));
}

void CL_UnixSocket::read_data(void *dest, int size_data)
{
	int total_read = 0;
	while (total_read < size_data)
	{
		int bytes_read = recv( sock,
						((char *) dest)+total_read, 
						size_data-total_read, 
						0);
		if (bytes_read == 0 || bytes_read == -1)
		{
			is_connection_lost = true;
			return;
		}
			
		total_read += bytes_read;
	}
}

void CL_UnixSocket::write_int(unsigned long value)
{
	unsigned long write_num = htonl(value);

	write_data(&write_num, sizeof(unsigned long));
}

void CL_UnixSocket::write_data(void *data, unsigned int size)
{
	unsigned int bytes_send = 0;
	while (bytes_send < size)
	{
		int b = ::send(sock, ((char *) data)+bytes_send, size-bytes_send, 0);
		if (b == -1)
		{
			is_connection_lost = true;
			return;
		}

		bytes_send += b;
	}
}

bool CL_UnixSocket::try_connect(unsigned long remote_ip_network_format, int port)
{
	sockaddr_in sock_addr;
	sock_addr.sin_family = AF_INET;
	sock_addr.sin_addr.s_addr = remote_ip_network_format;
	sock_addr.sin_port = htons(port);
	
	int res = ::connect(sock, (sockaddr *) &sock_addr, sizeof(sockaddr_in));
	if (!res)
	{
//		cout << "Connected. Writing network magic." << endl;
		write_int(NETWORK_MAGIC);
		return true;
	}
	else 
	{
//		cout << "Connect failed." << endl;
		return false;
	}
}

/******************************************************************************
					CL_UnixAcceptSocket implementation
******************************************************************************/

CL_UnixAcceptSocket::CL_UnixAcceptSocket(CL_Connections_Unix *provider)
: CL_UnixSocket(provider)
{
	port = -1;
}

CL_UnixAcceptSocket::~CL_UnixAcceptSocket()
{
}

bool CL_UnixAcceptSocket::bind(int _port)
{
	if (!init_socket()) return false;

	sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(_port);

	// Bind the socket to the specified port
	int err = ::bind(sock, (sockaddr *) &addr, sizeof(sockaddr_in));	
	if (err == -1) return false;

	// Call getsockname to determine allocated port-number,
	// since socket might have been created with _port==0 (any)
	unsigned int len = sizeof(sockaddr_in);
	err = getsockname(sock, (sockaddr *) &addr, &len);
	if (err == -1) return false;

	port = ntohs(addr.sin_port);
	
	// Make the socket listen for incoming connection attempts
	err = listen(sock, 64);
	if (err == -1) return false;

	return true;
}

bool CL_UnixAcceptSocket::peek()
{
	fd_set rfds;

	FD_ZERO(&rfds);
	FD_SET(sock, &rfds);

	timeval timeout;
	memset(&timeout, 0, sizeof(timeval));

	int retval = select(sock+1, &rfds, NULL, NULL, &timeout);
	cl_assert(retval != -1);

	if (retval == 0) return false;
	else return true;
}

CL_UnixSocket *CL_UnixAcceptSocket::accept()
{
	if (peek())
	{
		int res_sock = ::accept(sock, NULL, NULL);
		cl_assert(res_sock != -1);

		CL_UnixSocket *ret = new CL_UnixSocket(provider);
		ret->init_socket(res_sock);
		
		return ret;
	}
	else
	{
		return NULL;
	}
}

bool CL_UnixAcceptSocket::try_connect(unsigned long remote_ip_network_format, int port)
{
	cl_assert(false);

	return false;
}

CL_ConnectionPacket CL_UnixAcceptSocket::receive()
{
	cl_assert(false);
	CL_ConnectionPacket ret;
	return ret;
}

void CL_UnixAcceptSocket::send(CL_ConnectionPacket message)
{
	cl_assert(false);
}

bool CL_UnixAcceptSocket::connection_lost()
{
	cl_assert(false);
	return false;
}

/******************************************************************************
					CL_UDPConnection_Unix implementation
******************************************************************************/

CL_UDPConnection_Unix::CL_UDPConnection_Unix()
{
	sock = -1;
	port = -1;
}

CL_UDPConnection_Unix::~CL_UDPConnection_Unix()
{
	if (sock != -1) close(sock);
}

bool CL_UDPConnection_Unix::bind(unsigned int _port)
{
	int res;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1) return false;

	sockaddr_in addr;
	memset(&addr, 0, sizeof(sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(_port);

	res = ::bind(sock, (sockaddr *) &addr, sizeof(addr));
	if (res == -1)
	{
		close(sock);
		return false;
	}

	int enable = 1;
	res = setsockopt(
		sock,
		SOL_SOCKET,
		SO_BROADCAST,
		(const char *) &enable,
		sizeof(int));

	if (res == -1)
	{
		cout << "ClanLib Network: Could not allow broadcast!" << endl;
		close(sock);
		return false;
	}

	// Call getsockname to determine allocated port-number,
	// since socket might have been created with _port==0 (any)
	unsigned int len = sizeof(sockaddr_in);
	res = getsockname(sock, (sockaddr *) &addr, &len);
	if (res == -1)
	{
		cout << "ClanLib Network: Could not determine allocated port number!" << endl;
		close(sock);
		return false;
	}

	port = ntohs(addr.sin_port);

	return true;
}

bool CL_UDPConnection_Unix::peek()
{
	if (sock == -1) return false;

	fd_set rfds;

	FD_ZERO(&rfds);
	FD_SET(sock, &rfds);

	timeval timeout;
	memset(&timeout, 0, sizeof(timeval));

	int retval = select(sock+1, &rfds, NULL, NULL, &timeout);
	if (retval == -1) return false;

	return (retval > 0);
}

CL_UDPConnectionPacket CL_UDPConnection_Unix::receive()
{
	char *receive_buf = new char[8096];

	sockaddr_in addr;
	memset(&addr, 0, sizeof(sockaddr_in));
	addr.sin_family = AF_INET;
	unsigned int len_addr = sizeof(addr);
	int res = ::recvfrom(sock, receive_buf, 8096, 0, (sockaddr *) &addr, &len_addr);
	cl_assert(res != -1);

	CL_UDPConnectionPacket ret;
	ret.data = receive_buf;
	ret.size = res;
	ret.ip_addr = addr.sin_addr.s_addr;
	ret.port = ntohs(addr.sin_port);

	return ret;
}

void CL_UDPConnection_Unix::send(CL_UDPConnectionPacket message)
{
	sockaddr_in addr;
	memset(&addr, 0, sizeof(sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_BROADCAST /*message.ip_addr*/;
	addr.sin_port = htons(message.port);

	cl_assert(sock != -1);
	cl_assert(message.size < 2000);

	int res = ::sendto(
		sock,
		(char *) message.data,
		message.size,
		0,
		(sockaddr *) &addr,
		sizeof(sockaddr_in));

	if (res == -1)
	{
		cout << "sendto failed: " << strerror(errno) << endl;
	}

	cl_assert(res != -1);
}

void CL_UDPConnection_Unix::broadcast(CL_UDPConnectionPacket message)
{
	cl_assert(sock != -1);
	cl_assert(message.size < 2000);

	sockaddr_in addr;
	memset(&addr, 0, sizeof(sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_BROADCAST;
	addr.sin_port = htons(message.port);

	int res = ::sendto(
		sock,
		(char *) message.data,
		message.size,
		0,
		(sockaddr *) &addr,
		sizeof(sockaddr_in));
		
	if (res == -1)
	{
		cout << "sendto failed: " << strerror(errno) << endl;
	}

	cl_assert(res != -1);
}

unsigned int CL_UDPConnection_Unix::get_port()
{
	return port;
}

/******************************************************************************
                      Unix PipeConnection implementation
******************************************************************************/

CL_UnixPipeConnection::CL_UnixPipeConnection(int _read, int _write)
{
	read = _read;
	write = _write;
	is_connection_lost = false;
}

CL_UnixPipeConnection::~CL_UnixPipeConnection()
{
	close(read);
	close(write);
}

bool CL_UnixPipeConnection::peek()
{
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(read, &rfds);

	timeval timeout;
	memset(&timeout, 0, sizeof(timeval));

	int retval = select(read+1, &rfds, NULL, NULL, &timeout);
	return retval > 0;
}

CL_ConnectionPacket CL_UnixPipeConnection::receive()
{
	unsigned long size_data = read_int();

	unsigned char *data = new unsigned char[size_data];
	read_data(data, size_data);

	CL_ConnectionPacket ret;
	ret.size = size_data;
	ret.data = data;

	return ret;
}

void CL_UnixPipeConnection::send(CL_ConnectionPacket message)
{
	write_int(message.size);
	write_data(message.data, message.size);
}

bool CL_UnixPipeConnection::connection_lost()
{
	return is_connection_lost;
}

unsigned long CL_UnixPipeConnection::read_int()
{
	unsigned long ret;
	read_data(&ret, sizeof(unsigned long));

	return ret;
}

void CL_UnixPipeConnection::read_data(void *dest, int size_data)
{
	int total_read = 0;
	while (total_read < size_data)
	{
		int res = ::read(
			read, 
			((unsigned char *) dest)+total_read,
			size_data-total_read);

		if (res == 0)
		{
			cout << "OH MY GOD - TIME TO QUIT" << endl;
			is_connection_lost = true;
			return;
		}
			
		total_read += res;
	}
}

void CL_UnixPipeConnection::write_int(unsigned long value)
{
	write_data(&value, sizeof(unsigned long));
}

inline void CL_UnixPipeConnection::write_data(void *data, unsigned int size)
{
	::write(write, data, size);
}


#endif // USE_NETWORK


