/*
	ClanSound, sound mixing library written for ClanLib.
	Copyright (c)1998 Magnus Norddahl / ClanSoft.
*/

#ifdef USE_CLANSOUND

#include <unistd.h>
#include <fcntl.h>
#include <iostream.h>
#include <signal.h>

#include "clansound.h"
#include "soundserver.h"

CL_Array<int> c_samples;
CL_Array<int> c_sessions;

CS_PipeConnection *clientside_pipe = NULL;
int cs_server_read = -1;
int cs_server_write = -1;

void cs_signal_pipe(int)
{
	close(cs_server_read);
	close(cs_server_write);
	exit(0); // bye bye
}

int cs_init()
{
	if (cs_server_init() == false) // no sound available
	{
		cs_server_read = -1;
		cs_server_write = -1;
		return 0;
	}

	// start server:
	int pipes[4];

	pipe(pipes);
	pipe(pipes+2);

	cs_server_read = pipes[0];
	cs_server_write = pipes[3];
	
	// don't block writes to sound server - helps the performance...
//	fcntl(cs_server_write, F_SETFL, O_NONBLOCK);

	if (fork()==0) // start server
	{
		signal(SIGPIPE, cs_signal_pipe);
		signal(SIGHUP, cs_signal_pipe);
		signal(SIGCHLD, cs_signal_pipe);
		signal(SIGTERM, cs_signal_pipe);

//		cout << "ClanSound server is pid: " << getpid() << endl;
		
		cs_server_main(pipes[2], pipes[1]);
		close(pipes[2]);
		close(pipes[1]);
		exit(0);
	}

	clientside_pipe = new CS_PipeConnection(cs_server_read, cs_server_write);

//	cout << "ClanSound client is pid: " << getpid() << endl;

	return 0;
}


void cs_deinit()
{
	if (cs_server_write == -1) return;
	if (cs_server_read == -1) return;

	CL_DataWriter command(sizeof(int));
	command.write_int32(GOODBYE);
	clientside_pipe->send(CS_ConnectionPacket(&command));

	close(cs_server_write);
	if (cs_server_read != cs_server_write) close(cs_server_read);
	
	delete clientside_pipe;
	clientside_pipe = NULL;
}

int cs_create_sample(int format, void *data, int data_size)
{
	if (clientside_pipe == NULL) return -1;

	CL_DataWriter command(sizeof(int)*3+data_size);
	command.write_int32(CREATE_SAMPLE);
	command.write_int32(format);
	command.write_int32(data_size);
	command.write_data(data, data_size);
	clientside_pipe->send(CS_ConnectionPacket(&command));
/*
	CS_ConnectionPacket p = clientside_pipe->receive();
	CL_DataReader packet((unsigned char *) p.data, p.size, true);
	int id = packet.read_int32();
*/

	int id = c_samples.add(new int);

	return id;
}

void cs_release_sample(int sample_id)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(RELEASE_SAMPLE);
	command.write_int32(sample_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));

	delete c_samples[sample_id]; c_samples.del(sample_id);
}

int cs_create_session(int sample_id)
{
	if (clientside_pipe == NULL) return -1;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(CREATE_SESSION);
	command.write_int32(sample_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));
/*
	CS_ConnectionPacket p = clientside_pipe->receive();
	CL_DataReader packet((unsigned char *) p.data, p.size, true);
	int id = packet.read_int32();
*/

	int id = c_sessions.add(new int);

	return id;
}

int cs_create_session_streamed(int format)
{
	if (clientside_pipe == NULL) return -1;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(CREATE_STREAM);
	command.write_int32(format);
	clientside_pipe->send(CS_ConnectionPacket(&command));
/*
	CS_ConnectionPacket p = clientside_pipe->receive();
	CL_DataReader packet((unsigned char *) p.data, p.size, true);
	int id = packet.read_int32();
*/

	int id = c_sessions.add(new int);

	return id;
}

void cs_play(int session_id)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(PLAY_SESSION);
	command.write_int32(session_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

void cs_stop(int session_id)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(STOP_SESSION);
	command.write_int32(session_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

void cs_flush(int session_id)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(FLUSH_SESSION);
	command.write_int32(session_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

void cs_set_volume(int session_id, int volume)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*3);
	command.write_int32(SET_VOLUME);
	command.write_int32(session_id);
	command.write_int32(volume);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

void cs_set_pan(int session_id, int pan)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*3);
	command.write_int32(SET_PAN);
	command.write_int32(session_id);
	command.write_int32(pan);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

void cs_set_position(int session_id, int position)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*3);
	command.write_int32(SET_POSITION);
	command.write_int32(session_id);
	command.write_int32(position);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

void cs_set_frequency(int session_id, int frequency)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*3);
	command.write_int32(SET_FREQUENCY);
	command.write_int32(session_id);
	command.write_int32(frequency);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

bool cs_is_playing(int session_id)
{
	if (clientside_pipe == NULL) return false;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(IS_PLAYING);
	command.write_int32(session_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));

	CS_ConnectionPacket p = clientside_pipe->receive();
	CL_DataReader packet((unsigned char *) p.data, p.size, true);
	int playback_state = packet.read_int32();

	if (playback_state) return true;
	return false;
}

int cs_release_session(int session_id)
{
	if (clientside_pipe == NULL) return 0;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(RELEASE_SESSION);
	command.write_int32(session_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));

	delete c_sessions[session_id]; c_sessions.del(session_id);
	
	return 0;
}

int cs_update_stream(int session_id, void *data, int data_size)
{
	if (clientside_pipe == NULL) return 0;

	CL_DataWriter command(sizeof(int)*3+data_size);
	command.write_int32(UPDATE_STREAM);
	command.write_int32(session_id);
	command.write_int32(data_size);
	command.write_data(data, data_size);
	clientside_pipe->send(CS_ConnectionPacket(&command));

	return 0;
}

int cs_ask_update_stream(int session_id)
{
	if (clientside_pipe == NULL) return 0;
	
	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(ASK_UPDATE_STREAM);
	command.write_int32(session_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));

	CS_ConnectionPacket p = clientside_pipe->receive();
	CL_DataReader packet((unsigned char *) p.data, p.size, true);
	int bytes_requested = packet.read_int32();

	return bytes_requested;
}

void cs_eof_stream(int session_id)
{
	if (clientside_pipe == NULL) return;

	CL_DataWriter command(sizeof(int)*2);
	command.write_int32(EOF_STREAM);
	command.write_int32(session_id);
	clientside_pipe->send(CS_ConnectionPacket(&command));
}

#endif
