/*
	ClanSound, sound mixing library written for ClanLib.
	Copyright (c)1998 Magnus Norddahl / ClanSoft.

	OSS Documentation can be found at http://www.opensound.com
*/

#ifdef USE_CLANSOUND

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <iostream.h>

#include "oss.h"

// If USE_DRIVER_FRAGSIZE is defined, we use the fragsize (and number of them)
// used by the soundcard driver. This is not as efficient as using our own
// settings - mainly provided for debugging.
// #define USE_DRIVER_FRAGSIZE

int dev_dsp_fd;
int frag_size;

int cs_oss_init()
//: Initialize sound playback.
{
	dev_dsp_fd = open("/dev/dsp", O_WRONLY);
	if (dev_dsp_fd == -1)
	{
		cout << "ClanSound: Could not open /dev/dsp. No sound will be available." << endl;
		return 0;
	}
	
#ifndef USE_DRIVER_FRAGSIZE
	int frag_settings = 0x0003000b; // 0xMMMMSSSS
		// (where MMMM = num fragments, SSSS = fragment size)

	if (ioctl(dev_dsp_fd, SNDCTL_DSP_SETFRAGMENT, &frag_settings))
	{
		cout << "ClanSound: Failed to set soundcard fragment size. Sound may have a long latency." << endl;
	}

#endif
	
	int format = AFMT_S16_LE;
	ioctl(dev_dsp_fd, SNDCTL_DSP_SETFMT, &format);
	if (format != AFMT_S16_LE)
	{
		cout << "ClanSound: Requires 16 bit soundcard. No sound will be available." << endl;
		close(dev_dsp_fd);
		return 0;
	}
	
	int stereo = 1;
	ioctl(dev_dsp_fd, SNDCTL_DSP_STEREO, &stereo);
	if (stereo != 1)
	{
		cout << "ClanSound: Requires 16 bit stereo capable soundcard. No sound will be available." << endl;
		close(dev_dsp_fd);
		return 0;
	}
	
	int speed = 22050;
	ioctl(dev_dsp_fd, SNDCTL_DSP_SPEED, &speed);

	float percent_wrong = speed / (float) 22050;
	if (percent_wrong < 0.90 || percent_wrong > 1.10)
	{
		cout << "ClanSound: Mixing rate (22.05 kHz) not supported by soundcard." << endl;

		close(dev_dsp_fd);
		return 0;
	}
	
	// Try to improve mixing performance by using the same mixing buffer size
	// as the sound device does:
	int err = ioctl(dev_dsp_fd, SNDCTL_DSP_GETBLKSIZE, &frag_size);
	if (err == -1)
	{
		cout << "ClanSound: Warning, Couldn't get sound device blocksize. Using 0.25 sec mixing buffer." << endl;
		frag_size = 22050/2; // 0.25 sec mixing buffer used.
	}

//	cout << "ClanSound: Fragment size is " << frag_size << endl;

	audio_buf_info info;
	ioctl(dev_dsp_fd, SNDCTL_DSP_GETOSPACE, &info);

//	cout << "ClanSound: Num fragments available is " << info.fragments << endl;
	
	return 1;
}

void cs_oss_deinit()
//: Deinitializes sound playback.
{
	if (dev_dsp_fd != -1) close(dev_dsp_fd);
}

void cs_oss_silence()
//: Called when we have no samples to play - and wants to tell the soundcard
//: about this possible event.
{
	ioctl(dev_dsp_fd, SNDCTL_DSP_POST, 0);
}

bool cs_oss_is_full()
//: Returns true if all fragments are filled with data.
{
	audio_buf_info info;
	int err = ioctl(dev_dsp_fd, SNDCTL_DSP_GETOSPACE, &info);
	if (err == -1)
	{
		cout << "ClanSound: fragments free not supported by device!?" << endl;
		return false; // not supported by device!?
	}

	if (info.fragments == 0) return true;
	return false;
}

int cs_oss_get_frag_size()
//: Returns the buffer size used by device (returned as num [stereo] samples).
{
	return frag_size/4;
}

int cs_oss_get_fd()
//: Returns file descriptor used by device.
{
	return dev_dsp_fd;
}

void cs_oss_write_fragment(short *data)
//: Writes a fragment to the soundcard.
{
	write(dev_dsp_fd, data, frag_size);
}

#endif
