/*
	ClanSound, sound mixing library written for ClanLib.
	Copyright (c)1998 Magnus Norddahl / ClanSoft.
*/

#ifdef USE_CLANSOUND

#include <iostream.h>
#include <unistd.h>
#include <string.h>

#include "clansound.h"
#include "soundserver.h"
#include "session.h"

CS_Session_Sample::CS_Session_Sample(CS_Sample *_sample)
{
	sample = _sample;
	pos = 0;
	freq = 0;

	set_volume(128);
}

CS_Session_Sample::~CS_Session_Sample()
{
}

void CS_Session_Sample::play()
{
	playing = true;
}

void CS_Session_Sample::stop()
{
	playing = false;
}

void CS_Session_Sample::set_volume(int _vol)
{
	vol = _vol;
}

void CS_Session_Sample::set_pan(int _pan)
{
	pan = _pan;
}

void CS_Session_Sample::set_position(int _pos)
{
	pos = _pos;
}

void CS_Session_Sample::set_frequency(int _freq)
{
	freq = _freq;
}

bool CS_Session_Sample::is_playing()
{
	return playing;
}

void CS_Session_Sample::get_playback_data(int *data, int num_samples)
{
	if (freq == 0 || vol==0)
	{
		memset(data, 0, num_samples*sizeof(int)*2);
		return;
	}

	double speed = freq/(double) 22050;
	double fill_samples = (sample->num_samples-pos) / speed;

	if (fill_samples >= num_samples) fill_samples = num_samples;
	else // reached end of buffer - pad the rest with silence
	{
		memset(
			data,
			0,
			sizeof(int)*num_samples*2);
		
		playing = false;
	}

	int *ptr = data;
	for (int i=0; i<fill_samples; i++)
	{
		int cur_pos = (int) pos;

		// debug: (remove this one if it hasn't occoured in a long
		// time.
		// cs_assert(cur_pos < sample->num_samples);

		*(ptr++) = sample->data[cur_pos*2+0];
		*(ptr++) = sample->data[cur_pos*2+1];
		
		pos += speed;
	}

	// apply volume and panning:
	ptr = data;
	int left_pan = 128-pan;
	int right_pan = 128+pan;
	if (left_pan < 0) left_pan = 0;
	if (left_pan > 128) left_pan = 128;
	if (right_pan < 0) right_pan = 0;
	if (right_pan > 128) right_pan = 128;

	for (int i=0; i<fill_samples; i++)
	{
		// left channel:
		(*ptr) = (*ptr)*vol*left_pan/16384; ptr++;
		
		// right channel:
		(*ptr) = (*ptr)*vol*right_pan/16384; ptr++;
	}

	if (!playing) // stopped playing because buffer ended
	{
		pos = 0; // is this the behaviour of a directsound buffer? -- dok
	}
}

void CS_Session_Sample::mix_to(int *data, int num_samples)
{
	static int temp[16*1024*2];

	for (int i=0; i<num_samples; i+=16*1024)
	{
		int size = num_samples-i;
		if (size > 16*1024) size = 16*1024;
		
		get_playback_data(temp, size);
		for (int j=0; j<size*2; j++) data[i+j] += temp[j];
	}
}

/***********************************
         CS_Session_Stream
************************************/

CS_Session_Stream::CS_Session_Stream(int _format)
{
	pos = 0;
	freq = 0;
	set_volume(128);

	format = _format;
	pos = 0;
	playing = false;
	filled_pos = 0;
	first_time = true;
	
	buffer_size = 1024*256;
	ring_buffer = new short[buffer_size*2];
	stream_eof = false;
}

CS_Session_Stream::~CS_Session_Stream()
{
	delete[] ring_buffer;
}

void CS_Session_Stream::play()
{
	playing = true;
}

void CS_Session_Stream::stop()
{
	playing = false;
}

void CS_Session_Stream::flush()
{
	pos = 0;
	filled_pos = 0;
}

void CS_Session_Stream::set_volume(int _vol)
{
	vol = _vol;
}

void CS_Session_Stream::set_pan(int _pan)
{
	pan = _pan;
}

void CS_Session_Stream::set_frequency(int _freq)
{
	freq = _freq;
}

bool CS_Session_Stream::is_playing()
{
	return playing;
}

void CS_Session_Stream::eof()
{
	stream_eof = true;
}

void CS_Session_Stream::get_playback_data(int *data, int num_samples)
{
	if (freq == 0 || vol == 0)
	{
		memset(data, 0, num_samples*sizeof(int)*2);
		return;
	}

	double speed = freq/(double) 22050;

/*
	int samples_available = (int) (get_samples_available()*speed);
	if (num_samples > samples_available) // clear the rest of the buffer
	{
		memset(&data[samples_available], 0, num_samples-samples_available);
		num_samples = samples_available;
		
		if (stream_eof) playing = false; // reached end of stream.
	}
*/

	int *ptr = data;
	for (int i=0; i<num_samples; i++)
	{
		int cur_pos = (int) pos;

		if (cur_pos < buffer_size)
		{
			*(ptr++) = ring_buffer[cur_pos*2+0];
			*(ptr++) = ring_buffer[cur_pos*2+1];
		}

		pos += speed;
		if (pos >= buffer_size) pos -= buffer_size;
	}

//	if (num_samples >= samples_available) pos = filled_pos; // reached end of buffer

	// apply volume and panning:
	ptr = data;
	int left_pan = 128-pan;
	int right_pan = 128+pan;
	if (left_pan < 0) left_pan = 0;
	if (left_pan > 128) left_pan = 128;
	if (right_pan < 0) right_pan = 0;
	if (right_pan > 128) right_pan = 128;

	for (int i=0; i<num_samples; i++)
	{
		// left channel:
		(*ptr) = (*ptr)*vol*left_pan/16384; ptr++;
		
		// right channel:
		(*ptr) = (*ptr)*vol*right_pan/16384; ptr++;
	}

}

void CS_Session_Stream::mix_to(int *data, int num_samples)
{
	static int temp[16*1024*2];

	for (int i=0; i<num_samples; i+=16*1024)
	{
		int size = num_samples-i;
		if (size > 16*1024) size = 16*1024;
		
		get_playback_data(temp, size);
		for (int j=0; j<size*2; j++) data[i+j] += temp[j];
	}
}

int CS_Session_Stream::get_samples_available()
{
	if (pos < filled_pos) return (int) (pos+(buffer_size-filled_pos));
	if (pos > filled_pos) return (int) (filled_pos - pos);

	return 0;
}

int CS_Session_Stream::ask_update()
{
	int space_left = buffer_size;
	
	if (pos < filled_pos) space_left = (int) (buffer_size-filled_pos+pos);
	if (pos > filled_pos) space_left = (int) (pos-filled_pos);

	space_left -= 16;
	if (space_left <= 0) return 0;

	switch (format)
	{
	case CS_8BIT_MONO:
		break;
	
	case CS_8BIT_STEREO:
	case CS_16BIT_MONO:
		space_left *= 2;
		break;
		
	case CS_16BIT_STEREO:
		space_left *= 4;
		break;
		
	default:
		cs_assert(false);
	}
	
	return space_left;
}

int CS_Session_Stream::update(void *data, int data_size)
{
	if (first_time)
	{
		first_time = false;
		filled_pos = (int) pos;
	}

	int num_samples = data_size;
	int sample_size=0;

	switch (format)
	{
	case CS_8BIT_MONO:
		sample_size = 1;
		break;
	
	case CS_8BIT_STEREO:
	case CS_16BIT_MONO:
		sample_size = 2;
		break;
		
	case CS_16BIT_STEREO:
		sample_size = 4;
		break;
		
	default:
		cs_assert(false);
	}
	
	num_samples /= sample_size;

	if (filled_pos+num_samples > buffer_size) // exceeds end of ring buffer
	{
		int fill_length = buffer_size - filled_pos;
		char *_data = (char *) data;
	
		update_format(filled_pos, fill_length, data);
		update_format(0, num_samples - fill_length, _data+fill_length*sample_size);
		
		filled_pos = num_samples - fill_length;
	}
	else
	{
		update_format(filled_pos, num_samples, data);
		filled_pos += num_samples;
	}

	return 0;
}

void CS_Session_Stream::update_format(int pos, int num_samples, void *data)
{
	switch (format)
	{
	case CS_8BIT_MONO:
		{
			unsigned char *input_c = (unsigned char *) data;
			for (int i=0; i<num_samples; i++)
			{
				ring_buffer[(pos+i)*2+0] = input_c[i];
				ring_buffer[(pos+i)*2+1] = input_c[i];
			}
		}
		break;
	
	case CS_8BIT_STEREO:
		{
			unsigned char *input_c = (unsigned char *) data;
			for (int i=0; i<num_samples; i++)
			{
				ring_buffer[(pos+i)*2+0] = input_c[i*2+0];
				ring_buffer[(pos+i)*2+1] = input_c[i*2+1];
			}
		}
		break;

	case CS_16BIT_MONO:
		{
			short *input_s = (short *) data;
			for (int i=0; i<num_samples; i++)
			{
				ring_buffer[(pos+i)*2+0] = input_s[i];
				ring_buffer[(pos+i)*2+1] = input_s[i];
			}
		}
		break;
		
	case CS_16BIT_STEREO:
		{
			short *input_s = (short *) data;
			for (int i=0; i<num_samples; i++)
			{
				ring_buffer[(pos+i)*2+0] = input_s[i*2+0];
				ring_buffer[(pos+i)*2+1] = input_s[i*2+1];
			}
		}
		break;
		
	default:
		cs_assert(false);
	}
}

#endif
