/*
	ClanSound, sound mixing library written for ClanLib.
	Copyright (c)1998 Magnus Norddahl / ClanSoft.
*/

#ifdef USE_CLANSOUND
#define file_soundserver

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <iostream.h>
#include <errno.h>
#include <string.h>

#include "soundserver.h"
#include "sample.h"
#include "session.h"
#include "oss.h"
#include "mixer.h"

#include "array.h"

#include "pipeconnection.h"

bool cs_server_init()
{
	if (cs_oss_init()==0)
	{
		cout << "ClanSound: Failed to access /dev/dsp. No sound will be available." << endl;
		return false;
	}
	return true;
}

void cs_server_main(int client_read, int client_write)
{
	cs_mixer_init();

	int oss_fd = cs_oss_get_fd();

	int max_fd = oss_fd;
	if (client_read > max_fd) max_fd = client_read;

	CS_PipeConnection pipe(client_read, client_write);

	while (true)
	{
		fd_set rfds;
		fd_set wfds;
		
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		
		FD_SET(client_read, &rfds);
		FD_SET(oss_fd, &wfds);
		
		int retval = select(max_fd+1, &rfds, &wfds, NULL, NULL);
		if (retval)
		{
			if (FD_ISSET(client_read, &rfds)) // we have data from the client
			{
				while (pipe.peek())
				{
					if (cs_server_read_command(&pipe))
					{
						cs_mixer_deinit();
						cs_oss_deinit();
						
						return;
					}
				}
			}

			while (cs_oss_is_full() == false)

//			if (FD_ISSET(oss_fd, &wfds)) // time for some serious mixing!
				cs_mixer_mix();
		}
	}
}

int cs_server_read_command(CS_PipeConnection *pipe)
{
	CS_ConnectionPacket p = pipe->receive();
	CL_DataReader packet((unsigned char *) p.data, p.size, true);

	int command = packet.read_int32();

	switch (command)
	{
	case GOODBYE:
		return 1;

	case CREATE_SAMPLE:
		{
			int format = packet.read_int32();
			int data_size = packet.read_int32();

			char *data = new char[data_size];
			packet.read_data(data, data_size);
			
			/*int sample_id =*/ cs_server_create_sample(format, data, data_size);
/*
	// not needed anymore - we calc this on the client side.
	// (hopefully we won't be wrong) -- mbn

			CL_DataWriter reply(sizeof(int));
			reply.write_int32(sample_id);
			pipe->send(CS_ConnectionPacket(&reply));
*/
			delete[] data;
		}
		break;
		
	case RELEASE_SAMPLE:
		{
			int sample_id = packet.read_int32();
			cs_server_release_sample(sample_id);
		}
		break;
		
	case CREATE_SESSION:
		{
			int sample_id = packet.read_int32();
			/*int session_id =*/ cs_server_create_session(sample_id);
/*
	// not needed anymore - we calc this on the client side.
	// (hopefully we won't be wrong) -- mbn

			CL_DataWriter reply(sizeof(int));
			reply.write_int32(session_id);
			pipe->send(CS_ConnectionPacket(&reply));
*/
		}
		break;
		
	case CREATE_STREAM:
		{
			int format = packet.read_int32();
			/* int session_id =*/ cs_server_create_session_streamed(format);
/*
	// not needed anymore - we calc this on the client side.
	// (hopefully we won't be wrong) -- mbn

			CL_DataWriter reply(sizeof(int));
			reply.write_int32(session_id);
			pipe->send(CS_ConnectionPacket(&reply));
*/
		}
		break;
		
	case PLAY_SESSION:
		{
			int session_id = packet.read_int32();
			cs_server_play(session_id);
		}
		break;
		
	case STOP_SESSION:
		{
			int session_id = packet.read_int32();
			cs_server_stop(session_id);
		}
		break;
		
	case FLUSH_SESSION:
		{
			int session_id = packet.read_int32();
			cs_server_flush(session_id);
		}
		break;
		
	case SET_VOLUME:
		{
			int session_id = packet.read_int32();
			int volume = packet.read_int32();
			cs_server_set_volume(session_id, volume);
		}
		break;
		
	case SET_PAN:
		{
			int session_id = packet.read_int32();
			int pan = packet.read_int32();
			cs_server_set_pan(session_id, pan);
		}
		break;
		
	case SET_POSITION:
		{
			int session_id = packet.read_int32();
			int position = packet.read_int32();
			cs_server_set_position(session_id, position);
		}
		break;

	case SET_FREQUENCY:
		{
			int session_id = packet.read_int32();
			int frequency = packet.read_int32();
			cs_server_set_frequency(session_id, frequency);
		}
		break;
		
	case IS_PLAYING:
		{
			int session_id = packet.read_int32();
			int playback_state = 0;
			if (cs_server_is_playing(session_id)) playback_state = 1;
			
			CL_DataWriter reply(sizeof(int));
			reply.write_int32(playback_state);
			pipe->send(CS_ConnectionPacket(&reply));
		}
		break;
		
	case RELEASE_SESSION:
		{
			int session_id = packet.read_int32();
			cs_server_release_session(session_id);
		}
		break;

	case ASK_UPDATE_STREAM:
		{
			int session_id = packet.read_int32();
			int bytes_needed = cs_server_ask_update_stream(session_id);
			
			CL_DataWriter reply(sizeof(int));
			reply.write_int32(bytes_needed);
			pipe->send(CS_ConnectionPacket(&reply));
		}
		break;
		
	case UPDATE_STREAM:
		{
			int session_id = packet.read_int32();
			int data_size = packet.read_int32();
			
			char *data = new char[data_size];
			packet.read_data(data, data_size);

			cs_server_update_stream(session_id, data, data_size);

			delete[] data;
		}
		break;

	case EOF_STREAM:
		{
			int session_id = packet.read_int32();
			cs_server_eof_stream(session_id);
		}
		break;

	default:
		cout << "ClanSound server communication error!!! (unknown command received)" << endl;
		return 1;
	}
	
	return 0;
}

// -----------------------------------------------------------------------

int cs_server_create_sample(
	int format,
	void *data,
	int data_size)
{
	return samples.add(
		new CS_Sample(format, data, data_size));
}

void cs_server_release_sample(
	int sample_id)
{
	delete samples[sample_id];
	samples.del(sample_id);
}

int cs_server_create_session(
	int sample_id)
{
	return sessions.add(
		new CS_Session_Sample(samples[sample_id]));
}

int cs_server_create_session_streamed(
	int format)
{
	return sessions.add(new CS_Session_Stream(format));
}

void cs_server_play(
	int session_id)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->play();
}

void cs_server_stop(
	int session_id)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->stop();
}

void cs_server_flush(
	int session_id)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->flush();
}

bool cs_server_is_playing(
	int session_id)
{
	if (sessions[session_id] == NULL) return false;

	return sessions[session_id]->is_playing();
}

void cs_server_set_volume(
	int session_id,
	int volume)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->set_volume(volume);
}

void cs_server_set_pan(
	int session_id,
	int pan)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->set_pan(pan);
}

void cs_server_set_position(
	int session_id,
	int position)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->set_position(position);
}

void cs_server_set_frequency(
	int session_id,
	int frequency)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->set_frequency(frequency);
}

int cs_server_release_session(
	int session_id)
{
	delete sessions[session_id];
	sessions.del(session_id);

	return 0;
}

int cs_server_update_stream(
	int session_id,
	void *data,
	int data_size)
{
	if (sessions[session_id] == NULL) return 0;

	return sessions[session_id]->update(data, data_size);
}

int cs_server_ask_update_stream(
	int session_id)
{
	if (sessions[session_id] == NULL) return 0;

	return sessions[session_id]->ask_update();
}

void cs_server_eof_stream(
	int session_id)
{
	if (sessions[session_id] == NULL) return;

	sessions[session_id]->eof();
}

#endif
