/*
	$Id: timer_manager.h,v 1.2 1999/03/20 00:11:21 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		ClanTimer API - main header file.

*/

//! l1_component="System"

#ifndef header_timermanager
#define header_timermanager

class CL_TimerCallback
//: Abstract class to inherit in case of a timermanager-subscription
//: the manager makes calls to the callback()-function
{
	bool subscribed;
public:
	CL_TimerCallback();
	virtual ~CL_TimerCallback();

	virtual void subscribe(int milliseconds, float tolerance, bool periodic, bool delete_pointer = false);
//: CL_TimerManager wrapper function, which does nothing else, than call the CL_TimerManager
//: subscribe with 'this' as callback instance argument.
//!param: milliseconds - Specifies the callback granularity, eg. with what intervals the callback function should be called.
//!param: tolerance - Denotes the callback tolerance relative to the callback granularity, eg. how important it is, that the callback be performed on time. 0 denotes zero tolerance, whereas 1 for instance means, that it's ok that the callback varies by as much as 1*'milliseconds' milliseconds.
//!param: periodic - Boolean indicating whether this is a "one-shot" callback, which should be called once only, or it is a periodic callback, which should be called until a call to unsubscribe.
//!param: delete_pointer - Indicates if the instanced callback class should be deleted upon program exit, or a call to unsubscribe.

	virtual void unsubscribe();
//: CL_TimerManager wrapper function, which does nothing else, whan call the CL_TimerManager unsubscribe function with 'this' as argument.

	virtual int callback(int milliseconds_passed)=0;
//: Virtual function, which must be implemented. This function is called by the 
//: CL_TimerManager, in case the instance is subscribed to callbacks.
//!param: milliseconds_passed - the number of milliseconds passed, since the last call to the callback function
//!retval: If return value is != 0, this value denotes the new interval, on which this function should be called.
};

class CL_TimerManager
//: The CL_TimerManager class is ClanLib's interface to callback functions. Games
//: using MOD players or similar often need to use the TimerManager.
{
public:
	virtual ~CL_TimerManager() { ; }

	static bool subscribe(int milliseconds, CL_TimerCallback *callback_class, float tolerance, bool periodic, bool delete_pointer = true);
//: Static function used to subscribe a new timer-class to the timer.
//: <br>
//: Asks the timermanager to call 'callback_class->callback()' every 'milliseconds' 
//: millisecond, with a time-tolerance of 'tolerance' measured relative to the 
//: interval. Milliseconds=20 and tolerance=0.5 eg. means that callback_class 
//: has to be called every 10-30-th millisecond.
//: <br>
//: 'periodic' indicates whether the callback_class should be called only once, 
//: or every 'millisecond' millisecond, until unsubscribe is called. 
//: <br>
//: Pr. default the timermanager deletes the pointer to the callback_class when 
//: the manager is destroyed at program-termination-time, but if 'delete_pointer'==false 
//: this won't happen - likewise, the callback_class gets deleted when unsubscribing.
//!param: milliseconds - Specifies the callback granularity, eg. with what intervals the callback function should be called.
//!param: callback_class - Must point to an instance of the CL_TimerCallback class, which is to be called with the given interval.
//!param: tolerance - Denotes the callback tolerance relative to the callback granularity, eg. how important it is, that the callback be performed on time. 0 denotes zero tolerance, whereas 1 for instance means, that it's ok that the callback varies by as much as 1*'milliseconds' milliseconds.
//!param: periodic - Boolean indicating whether this is a "one-shot" callback, which should be called once only, or it is a periodic callback, which should be called until a call to unsubscribe.
//!param: delete_pointer - Indicates if the instanced callback class should be deleted upon program exit, or a call to unsubscribe.
//!retval: Returns whether the subscribtion attempt succeeded.

	static void unsubscribe(CL_TimerCallback *callback_class);
//: Unsubscribes the given callback_class, meaning no more calls will be made to it.
//!param: callback_class - A pointer to the instanced callback class previously given as an argument to the subscribe function. This class may be deleted on unsubscription, depending on the delete_pointer value passed during subscribtion.
};

#endif
