/*
	$Id: init_linux.cpp,v 1.13 1999/09/15 21:50:02 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Initializes ClanLib under Linux. Home of the legendary main().

*/

#include <sys/time.h>
#include <unistd.h>
#include <dlfcn.h>
#include <dirent.h>
#include <signal.h>
#include <assert.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include "System/API/system.h"
#include "System/API/clanapp.h"
#include "System/API/cl_assert.h"
#include "Input/API/input.h"
#include "Input/API/keyboard.h"
#include "Sound/API/sound.h"
#include "Display/API/display.h"
#include "Display/API/displaycard.h"
#include "System/API/error.h"
#include "clanlib.h"
#include "implementation.h"
#include "appconf.h"

extern "C"
{
	#include <Hermes/Hermes.h>
}

#include "init_linux.h"


#define TARGET_DISPLAY 0
#define TARGET_SOUND 1
#define TARGET_NETWORK 2


struct target_structure {
	void *dlhandle;
	clan_id_proc module_id;
	clan_id_proc module_abbr;
	clan_init_proc module_init;
};


CL_DisplayCard *cl_current_displaycard = NULL;
CL_SoundCard *cl_current_soundcard = NULL;
long _begin_time;



int select_display_target(const struct dirent *d)
{ 
	if (strstr(d->d_name,"libclan-display") == d->d_name &&
		strstr(d->d_name,".so") != NULL &&
		strstr(d->d_name,CL_VERSION_STRING) != NULL)
	{
		return 1;
	}
	else
	{
		return 0; 
	}
}

int select_sound_target(const struct dirent *d)
{ 
	if (strstr(d->d_name,"libclan-sound")==d->d_name &&
		strstr(d->d_name,".so") != NULL &&
		strstr(d->d_name,CL_VERSION_STRING) != NULL)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}


int select_network_target(const struct dirent *d)
{ 
	if (strstr(d->d_name,"libclan-network")==d->d_name &&
		strstr(d->d_name,".so") != NULL &&
		strstr(d->d_name,CL_VERSION_STRING) != NULL)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

#ifdef USE_DYN

static int num_targets_display=0;
static int num_targets_sound=0;
static int num_targets_network=0;

target_structure* detect_targets(int target_type,const char* target_dir) 
{
	struct dirent **filenames;
	int res = 0;

//	cout << "Scanning " << target_dir << endl;

	switch(target_type)
	{
	case TARGET_DISPLAY:
		res = scandir(
			target_dir,
			&filenames,
			select_display_target,
			alphasort);

		num_targets_display = res;
		break;

	case TARGET_SOUND:
		res = scandir(
			target_dir,
			&filenames,
			select_sound_target,
			alphasort);

		num_targets_sound = res;
		break;

	case TARGET_NETWORK:
		res = scandir(
			target_dir,
			&filenames,
			select_network_target,
			alphasort);

		num_targets_network=res;
		break;

	default:
		cl_assert(false);
	}

	if (res > 0)
	{
		char fullname[256];

		target_structure *targets=new target_structure[res];

		for (int i=0;i<res;i++)
		{
			targets[i].dlhandle=0;

			strcpy(fullname,target_dir);
			strcat(fullname,"/");
			strcat(fullname,filenames[i]->d_name);

//			cout << "Trying " << filenames[i]->d_name << "... ";
			
			targets[i].dlhandle=dlopen(fullname,RTLD_NOW);
			if (!targets[i].dlhandle)
			{
//				cout << "dlopen failed!" << endl;
//				cout << "Error: " << dlerror() << endl;
				continue;
			}

			char *error;

			targets[i].module_id = (clan_id_proc) dlsym(
					targets[i].dlhandle,
					"clan_module_identify");

			if ((error = dlerror()) != NULL)
			{
				cout << "identification failed!" << endl;
				continue;
			}

			targets[i].module_abbr = (clan_id_proc) dlsym(
				targets[i].dlhandle,
				"clan_module_abbreviation");

			if ((error = dlerror()) != NULL)
			{
				cout << "abbreviation identification failed!" << endl;
				continue;
			}
			
			targets[i].module_init = (clan_init_proc) dlsym(
				targets[i].dlhandle,
				"clan_module_init");

			if ((error = dlerror()) != NULL)
			{
				cout << "module initialisation unresolved!" << endl;
				continue;
			}

//			cout << "OK!" << endl;
		}

		return targets;
	}

	return 0;
}

#else

#ifdef USE_X11
#include "implementation_xwindow.h"
#endif

#ifdef USE_PTC
#include "implementation_ptc.h"
#endif

#ifdef USE_SVGALIB
#include "implementation_svgalib.h"
#endif

#ifdef USE_GGI
#include "implementation_ggi.h"
#endif

#ifdef USE_FBDEV
#include "implementation_fbdev.h"
#endif

#ifdef USE_OPENGL
#include "implementation_glx.h"
#endif

#ifdef USE_CLANSOUND
#include "implementation_clansound.h"
#endif

#ifdef USE_NETWORK
#include "Network/Unix/network_unix.h"
#endif

static int num_targets_display=6;
static int num_targets_sound=1;
static int num_targets_network=1;

target_structure* detect_targets(int target_type,const char* target_dir)
{
	static target_structure disp_targets[6]=
	{
		{(void*)0xff,xwindow_identify,xwindow_abbreviation,xwindow_init},
#ifdef USE_PTC
		{(void*)0xff,ptc_identify,ptc_abbreviation,ptc_init},
#else
		{(void*)0,(clan_id_proc)0,(clan_id_proc)0,(clan_init_proc)0},
#endif
#ifdef USE_SVGALIB
		{(void*)0xff,svgalib_identify,svgalib_abbreviation,svgalib_init},
#else
		{(void*)0,(clan_id_proc)0,(clan_id_proc)0,(clan_init_proc)0},
#endif
#ifdef USE_GGI
		{(void*)0xff,ggi_identify,ggi_abbreviation,ggi_init},
#else
		{(void*)0,(clan_id_proc)0,(clan_id_proc)0,(clan_init_proc)0},
#endif
#ifdef USE_FBDEV
		{(void*)0xff,fbdev_identify,fbdev_abbreviation,fbdev_init},
#else
		{(void*)0,(clan_id_proc)0,(clan_id_proc)0,(clan_init_proc)0},
#endif
#ifdef USE_OPENGL
		{(void*)0xff,glx_identify,glx_abbreviation,glx_init},
#else
		{(void*)0,(clan_id_proc)0,(clan_id_proc)0,(clan_init_proc)0},
#endif
	};

#ifdef USE_CLANSOUND
	static target_structure sound_targets[1]=
	{
		{(void*)0xff,clansound_identify,clansound_abbreviation,clansound_init}
	};
#endif

#ifdef USE_NETWORK
	static target_structure net_targets[1]=
	{
		{(void*)0xff,network_identify,network_abbreviation,network_init}
	};
#endif

	switch (target_type)
	{
	case TARGET_DISPLAY:
		return disp_targets;

#ifdef USE_CLANSOUND
	case TARGET_SOUND:
		return sound_targets;
#endif

#ifdef USE_NETWORK
	case TARGET_NETWORK:
		return net_targets;
#endif
	}

	// Make the compiler happy
	return (target_structure*)0x0;
}


#endif

void CL_System::init_display(int /*flags*/)
{
	FileConfig config("clanlib");
	CL_String display = config.readEntry("display", "ask");
	CL_String scandir = config.readEntry("targetdir", CL_SCANDIR);

	target_structure *targets = detect_targets(TARGET_DISPLAY,scandir);
	target_structure *chosen_display = 0;

	char choice[16];

	if (display == "ask")
	{
		cout << endl << "Please select a display target by entering its abbreviation: " << endl;
		
		for (int i=0;i<num_targets_display;i++) 
		{
			if (targets[i].dlhandle!=0)
			{
				cout << "	[" << targets[i].module_abbr() << "] ";
				cout << targets[i].module_id() << endl;
			}
		}

		cout << "Your choice: ";
		cin >> choice;
	}
	else
	{
		strcpy(choice, display);
	}

	for (int i=0;i<num_targets_display;i++) 
	{
		if (targets[i].dlhandle!=0)
		{
			if (!strcmp(targets[i].module_abbr(),choice))
			{
				chosen_display=&targets[i];
				break;
			}
		}
	}

	if (!chosen_display)
	{
		throw CL_Error("Cannot open display target");
	}
	else
	{
		for (int i=0;i<num_targets_display;i++)
		{
			if (targets[i].dlhandle!=0)
			{
				if (!strcmp(targets[i].module_abbr(),display))
				{
					chosen_display=&targets[i];
					break;
				}
			}
		}

		if (!chosen_display) throw CL_Error("Cannot open display target");
	}

	chosen_display->module_init();

	CL_Display::num_cards = CL_Display::cards.get_num_items();
	cl_current_displaycard = CL_Display::cards[0];
}

void CL_System::init_sound()
{
	FileConfig config("clanlib");
	CL_String scandir = config.readEntry("targetdir", CL_SCANDIR);

	target_structure* targets=detect_targets(TARGET_SOUND,scandir);

	if (num_targets_sound<=0 || targets[0].dlhandle==0)
		throw CL_Error("Unable to load any sound implementations");

	targets[0].module_init();

	CL_Sound::num_cards = CL_Sound::cards.get_num_items();
	cl_current_soundcard = CL_Sound::cards[0];
}

void CL_System::init_network()
{
	FileConfig config("clanlib");
	CL_String scandir = config.readEntry("targetdir", CL_SCANDIR);
	target_structure* targets=detect_targets(TARGET_NETWORK,scandir);

	if (num_targets_network<=0 || targets[0].dlhandle==0)
		throw CL_Error("Unable to load any network implementations");

	targets[0].module_init();
}

void CL_System::deinit_display()
{
	for (int i=0; i<CL_Display::num_cards; i++) delete CL_Display::cards[i];
	CL_Display::cards.clear();
	CL_Display::num_cards = 0;
}

void CL_System::deinit_sound()
{
	for (int i=0; i<CL_Sound::num_cards; i++) delete CL_Sound::cards[i]; 
	CL_Sound::cards.clear();
	CL_Sound::num_cards = 0;
}

void CL_System::deinit_network()
{
#ifdef USE_NETWORK
	// TODO: Add deinit code here...
#endif
}

void deinit()
{
	// Deinitialization:
	CL_System::deinit_display();
	CL_System::deinit_sound();
	CL_System::deinit_network();

	int num_keyboards = CL_Input::keyboards.get_num_items();
	int num_joysticks = CL_Input::joysticks.get_num_items();
	int num_pointers = CL_Input::pointers.get_num_items();

	for (int i=0; i<num_keyboards; i++) delete CL_Input::keyboards[i];
	for (int i=0; i<num_joysticks; i++) delete CL_Input::joysticks[i];
	for (int i=0; i<num_pointers; i++) delete CL_Input::pointers[i];
}

int main(int argc, char **argv)
{
	Hermes_Init();

	timeval tv;
	gettimeofday(&tv, NULL);
	_begin_time = (long) tv.tv_sec*(long) 1000+(long) tv.tv_usec/(long) 1000;

	cout << endl;
	cout << "ClanLib - '" << CL_RELEASE_NAME << "' (release " << CL_VERSION_STRING << ")" << endl;
	cout << endl;

	if (CL_ClanApplication::app == NULL)
	{
		cout << "ClanLib: No global CL_ClanApplication instance!!!" << endl;
		return 255;
	}

	int retval = CL_ClanApplication::app->main(argc, argv);

	deinit();

	return retval;
}

bool CL_System::keep_alive()
{
	bool retval = false;

	int num = CL_System_Generic::keep_alives.get_num_items();
	for (int i=0; i<num; i++)
	{
		if (CL_System_Generic::keep_alives[i]->keep_alive()) retval = true;
	}

	return retval;
}

int CL_System::get_time()
{
	timeval tv;
	gettimeofday(&tv, NULL);

	long tid = (long) tv.tv_sec*(long) 1000 + (long) tv.tv_usec/(long) 1000 - _begin_time;

	return (int) (tid);
}

void CL_System::sleep(int millis)
{
	usleep(millis*1000);
}

void CL_Display::select_card(CL_DisplayCard *card)
{
	cl_current_displaycard = card;
}

CL_DisplayCard *CL_Display::get_current_card()
{
	cl_assert(cl_current_displaycard != NULL);
	return cl_current_displaycard;
}

void CL_Sound::select_card(CL_SoundCard *card)
{
	cl_current_soundcard = card;
}

CL_SoundCard *CL_Sound::get_current_card()
{
	cl_assert(cl_current_soundcard != NULL);
	return cl_current_soundcard;
}

CL_Array <CL_System_KeepAlive> CL_System_Generic::keep_alives;
