/*
	$Id: layermanager.cpp,v 1.2 1999/02/03 22:37:11 starch Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Layer Manager

*/

#include "layermanager.h"

CL_LayerManager::CL_LayerManager(int _num_layers) : layers(_num_layers)
{
	num_layers = _num_layers;
	cur_layer_num = 0;
	current_layer = NULL;
}

CL_LayerManager::~CL_LayerManager()
{
	for (int i=0;i<layers.get_num_items();i++)
	{
		if (layers[i] != NULL)
		{
			delete layers[i];
		}
	}
}

void CL_LayerManager::put_screen(CL_Surface *surf, int x, int y, int frame, int layer)
{
	// sanity check
	cl_assert(surf != NULL);
	cl_assert(frame >= 0);

	SingleLayer *l = get_layer(layer);
	cl_assert(l!=NULL);
	l->add(surf, x, y, frame);
}

void CL_LayerManager::select_layer(int layer)
{
	cl_assert(layer >= 0 && layer < 100);
	cur_layer_num = layer;
	current_layer = get_layer(cur_layer_num);
	cl_assert(current_layer != NULL);
}

int	 CL_LayerManager::get_currentlayer()
{
	return cur_layer_num;
}

void CL_LayerManager::move_layer(int x_delta, int y_delta)
{
	current_layer->move(x_delta, y_delta);
}

void CL_LayerManager::move_layer(int layer, int x_delta, int y_delta)
{
	get_layer(layer)->move(x_delta, y_delta);
}

void CL_LayerManager::get_layerpos(int *x_delta, int *y_delta)
{
	current_layer->get_pos(x_delta, y_delta);
}

void CL_LayerManager::get_layerpos(int layer, int *x_delta, int *y_delta)
{
	get_layer(layer)->get_pos(x_delta, y_delta);
}

void CL_LayerManager::swap_layers(int layer1, int layer2)
{
	cl_assert(layer1 >= 0 && layer2 >= 0 &&
			 layer1 < 100 && layer2 < 100);

	SingleLayer *l1 = layers[layer1];
	SingleLayer *l2 = layers[layer2];
	if (l1 == NULL || l2 == NULL) return;

	layers.add(l1, layer2);
	layers.add(l2, layer1);
}

void CL_LayerManager::show_layers()
{
	for (int i=layers.get_num_items()-1;i>=0;i--)
	{
		if (layers[i] != NULL) layers[i]->show();
	}
}

void CL_LayerManager::show_singlelayer(int layer)
{
	if (layers[layer] != NULL) layers[layer]->show();
}

void CL_LayerManager::reset_layers()
{
	for (int i=0;i<layers.get_num_items();i++)
	{
		if (layers[i] != NULL) layers[i]->reset();
	}
}

void CL_LayerManager::reset_singlelayer(int layer)
{
	if (layers[layer] != NULL) layers[layer]->reset();
}

void CL_LayerManager::wrap_layer(int layer, int wrap_x, int wrap_y)
{
	if (layers[layer] != NULL) layers[layer]->wrap_layer(wrap_x, wrap_y);
}

CL_LayerManager::SingleLayer *CL_LayerManager::get_layer(int num)
{
	cl_assert(num >= -1 && num < 100);

	SingleLayer *l;

	if (num==-1)
	{
		if (current_layer == NULL)
		{
			l = layers[0];
			if (l == NULL)
			{
				l = new SingleLayer;
				layers.add(l, 0);
			}
			current_layer = l;
		}

		return current_layer;
	}
	else
	{
		l = layers[num];
		if (l == NULL)
		{
			l = new SingleLayer;
			layers.add(l, num);
		}
	}

	return l;
}

CL_LayerManager::SingleLayer::SingleLayer()
{
	x = y = 0;
}

CL_LayerManager::SingleLayer::~SingleLayer()
{
	reset();
}

void CL_LayerManager::SingleLayer::move(int _x, int _y)
{
	x = _x;
	y = _y;
}

void CL_LayerManager::SingleLayer::get_pos(int *_x, int *_y)
{
	*_x = x;
	*_y = y;
}

void CL_LayerManager::SingleLayer::add(CL_Surface *_s, int _x, int _y, int _frame)
{
	DrawItem *new_item = new DrawItem;
	cl_assert(new_item != NULL);
	
	new_item->s = _s;
	new_item->x = _x;
	new_item->y = _y;
	new_item->frame = _frame;

	draw_items.add(new_item);
}

void CL_LayerManager::SingleLayer::show()
{
	CL_Iterator<DrawItem> counter(draw_items);
	while (counter.next() != NULL)
	{
		DrawItem *c = counter();
		c->s->put_screen(c->x+x, c->y+y, c->frame);
	}
}

void CL_LayerManager::SingleLayer::reset()
{
	CL_Iterator<DrawItem> counter(draw_items);
	while (counter.next() != NULL)
	{
		DrawItem *del_item = counter();
		counter.remove();
		delete del_item;
	}
}

void CL_LayerManager::SingleLayer::wrap_layer(int wrap_x, int wrap_y)
{
	CL_Iterator<DrawItem> counter(draw_items);
	while (counter.next() != NULL)
	{
		DrawItem *cur = counter();
		if (cur->x+x >= wrap_x) cur->x -= wrap_x;
		if (cur->y+y >= wrap_y) cur->y -= wrap_y;
	}
}
