/*
	$Id: menu.cpp,v 1.4 1999/08/23 08:39:21 sphair Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#include "../../clanlib.h"
#include "../../clanlayer2.h"

#include "menu.h"

CL_MenuItem::CL_MenuItem( const CL_String _text, int _id, int _parent/*=0*/ )
{
	text = _text;
	id = _id;
	parent = _parent;
}

bool CL_MenuItem::has_children()
{
	return children.get_num_items() > 0;
}

void CL_MenuItem::add_child( CL_MenuItem* child )
{
	children.add( child );
}

int CL_MenuItem::get_id() const
{
	return id;
}

int CL_MenuItem::get_parent() const
{
	return parent;
}

CL_String CL_MenuItem::get_text()
{
	return text;
}

void CL_MenuItem::set_text(const CL_String _text)
{
	text = _text;
}


CL_MenuItem_Value::CL_MenuItem_Value( const CL_String _text, int _id, int _parent, int _min, int _max, int _value ) : CL_MenuItem( _text, _id, _parent )
{
	value = _value;
	min = _min;
	max = _max;
	test_value();
}

int CL_MenuItem_Value::get_value()
{
	return value;
}

int CL_MenuItem_Value::get_min()
{
	return min;
}

int CL_MenuItem_Value::get_max()
{
	return max;
}

void CL_MenuItem_Value::set_min( int _min )
{
	min = _min;
	test_value();
}

void CL_MenuItem_Value::set_max( int _max )
{
	max = _max;
	test_value();
}

void CL_MenuItem_Value::set_value( int _value )
{
	value = _value;
	test_value();
}

void CL_MenuItem_Value::inc_value()
{
	value++;
	test_value();
}

void CL_MenuItem_Value::dec_value()
{
	value--;
	test_value();
}

void CL_MenuItem_Value::test_value()
{
	if (value > max)
	{
	  	value = max;
	}
	if (value < min)
	{
		value = min;
	}
}

// *********************
CL_MenuItem_String::CL_MenuItem_String( const CL_String _text, int _id, int _parent, CL_String _string) : CL_MenuItem( _text, _id, _parent )
{
	string = _string;
}

CL_String& CL_MenuItem_String::get_string()
{
	return string;
}

void CL_MenuItem_String::set_string( CL_String  _string )
{
	string = _string;
}

// *********************
CL_MenuItem_StringList::CL_MenuItem_StringList( const CL_String _text, int _id, int _parent, CL_Array<CL_String> _string_list, int _value ) : CL_MenuItem( _text, _id, _parent )
{
	set_strings( _string_list );
	value = _value;
	test_value();
}

CL_String CL_MenuItem_StringList::get_string()
{
	return *(string_list[value]);
}

CL_Array<CL_String> CL_MenuItem_StringList::get_strings()
{
	return string_list;
}

void CL_MenuItem_StringList::set_value( int _value )
{
	value = _value;
	test_value();
}

int CL_MenuItem_StringList::get_value()
{
	return value;
}

void CL_MenuItem_StringList::set_strings( CL_Array<CL_String>  _string_list )
{
	min = 0;
	max = _string_list.get_num_items() - 1;
	for (int i=min; i<=max; i++)
		string_list.add( new CL_String( *(_string_list[i]) ) );
}

void CL_MenuItem_StringList::inc_value()
{
	value++;
	test_value();
}

void CL_MenuItem_StringList::dec_value()
{
	value--;
	test_value();
}

void CL_MenuItem_StringList::test_value()
{
	if (value > max)
	{
	  	value = max;
	}
	if (value < min)
	{
		value = min;
	}
}

// ************************
CL_Menu::CL_Menu( const CL_String& name, CL_MenuDisplay *_menu_display )
{
	menu_display = _menu_display;
	current_run_id = 0;
	current_selection = 1;

	key_down = false;
	key_up = false;
	key_enter = false;
	key_esc = false;

	enter = false;

	key_buffer = CL_Input::keyboards[0]->get_buffer(0);

	items.add( new CL_MenuItem(name, 0, -1) );
}

CL_Menu::~CL_Menu()
{
}

int CL_Menu::execute()
{
	enter = false;
	key_esc = true;
	bool must_redraw = true;
	
	while (1)
	{
		CL_MenuItem* current = get_item_by_id(current_run_id);

		if (must_redraw)
		{
			CL_Iterator<CL_MenuItem> item_counter( current->children );

			menu_display->begin_draw();
			menu_display->draw_title(current->get_text());
			menu_display->draw_elements(item_counter, current_selection);
			must_redraw = menu_display->end_draw();
		}
		
		CL_System::keep_alive();

		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_DOWN) && !key_down)
		{
			current_selection++;
			if (current_selection > current->children.get_num_items())
				current_selection = get_item_by_id(current_run_id)->children.get_num_items();
			must_redraw = true;
			menu_display->play_sfx_break();
		}
		key_down = CL_Keyboard::get_keycode(CL_Keyboard::KEY_DOWN);
	
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_UP) && !key_up)
		{
			current_selection--;
			if (current_selection < 1) current_selection = 1;
			must_redraw = true;
			menu_display->play_sfx_break();
		}
		key_up =CL_Keyboard::get_keycode(CL_Keyboard::KEY_UP);
	
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_RIGHT) && !key_right)
		{
			if (current->children.get_item(current_selection-1)->get_type() == CL_MenuItem::MT_VALUE)
			{
				((CL_MenuItem_Value*)(current->children.get_item(current_selection-1)))->inc_value();
			}
			if (current->children.get_item(current_selection-1)->get_type() == CL_MenuItem::MT_STRINGLIST)
			{
				((CL_MenuItem_StringList*)(current->children.get_item(current_selection-1)))->inc_value();
			}
			menu_display->play_sfx_break();
		}
		key_right = CL_Keyboard::get_keycode(CL_Keyboard::KEY_RIGHT);
		
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_LEFT) && !key_left)
		{
			if (current->children.get_item(current_selection-1)->get_type() == CL_MenuItem::MT_VALUE)
			{
				((CL_MenuItem_Value*)(current->children.get_item(current_selection-1)))->dec_value();
			}
			if (current->children.get_item(current_selection-1)->get_type() == CL_MenuItem::MT_STRINGLIST)
			{
				((CL_MenuItem_StringList*)(current->children.get_item(current_selection-1)))->dec_value();
			}
			menu_display->play_sfx_break();
		}
		key_left = CL_Keyboard::get_keycode(CL_Keyboard::KEY_LEFT);
	
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_ENTER) && !key_enter)
		{
			if (current->children.get_item(current_selection-1)->has_children())
			{
				current_run_id = current->children.get_item(current_selection-1)->get_id();
				current_selection = 1;
				must_redraw = true;
				menu_display->play_sfx_clear();
			} else
			{
				enter = true;
				if (current->children.get_item(current_selection-1)->get_type() == CL_MenuItem::MT_STRING)
				{
					enter_string( (CL_MenuItem_String*)(current->children.get_item(current_selection-1)) );
				}
			}
		}
		key_enter = CL_Keyboard::get_keycode(CL_Keyboard::KEY_ENTER);

		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_ESCAPE) && !key_esc)
		{
			if (current->get_id() != 0)
			{
				current_run_id = current->get_parent();
				current_selection = 1;
				must_redraw = true;
				menu_display->play_sfx_clear();
			}
		}
		key_esc = CL_Keyboard::get_keycode(CL_Keyboard::KEY_ESCAPE);

		if (key_left || key_right || enter)
		{
			enter = false;
			return current->children.get_item(current_selection-1)->get_id();
		}
	}
	return -1;			
}

void CL_Menu::enter_string(CL_MenuItem_String* item)
{
	return; // disabled for repairs!
/*
	while (CL_Keyboard::get_keycode(CL_Keyboard::KEY_ENTER))
	{
		CL_System::keep_alive();
	}
	
	key_buffer->clear();
	CL_String new_string = item->get_string();
	
	while (1)
	{
		resources->sur_titlescreen->put_screen(0, 0, 0 );
//		CL_Display::fill_rect( 200,300, 600,400, 0,0,0,0.5f);

		resources->fnt_clansoft->print_left( 230, 330, item->get_text() );
		
		resources->fnt_clansoft->print_left( 380, 330, new_string );
		
		CL_Display::flip_display();
		CL_System::keep_alive();
		
		switch (key_buffer->get_key())
		{
			case CL_Keyboard::KEY_ENTER:
				item->set_string( new_string );
				return;
			break;
			case CL_Keyboard::KEY_BACKSPACE:
				new_string = new_string.mid( 0, new_string.get_length()-1 );
			break;
			case CL_Keyboard::KEY_SPACE:
				new_string += " ";
			break;
			case CL_Keyboard::KEY_A:
				new_string += "a";
			break;
			case CL_Keyboard::KEY_B:
				new_string += "b";
			break;
			case CL_Keyboard::KEY_C:
				new_string += "c";
			break;
			case CL_Keyboard::KEY_D:
				new_string += "d";
			break;
			case CL_Keyboard::KEY_E:
				new_string += "e";
			break;
			case CL_Keyboard::KEY_F:
				new_string += "f";
			break;
			case CL_Keyboard::KEY_G:
				new_string += "g";
			break;
			case CL_Keyboard::KEY_H:
				new_string += "h";
			break;
			case CL_Keyboard::KEY_I:
				new_string += "i";
			break;
			case CL_Keyboard::KEY_J:
				new_string += "j";
			break;
			case CL_Keyboard::KEY_K:
				new_string += "k";
			break;
			case CL_Keyboard::KEY_L:
				new_string += "l";
			break;
			case CL_Keyboard::KEY_M:
				new_string += "m";
			break;
			case CL_Keyboard::KEY_N:
				new_string += "n";
			break;
			case CL_Keyboard::KEY_O:
				new_string += "o";
			break;
			case CL_Keyboard::KEY_P:
				new_string += "p";
			break;
			case CL_Keyboard::KEY_Q:
				new_string += "q";
			break;
			case CL_Keyboard::KEY_R:
				new_string += "r";
			break;
			case CL_Keyboard::KEY_S:
				new_string += "s";
			break;
			case CL_Keyboard::KEY_T:
				new_string += "t";
			break;
			case CL_Keyboard::KEY_U:
				new_string += "u";
			break;
			case CL_Keyboard::KEY_V:
				new_string += "v";
			break;
			case CL_Keyboard::KEY_W:
				new_string += "w";
			break;
			case CL_Keyboard::KEY_X:
				new_string += "x";
			break;
			case CL_Keyboard::KEY_Y:
				new_string += "y";
			break;
			case CL_Keyboard::KEY_Z:
				new_string += "z";
			break;
		}
		if (new_string.get_length() == 1)
		{
			new_string.to_upper();
		}
	}
*/
}

void CL_Menu::add_item( const CL_String& text, int id, int parent )
{
	CL_MenuItem* new_item = new CL_MenuItem( text, id, parent );

	get_item_by_id(parent)->add_child(new_item);
	items.add( new_item );
}

void CL_Menu::add_value( const CL_String& text, int id, int parent, int min, int max, int value )
{
	CL_MenuItem_Value* new_item = new CL_MenuItem_Value( text, id, parent, min, max, value );

	get_item_by_id(parent)->add_child(new_item);
	items.add( new_item );
}

void CL_Menu::add_string( const CL_String& text, int id, int parent, CL_String string )
{
	CL_MenuItem_String*  new_item = new CL_MenuItem_String( text, id, parent, string);
	
	get_item_by_id(parent)->add_child(new_item);
	items.add( new_item );
}

void CL_Menu::add_stringlist( const CL_String& text, int id, int parent, CL_Array<CL_String> string_list, int cur_string )
{
	CL_MenuItem_StringList*  new_item = new CL_MenuItem_StringList( text, id, parent, string_list, cur_string );
	
	get_item_by_id(parent)->add_child(new_item);
	items.add( new_item );
}

CL_MenuItem* CL_Menu::get_item_by_id( int id )
{
	CL_Iterator<CL_MenuItem> item_counter(items);
	while (item_counter.next() != NULL)
	{
		if (item_counter()->get_id() == id)
		{
			return item_counter();
		}
	}
	return NULL;
}

// ------------------- Standard, totally borring menu type ------------------
//  the one used in ClanBomber. :-)

CL_MenuDisplay_Simple::CL_MenuDisplay_Simple(
	CL_Surface *sur_select_icon,
	CL_Font *font,
	CL_SoundBuffer *sfx_break,
	CL_SoundBuffer *sfx_clear)
{
	m_sur_select_icon = sur_select_icon;
	m_font = font;
	m_sfx_break = sfx_break;
	m_sfx_clear = sfx_clear;
}

void CL_MenuDisplay_Simple::begin_draw()
{
	CL_Display::clear_display(0.0f, 0.0f, 0.3f);
}

void CL_MenuDisplay_Simple::draw_title(const char *title)
{
/*
			int vert = 260 - (current->children.num_items*40) / 2;
			resources->fnt_clansoft->print_center(
				CL_Display::get_width()/2,
				vert,
				current->get_text() );

			vert += 80;
*/
}

void CL_MenuDisplay_Simple::draw_elements(
	CL_Iterator<CL_MenuItem> &item_counter,
	int current_selection)
{
	int vert = 0;
	while (item_counter.next() != NULL) vert += m_font->get_height();

	vert = CL_Display::get_height()/2-vert/2;

	int act_draw = 0;
	while (item_counter.next() != NULL)
	{
		act_draw++;
		if (act_draw == current_selection)
		{
//			m_font->print_left( 235, vert-4, CL_String(">") );
			m_sur_select_icon->put_screen(235, vert);
		}
		if (item_counter()->get_type() == CL_MenuItem::MT_VALUE)
		{
			if ((((CL_MenuItem_Value*)item_counter())->get_min() == 0) && (((CL_MenuItem_Value*)item_counter())->get_max() == 1))
			{
				m_font->print_right( 560, vert, CL_String( ((CL_MenuItem_Value*)item_counter())->get_value() ? "Yes" : "No" ) );				
			}
			else
			{
				m_font->print_right( 560, vert, CL_String( ((CL_MenuItem_Value*)item_counter())->get_value() ) );
			}
		}
		if (item_counter()->get_type()==CL_MenuItem::MT_STRING)
		{
			m_font->print_right( 560, vert, ((CL_MenuItem_String*)item_counter())->get_string() );
		}
		if (item_counter()->get_type()==CL_MenuItem::MT_STRINGLIST)
		{
			m_font->print_right( 560, vert, ((CL_MenuItem_StringList*)item_counter())->get_string() );
		}
		m_font->print_left( 255, vert, item_counter()->get_text() );
		vert += m_font->get_height();
	}

	CL_Display::flip_display();
}

bool CL_MenuDisplay_Simple::end_draw()
{
	return false;
}

void CL_MenuDisplay_Simple::play_sfx_break()
{
	if (m_sfx_break != NULL) m_sfx_break->play();
}

void CL_MenuDisplay_Simple::play_sfx_clear()
{
	if (m_sfx_clear != NULL) m_sfx_clear->play();
}
