/*
	$Id: menu.h,v 1.2 1999/07/06 18:13:56 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#ifndef header_menu
#define header_menu

class CL_Menu;
class CL_MenuItem
{
 friend CL_Menu;
 public:
  CL_MenuItem( const CL_String _text, int _id, int _parent=0 );
  virtual ~CL_MenuItem() {};

  void add_child( CL_MenuItem* child );
  bool has_children();
  int get_id() const;
  int get_parent() const;
  CL_String get_text();
  void set_text( const CL_String _text );

  typedef enum
  {
	MT_NORMAL,
	MT_VALUE,
	MT_STRING,
	MT_STRINGLIST
  } MENUTYPE;

  virtual MENUTYPE get_type() { return MT_NORMAL; };

 protected:
  CL_String		text;
  int			id;
  int			parent;
  CL_List<CL_MenuItem>	children;
};

class CL_MenuItem_Value : public CL_MenuItem
{
friend CL_Menu;
public:
	CL_MenuItem_Value( const CL_String _text, int _id, int _parent, int _min, int _max, int _value );
	virtual ~CL_MenuItem_Value() {};

	int get_value();
	void set_value(int _value);
	int get_min();
	int get_max();
	void set_min( int _min );
	void set_max( int _max );
	void inc_value();
	void dec_value();

	virtual MENUTYPE get_type() { return MT_VALUE; };

protected:
	void test_value();
	int value;
	int min;
	int max;
};

class CL_MenuItem_String : public CL_MenuItem
{
friend CL_Menu;
public:
	CL_MenuItem_String( const CL_String _text, int _id, int _parent, CL_String _string);
	virtual ~CL_MenuItem_String() {};

	CL_String& get_string();
	void set_string(CL_String _string);

	virtual MENUTYPE get_type() { return MT_STRING; };

protected:
	CL_String string;
};

class CL_MenuItem_StringList : public CL_MenuItem
{
friend CL_Menu;
public:
	CL_MenuItem_StringList( const CL_String _text, int _id, int _parent, CL_Array<CL_String> _string_list, int _value );
	virtual ~CL_MenuItem_StringList() {};

	CL_String get_string();
	CL_Array<CL_String> get_strings();
	void set_strings(CL_Array<CL_String> _string_list);
	void inc_value();
	void dec_value();
	int get_value();
	void set_value(int _value);

	virtual MENUTYPE get_type() { return MT_STRINGLIST; };

protected:
	CL_Array<CL_String> string_list;
	void test_value();
	int value;
	int min;
	int max;
};

class CL_MenuDisplay // inherit this class to customize menues.
{
public:
	virtual ~CL_MenuDisplay() {;}

protected: // the following functions are called by CL_Menu:
friend CL_Menu;

	virtual void begin_draw()=0;
	virtual void draw_title(const char *title)=0;
	virtual void draw_elements(
		CL_Iterator<CL_MenuItem> &item_counter,
		int current_selection)=0;
	virtual bool end_draw()=0;

	virtual void play_sfx_break()=0;
	virtual void play_sfx_clear()=0;
};

class CL_Menu
{
public:
	CL_Menu(const CL_String& name, CL_MenuDisplay *menu_display);
	~CL_Menu();

	void add_item( const CL_String& text, int id, int parent=0 );
	void add_value( const CL_String& text, int id, int parent, int min, int max, int value );
	void add_string( const CL_String& text, int id, int parent, CL_String string );
	void add_stringlist( const CL_String& text, int id, int parent, CL_Array<CL_String> string_list, int cur_string );
	int execute();
	CL_MenuItem* get_item_by_id( int id );

protected:
	CL_MenuDisplay *menu_display;
	CL_List<CL_MenuItem> items;
	int current_run_id;
	int current_selection;
	void enter_string(CL_MenuItem_String* _item);
	
	CL_InputBuffer* key_buffer;

	bool key_up, key_down, key_right, key_left, key_enter, key_esc, enter;
};

class CL_MenuDisplay_Simple : public CL_MenuDisplay
{
public:
	CL_MenuDisplay_Simple(
		CL_Surface *sur_select_icon,
		CL_Font *font,
		CL_SoundBuffer *sfx_break = NULL,
		CL_SoundBuffer *sfx_clear = NULL);

protected: // the following functions are called by CL_Menu:

	virtual void begin_draw();
	virtual void draw_title(const char *title);
	virtual void draw_elements(
		CL_Iterator<CL_MenuItem> &item_counter,
		int current_selection);
	virtual bool end_draw();

	virtual void play_sfx_break();
	virtual void play_sfx_clear();

private:
	CL_Surface *m_sur_select_icon;
	CL_Font *m_font;
	CL_SoundBuffer *m_sfx_break;
	CL_SoundBuffer *m_sfx_clear;
};

#endif
