/*
	$Id: netdata_controller.cpp,v 1.2 1999/09/02 14:45:43 sphair Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#include "netdata_controller.h"

#ifndef min
#define min(a, b) ((a<b) ? (a) : (b))
#endif

/*****************************************************************************/

CL_NetDataInfo::CL_NetDataInfo(
	const CL_NetComputer *netdest, CL_NetDataObject *netobject, 
	int netid, int nettype, int netpriority)
{
	dest=netdest;
	object=netobject;

	id=netid;
	type=nettype;
	priority=netpriority;

	data_size=data_left=object->get_size();
}

/*****************************************************************************/

CL_NetDataInfo::CL_NetDataInfo(CL_NetDataObject *netobject, int netid, int nettype, int size)
{
	dest=NULL;
	object=netobject;

	id=netid;
	type=nettype;
	priority=0;

	data_size=size;
	data_left=0;
}

/*****************************************************************************/

void CL_NetFileController::send(const CL_NetComputer *dest, char *filename, int type, int priority)
{
	cl_assert(game!=NULL);

	CL_NetFileObject *object=new CL_NetFileObject(filename);
	CL_NetDataController::send(dest, object, type, priority);
}

/*****************************************************************************/

CL_NetDataController::CL_NetDataController()
{
	game=NULL;
	channel=0;

	next_object_id = 0;
}

void CL_NetDataController::init(CL_NetDataCreator *netcreator, CL_NetGame *netgame, int netchannel)
{
	creator=netcreator;
	game=netgame;
	channel=netchannel;
}

void CL_NetDataController::send(const CL_NetComputer *dest, CL_NetDataObject *object, int type, int priority)
{
	cl_assert(game!=NULL);

	CL_NetDataInfo *info=new CL_NetDataInfo(dest, object, next_object_id++, type, priority);
	serverTransfers.add(info);
}

bool CL_NetDataController::update_server(int priority, const CL_NetComputer *dest)
{
	bool is_finished=true;
	
	cl_assert(game!=NULL);

	for (int i=0; i<serverTransfers.get_num_items(); i++)
	{
		if (serverTransfers[i] != NULL)
		{
			CL_NetDataInfo *info=serverTransfers[i];

			if((info->dest==dest || dest==NULL) && info->priority <= priority) {
				CL_OutputSource_Memory output;

				if(info->data_left==0) {
					output.write_int32(EndTransfer);
					output.write_int32(info->id);

					info->object->serialize_save(&output, info->data_size, 0);
					serverTransfers.del(i);
				}
				else {
					if(info->data_left==info->data_size) {
						output.write_int32(InitTransfer);
						output.write_int32(info->id);
						output.write_int32(info->type);
						output.write_int32(info->object->get_size());
					}
					else {
						output.write_int32(Transfer);
						output.write_int32(info->id);
					}

					int packet_size = min(MAX_PACKET_SIZE, info->data_left);
					info->data_left -= packet_size;
					int data_pos = info->data_size - info->data_left - packet_size;

					output.write_int32(data_pos);
					info->object->serialize_save(&output, data_pos, packet_size);

//					cout << "Sending packet " << (data_pos / MAX_PACKET_SIZE)+1 << endl;
				}
	
				game->send(
					channel, info->dest,
					CL_NetMessage(output.get_data(), output.size()));
				is_finished=false;
			}
		}
	}

	return is_finished;
}

void CL_NetDataController::update_client()
{
	cl_assert(game!=NULL);

	while (game->peek(channel))
	{
		CL_NetMessage message = game->receive(channel);
		CL_InputSource_Memory input(message.data, message.size, true);
		
		cl_assert(input.size() > 0);

		TransferMode command = (TransferMode) input.read_int32();
		int netobject_id = input.read_int32();

		if(command==InitTransfer) {
			int netobject_type = input.read_int32();
			int netobject_size = input.read_int32();
			int pos=input.read_int32();
			
			if (creator != NULL)
			{
				CL_NetDataObject *object = creator->create_data(netobject_type, netobject_size);
				
				if (object != NULL)
				{
					CL_NetDataInfo *info;
					info = new CL_NetDataInfo(object, netobject_id, netobject_type, netobject_size); 
					clientTransfers.add(info);

					info->object->serialize_load(&input, pos, input.size()-20);
				}
			}
		}

		else if(command==Transfer) {
			int pos=input.read_int32();

			int i;
			for (i=0; i<clientTransfers.get_num_items(); i++)
			{
				if(clientTransfers[i]->id == netobject_id) {
					clientTransfers[i]->object->serialize_load(&input, pos, input.size()-12);
					break;
				}
			}

			cl_assert(i!=clientTransfers.get_num_items());
		}

		else if(command==EndTransfer) {
			int i;
			for (i=0; i<clientTransfers.get_num_items(); i++)
			{
				if(clientTransfers[i]->id == netobject_id) {
					clientTransfers[i]->object->serialize_load(&input, clientTransfers[i]->data_size, 0);
					break;
				}
			}

			clientTransfers.del(i);
		}
	}
}
