/*
	$Id: resource_manager.cpp,v 1.17 1999/07/18 18:36:04 arjanl Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

#include "clanlib.h"
#include "resource_manager.h"
#include "resourceoptions.h"
#include "resourcetype_surface.h"
#include "resourcetype_sample.h"
#include "resourcetype_font.h"
#include "resourcetype_palette.h"
#include "resourcetype_raw.h"
#include "resourcetype_integer.h"
#include "resourcetype_string.h"
#include "resourcetype_boolean.h"

#include "Layer1/System/API/error.h" 

#define MAX_SECTION_LENGTH 200

/******************************************************************************
			class CL_ResourceManager
******************************************************************************/

bool CL_ResourceManager::std_resourcetypes_added = false;

CL_ResourceManager::CL_ResourceManager(
	const char *config_file,
	CL_InputSourceProvider *provider,
	bool read_directly_from_source,
	bool delete_inputsource_provider)
{
	init(
		config_file,
		provider,
		read_directly_from_source,
		delete_inputsource_provider);
}

/* old constructor using non-default config_id */
CL_ResourceManager::CL_ResourceManager(
	const char *config_id,
	const char *datafile_name)
{
	init(
		config_id,
		CL_InputSourceProvider::create_datafile_provider(datafile_name),
		false,
		true);
}

CL_ResourceManager::CL_ResourceManager(
	const char *file_name, const bool is_datafile )
{
	if( is_datafile )
	{
		init(
			default_scriptfile_id,
			CL_InputSourceProvider::create_datafile_provider(file_name),
			false,
			true);
	}
	else
	{
		init(
			file_name,
			NULL,
			true,
			true);
	}
}


void CL_ResourceManager::init(
	const char *_config_file,
	CL_InputSourceProvider *provider,
	bool read_directly_from_source,
	bool delete_inputsource_provider)
{
	add_resourcetypes();

	ignore_datafile = read_directly_from_source;
	root_section = NULL;
	config_handle = _config_file;

	CL_InputSourceProvider *config_provider = NULL;
	if (provider == NULL)
	{
		config_provider = CL_InputSourceProvider::create_file_provider("");
	}
	else
	{
		config_provider = provider;
	}

	CL_InputSource *config_file = NULL;
	try
	{
		config_file = config_provider->open_source(_config_file);
	}
	catch (CL_Error err)
	{
		// backward compatibility:
		//   try open default_scriptfile_id if config_file is the incorrect
		//   name.

		if (read_directly_from_source == false)
			config_file = config_provider->open_source(default_scriptfile_id);
		else
			throw err;
	}
	
	if( config_file==NULL )
		cout << "Configfile not found. Did you compile the datafile with the correct sectionname?" << endl;
	cl_assert(config_file != NULL);

	total_size = config_file->size();
	config_file_data = new char[total_size];
	config_file->read(config_file_data, total_size);

	delete config_file;

	cur_pos = 0;
	cur_line = 1;
	in_remark = false;
	data_in_datafile = false;

	root_section = new CL_ResourceSection("", "", this, true);

//	cout << "Reading config from input '" << config_handle << "'" << endl;
	bool result = root_section->read_config();
	if (result == false) throw CL_Error("Failed to read input");
//	cl_assert(result);

	if (provider != NULL) // mbn: use current datafile if orig source
	{
		data_in_datafile = true;
		datafile_name = "resource-hack.dat";
		resource_provider = provider->clone();
	}
	else if (data_in_datafile)
	{
//		cout << "Opening datafile '" << datafile_name << "'" << endl;
		resource_provider =
			CL_InputSourceProvider::create_datafile_provider(datafile_name);
	}
	else
	{
		resource_provider =
			CL_InputSourceProvider::create_file_provider("");
	}

	if (provider == NULL || delete_inputsource_provider) delete config_provider;
}

CL_ResourceManager::~CL_ResourceManager()
{
	delete[] config_file_data;
	delete root_section;
	delete resource_provider;
}

bool CL_ResourceManager::load_all_resources()
{
	return root_section->load();
}

bool CL_ResourceManager::load_section(const char *section_name)
{
	CL_ResourceSection *section;
	try
	{
		section = root_section->get_section(section_name);
	}
	catch (CL_Error err)
	{
		return false;
	}
//	if (section == NULL) return false;

	return section->load();
}

bool CL_ResourceManager::is_operator(char c)
{
	return c == ',' || c == '=' || c == '(' || c == ')' || c == ';' ||
		   c == '{' || c == '}';
}

bool CL_ResourceManager::is_space(char c)
{
	return c == ' ' || c == '\t' || c == 13 || c == 10;
}

bool CL_ResourceManager::is_remark(char c, int position)
{
	if (in_remark)
	{
		if (c == '\n') 
		{	
			in_remark = false;
		}
		return true;
	}
	else
	{
		if (c == '/')
		{
			if (position+1 < total_size && config_file_data[position+1] == '/')
			{
				in_remark = true;
				return true;
			}
		}
		else if (c == '#')
		{
			in_remark = true;
			return true;
		}
	}
	return false;
}
	
CL_String CL_ResourceManager::get_next_token()
{
	while (cur_pos < total_size && 
           (is_remark(config_file_data[cur_pos], cur_pos) ||
            is_space(config_file_data[cur_pos])))
	{
		if (config_file_data[cur_pos] == '\n') cur_line++;
		cur_pos++;
	}
	if (cur_pos == total_size) return "";

	if (is_operator(config_file_data[cur_pos]))
	{
		char temp[2];
		temp[0] = config_file_data[cur_pos];
		temp[1] = 0;
		CL_String ret(temp);
		cur_pos++;

		return ret;
	}
	else
	{
		int temp_pos = cur_pos;
		if (config_file_data[temp_pos] == '"')
		{
			int num_chars = 0;
			temp_pos++;
			while (temp_pos < total_size)
			{
				if (config_file_data[temp_pos] == '"')
				{
					if (temp_pos+1>=total_size || config_file_data[temp_pos+1] != '"') break;
					else
					{
						temp_pos++;
					}
				}
				temp_pos++;
				num_chars++;
			}
			if (temp_pos == total_size)
			{
				write_error("Missing '""' following '""'-begin");
				return "";
			}

			char *temp = new char[num_chars+1];
			temp[num_chars] = 0;
			int ofs = 1;
			for (int i=0;i<num_chars;i++)
			{
				temp[i] = config_file_data[cur_pos+ofs+i];
				if (config_file_data[cur_pos+ofs+i] == '"') 
				{
					ofs++;
				}
			}
			cur_pos = temp_pos+1;

			CL_String ret(temp);
			delete temp;

			return ret;
		}
		else
		{
			while (temp_pos < total_size &&
				   !is_space(config_file_data[temp_pos]) &&
				   !is_operator(config_file_data[temp_pos]) &&
				   !is_remark(config_file_data[temp_pos], cur_pos))
			{
				temp_pos++;
			}
			int tok_size = temp_pos-cur_pos;
			char *temp = new char[tok_size+1];
			temp[tok_size] = 0;
			memcpy(temp, &config_file_data[cur_pos], tok_size);
			cur_pos += tok_size;

			CL_String ret(temp);
			delete temp;

			return ret;
		}
	}
}

void CL_ResourceManager::write_error(CL_String err_msg)
{
	cout << "Error occured during read of config file '" <<
		     config_handle << "' (line " << cur_line << "): " << err_msg << endl;
}

void CL_ResourceManager::get_section_resources(
			CL_String section_prefix, 
			CL_ResourceSection *sect, 
			CL_List<CL_String> *ret, 
			const char *match_type) 
{
	CL_List<CL_Resource> *res = sect->get_resources();
	CL_Iterator<CL_Resource> counter(res);
	while (counter.next() != NULL) 
	{
		if (match_type == NULL || counter()->get_type() == match_type)
		{
			if (section_prefix == "")
			{
				// Hmmmmm, apparently it is necessary to store the name in a variable before
				// passing it to the string constructor - VC60 causes an internal compiler
				// error otherwise :-(
				CL_String res_name = counter()->get_name();
				ret->add(new CL_String(res_name));
			}
			else
			{
				ret->add(new CL_String(section_prefix+"/"+counter()->get_name()));
			}
		}
	}
	
	CL_List<CL_ResourceSection> *sections = sect->get_sections();
	CL_Iterator<CL_ResourceSection> counter2(sections);
	while (counter2.next() != NULL)
	{
		CL_String new_prefix;
		if (section_prefix == "") new_prefix = counter2()->get_name();
		else new_prefix = section_prefix + "/" + counter2()->get_name();

		get_section_resources(new_prefix, counter2(), ret, match_type);
	}
}

CL_List<CL_String> *CL_ResourceManager::get_all_resources()
{
	CL_List<CL_String> *ret = new CL_List<CL_String>;
	get_section_resources("", root_section, ret, NULL); 
	return ret;
}

CL_List<CL_String> *CL_ResourceManager::get_resources_of_type(CL_String type_id)
{
//	type_id.to_lower();

	CL_List<CL_String> *ret = new CL_List<CL_String>;
	get_section_resources("", root_section, ret, type_id); 
	return ret;
}

CL_InputSourceProvider *CL_ResourceManager::get_resource_provider()
{
	return resource_provider;
}

CL_Resource *CL_ResourceManager::get_resource(CL_String res_id)
{
	return root_section->get_resource(res_id);
}

void CL_ResourceManager::add_resourcetypes()
{
	if (std_resourcetypes_added) return;
	std_resourcetypes_added = true;
	new CL_Res_Surface;
	new CL_Res_Sample;
	new CL_Res_Font;
	new CL_Res_Palette;
	new CL_Res_Raw;
	new CL_Res_Integer;
	new CL_Res_String;
	new CL_Res_Boolean;
}

void CL_ResourceManager::init_read_from_datafile(CL_String _datafile_name)
{
	if (!ignore_datafile && data_in_datafile == false)
	{
		datafile_name = _datafile_name;
		data_in_datafile = true;
	}
}

CL_Resource *CL_ResourceManager::try_create_resource(CL_String name, CL_String full_name, CL_String res_id, CL_ResourceOptions *options)
{
	CL_Resource *ret = NULL;

	CL_Iterator<CL_ResourceType> counter(CL_ResourceType::resource_types);
	while (counter.next() != NULL)
	{
		if (data_in_datafile)
		{
			ret = counter()->try_create_from_datafile(name, full_name, res_id, options, this);
		}
		else
		{
			ret = counter()->try_create_from_file(name, full_name, res_id, options, this);
		}

		if (ret != NULL)
		{
			break;
		}
	}
	
	if (ret == NULL)
	{
		CL_String err("Failed to create resource: ");
		err << full_name;
		throw CL_Error(err);
	}
	return ret;
}

/******************************************************************************
							class CL_ResourceSection
******************************************************************************/

CL_ResourceManager::CL_ResourceSection::CL_ResourceSection(
	CL_String section_name, 
	CL_String prefix,
	CL_ResourceManager *_parent, 
	bool root_section) : CL_Resource("")
{
	name = section_name;
	if (prefix == "") full_name = name;
	else full_name = prefix + "/" + name;
//	cout << "Resource section: " << full_name << endl;
	parent = _parent;
	is_root_section = root_section;
}

CL_ResourceManager::CL_ResourceSection::~CL_ResourceSection()
{
	CL_Iterator<CL_Resource> counter(resources);
	while (counter.next() != NULL)
	{
		delete counter();
	}

	CL_Iterator<CL_ResourceSection> counter2(sub_sections);
	while (counter2.next() != NULL)
	{
		delete counter2();
	}
}

CL_Resource *CL_ResourceManager::CL_ResourceSection::get_resource(const char *id)
{
	CL_Iterator<CL_Resource> counter(resources);
	while (counter.next() != NULL)
	{
		if (counter()->get_name() == id) 
		{
			return counter();
		}
	}

	char *pos = strchr(id, '/');
	if (pos != NULL)
	{
		char temp[MAX_SECTION_LENGTH];
		memcpy(temp, id, pos-id);
		temp[pos-id] = 0;

		CL_Iterator<CL_ResourceSection> counter2(sub_sections);
		while (counter2.next() != NULL)
		{
			if (counter2()->get_name() == temp) 
			{
				return counter2()->get_resource(pos+1);
			}
		}
	}

	CL_String err("Couldn't get resource: ");
	err << id;
	throw CL_Error(err);
	return NULL;
}

CL_ResourceManager::CL_ResourceSection *CL_ResourceManager::CL_ResourceSection::get_section(const char *id)
{
	CL_Iterator<CL_ResourceSection> counter(sub_sections);
	while (counter.next() != NULL)
	{
		if (counter()->get_name() == id) return counter();
	}

	char *pos = strchr(id, '/');
	if (pos != NULL)
	{
		char temp[MAX_SECTION_LENGTH];
		memcpy(temp, id, pos-id);
		temp[pos-id] = 0;

		CL_Iterator<CL_ResourceSection> counter2(sub_sections);
		while (counter2.next() != NULL)
		{
			if (counter2()->get_name() == temp) 
			{
				return counter2()->get_section(pos+1);
			}
		}
	}

	CL_String err("Couldn't get section: ");
	err << id;
	throw CL_Error(err);
	return NULL; // MS Visual C++ need this line!
}

bool CL_ResourceManager::CL_ResourceSection::read_config()
{
	CL_String token = parent->get_next_token();
	do
	{
//		token.to_lower();
		// A new section
		if (token == "section")
		{
			CL_String section_name = parent->get_next_token();

			CL_String next = parent->get_next_token();
			if (next != "{")
			{
				parent->write_error("Missing '{'");
				return false;
			}
			CL_ResourceSection *new_section = new CL_ResourceSection(section_name, full_name, parent, false);
			sub_sections.add(new_section);
			if (!new_section->read_config()) return false;
		}
		// Is it the end of the section?
		else if (token == "}") 
		{
			break;
		}
		// Must be a resource then
		else
		{
			if (is_root_section)
			{
				if (token == "location")
				{
					CL_String next = parent->get_next_token();
					if (next != "=")
					{
						parent->write_error("Missing '=' following global 'location'-declaration");
						return false;
					}
					next = parent->get_next_token();
//					next.to_lower();
					bool syntax_ok = false;
					if (next == "datafile")
					{
						if (parent->get_next_token() == "(")
						{
							CL_String datafile_name = parent->get_next_token();
							parent->init_read_from_datafile(datafile_name);

							next = parent->get_next_token();
							if (next == ")" && parent->get_next_token() == ";")
							{
								syntax_ok = true;
							}
						}
					}
					if (!syntax_ok)
					{
						parent->write_error("Error encountered inside global location-declaration");
						return false;
					}
					token = parent->get_next_token();
					continue;
				}
			}

			CL_String name = token;
			token = parent->get_next_token();
//			cout << "Reading resource '" << name << "'" << endl;

			if (token != "=")
			{
				CL_String err;
				err << "Missing '=' following declaration of resource '" << name << "'";
				parent->write_error(err);
				return false;
			}
			CL_String res_id = parent->get_next_token();

			CL_ResourceOptions *options = new CL_ResourceOptions(parent);
			if (!options->read_options(name)) 
			{
				delete options;
				return false;
			}

			token = parent->get_next_token();
			if (token != ";") 
			{
				CL_String err;
				err << "Missing ';' following declaration of resource '" << name << "'";
				parent->write_error(err);
				delete options;
				return false;
			}

			CL_Resource *res = parent->try_create_resource(name, full_name, res_id, options);
			if (res != NULL)
			{
				resources.add(res);
			}
			else
			{
				CL_String err;
				err << "Unknown or no type declared for resource '" << name << "'";
				parent->write_error(err);
				delete options;
				return false;
			}
		}
		token = parent->get_next_token();
	} while (token != "");
	
	return true;
}

bool CL_ResourceManager::CL_ResourceSection::load()
{
	CL_Iterator<CL_Resource> counter(resources);
	while (counter.next() != NULL)
	{
		if (!counter()->load()) return false;
	}

	CL_Iterator<CL_ResourceSection> counter2(sub_sections);
	while (counter2.next() != NULL)
	{
		if (!counter2()->load()) return false;
	}

	return true;
}
