/*
	$Id: resource_manager.h,v 1.10 1999/08/22 17:18:47 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

//! l2_component="Resources"

#ifndef header_resource_manager
#define header_resource_manager

#include "../../clanlib.h"

#include "resource.h"
#include "resourcetype.h"
#include "resourceoptions.h"
#include "resourceoption.h"

#define default_scriptfile_id "_script_section"

class CL_ResourceOption;
class CL_ResourceOptions;
class CL_ResourceManager;

class CL_ResourceManager
//: The ClanLib resource manager.
// The resource manager is used to retrieve resources from a given resource source.
// This can either be a resource script file (used as input to the
// datafile compiler), or a datafile with all the resources included into
// one large gzipped resource file.
// <br>
// To speedup loading of resources in a game, you can load entire sections
// of resources at once. When a resource in the section is requested, it is
// returned instantly without having to access the disk. This is especially
// useful to make sure all the game resources are loaded before the game is
// started.
// <br>
// Resources are normally not retrieved using the get_resource() function.
// Instead, you should load the resource using the appropiate resource type
// class. For instance, a surface is easiest loaded like this:
// <code>
// CL_ResourceManager res_manager("my_datafile.dat", true);
// CL_Surface *my_surface = CL_Res_Surface::load("my_surface", res_manager);
// </code>
// <br>
// Getting the same resource twice won't create a new instance of the
// resource, but you will still have to clean up the resource after
// yourself. In other words: If you load a resource twice, you only have to
// delete it once.
{
	friend CL_InputSourceProvider;
public:
	CL_ResourceManager(const char *config_file,
		CL_InputSourceProvider *provider = NULL,
		bool read_directly_from_source=false,
		bool delete_inputsource_provider=false);
	//: Resource Manager constructor.
	//!param: config_file - the name of the file in which the resources are defined
	//!param: provider - the optional inputprovider in which, the resource file is stored
	//!param: read_directly_from_source - if true, any resources are loaded directly from their source, meaning that any datafile-directive is ignored.

	CL_ResourceManager(const char *config_id, const char *datafile_name);
	//: Resource Manager constructor. (obsolete)
	//!param: config_id - the datafile-id of the resource script-file
	//!param: datafile_name - the name of the datafile to open

	CL_ResourceManager( const char *file_name, const bool is_datafile );
	//: Resource Manager constructor.
	//!param: file_name - the name of the file to open
	//!param: is_datafile - indicates if the file is a scriptfile or a datafile

	virtual ~CL_ResourceManager();

	bool load_all_resources();
	//: Loads all resources globally in the config file into memory.
	//!retval: Returns false if any resources in the config file couldn't be loaded

	bool load_section(const char *section_name);
	//: Loads all resources in a given section into memory.
	//!param: section_name - the section to be loaded
	//!retval: Returns false if any resources in the section couldn't be loaded

	CL_Resource *get_resource(CL_String res_id);
	//: Returns a pointer to the CL_Resource representing the given resource
	//!param: Name of the resource in the configuration file - note: the name of resources are absolute, eg. including section names with slashes (Graphics/Pacman/sur_pacman)

	CL_List<CL_String> *get_all_resources();
	//: Returns a list of all resources available. Primarily used by the ClanCompiler to 
	//: build datafiles from resources.
	//!retval: The list of resources available. You'll have to delete all strings in the returned
	//!retval: list manually (including the list-instance itself).

	CL_List<CL_String> *get_resources_of_type(CL_String type_id);
	//: Returns a list of all resources available matching a given type. 
	//: Primarily used by the ClanCompiler to build datafiles from resources.
	//!retval: The list of resources available. You'll have to delete all strings in the 
	//!retval: list manually (including the list-instance itself).

	CL_InputSourceProvider *get_resource_provider();
	//: Returns a pointer to the inputsourceprovider, in which all resources are stored
	//: (this can be a fileprovider or a datafileprovider depending on the informations contained
	//: in the configuration file).
	//!retval: Pointer to inputsourceprovider containing resource data.

private:

	void init(
		const char *config_file,
		CL_InputSourceProvider *provider = NULL,
		bool read_directly_from_source=false,
		bool delete_inputsource_provider=false);

	static bool std_resourcetypes_added;
	void add_resourcetypes();

	class CL_ResourceSection : public CL_Resource
	{
		CL_List<CL_Resource> resources;
		CL_List<CL_ResourceSection> sub_sections;
		CL_ResourceManager *parent;
		bool is_root_section;

	public:
		CL_ResourceSection(CL_String section_name, CL_String prefix, CL_ResourceManager *parent, bool root_section);
		~CL_ResourceSection();

		CL_Resource *get_resource(const char *id);
		CL_ResourceSection *get_section(const char *id);

		CL_List<CL_Resource> *get_resources() { return &resources; }
		CL_List<CL_ResourceSection> *get_sections() { return &sub_sections; }

		bool read_config();
		bool load();
	};

	bool is_space(char c);
	bool is_operator(char c);
	bool is_remark(char c, int position);
	CL_String get_next_token();
	void write_error(CL_String err_msg);

	friend CL_Resource;
	friend CL_ResourceSection;
	friend CL_ResourceOptions;

	void get_section_resources(CL_String section_prefix, CL_ResourceSection *sect, CL_List<CL_String> *ret_list, const char *match_type);
	CL_Resource *try_create_resource(CL_String name, CL_String prefix, CL_String res_id, CL_ResourceOptions *options);

	CL_ResourceSection *root_section;
	CL_InputSourceProvider *resource_provider;

	void init_read_from_datafile(CL_String datafile_name);
	bool data_in_datafile;
	bool ignore_datafile;
	CL_String datafile_name;
		
	char *config_file_data;
	int total_size;
	int cur_pos;
	int cur_line;
	bool in_remark;

	CL_String config_handle;
};
#endif
