/*
	$Id: resourcetype_font.cpp,v 1.11 1999/07/12 22:54:44 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Font resource type

*/

#include "resourcetype_font.h"
#include "../Misc/font_description.h"
#include "../Misc/font.h"
#include "../SurfaceProviders/provider_targa.h"
#include "../SurfaceProviders/provider_pcx.h"

/******************************************************************************
						class CL_Res_Font
******************************************************************************/

CL_Res_Font::CL_Res_Font() : CL_ResourceType("font")
{
}

CL_Resource *CL_Res_Font::try_create_from_file(
	CL_String name,
	CL_String prefix,
	CL_String file_location,
	CL_ResourceOptions *options,
	CL_ResourceManager *parent)
{
	CL_String ext = file_location.right(4);
	ext.to_lower();

	bool is_font_type = false;
	if (options->exists("type")) 
	{
		if (options->get_value("type") != "font") return NULL;
		is_font_type = true;
	}

	if (is_font_type ||
		options->exists("font") || ext == ".pcx" || ext == ".tga")
	{
		return new CL_Font_Resource(
			name,
			prefix,
			file_location,
			options,
			parent,
			false);
	}

	return NULL;
}

CL_Resource *CL_Res_Font::try_create_from_datafile(
	CL_String name,
	CL_String prefix,
	CL_String file_location,
	CL_ResourceOptions *options,
	CL_ResourceManager *parent)
{
	// this function isn't needed -- really!
/*
	return try_create_from_file(
		name,
		prefix,
		file_location,
		options,
		parent);
*/

	// hmm, maybe it is afterall,
	// but thats only because CL_ResourceManager is crap!

	CL_String ext = file_location.right(4);
	ext.to_lower();

	bool is_font_type = false;
	if (options->exists("type")) 
	{
		if (options->get_value("type") != "font") return NULL;
		is_font_type = true;
	}
	
	if (is_font_type ||
		options->exists("font") || ext == ".pcx" || ext == ".tga")
	{
		return new CL_Font_Resource(
			name,
			prefix,
			file_location,
			options,
			parent,
			true);
	}

	return NULL;
}

CL_Font *CL_Res_Font::load(
	CL_String res_id,
	CL_ResourceManager &manager)
{
	CL_Resource *res = manager.get_resource(res_id);
	cl_assert(res != NULL);

	if (res->get_type() != "font")
		throw CL_Error("Resource is not a font!!!");

	if (res->load())
	{
		CL_Font_Resource *res2 = (CL_Font_Resource *) res;
		cl_assert(res2->get_font_desc() != NULL);

		return new CL_Font(res2->get_font_desc());
	}
	else
	{
		throw CL_Error("Failed to load font!!!");
		return NULL; // Visual C++ need this line!
	}
}

/******************************************************************************
						class CL_Font_Resource
******************************************************************************/

CL_Font_Resource::CL_Font_Resource(
	CL_String name,
	CL_String prefix,
	CL_String location,
	CL_ResourceOptions *options,
	CL_ResourceManager *parent,
	bool from_datafile)
: CL_Resource("font") 
{
	this->name = name;
	this->prefix = prefix;
	this->location = location;
	this->options = options;
	this->parent = parent;
	this->from_datafile = from_datafile;

	font_desc = NULL;
}

void CL_Font_Resource::write(CL_OutputSource *output)
{
	load();

	output->write_int32(font_desc->space_len);
	output->write_int32(font_desc->subtract_width);
	output->write_short16(strlen(font_desc->letters)+1);
	output->write(font_desc->letters, strlen(font_desc->letters)+1);
	
	int len = strlen(font_desc->letters);
	if (len > font_desc->letter_providers.get_num_items())
	{
		CL_String err("Font loading error - ");
		err << "Letters expected:  " << len;
		err << ", letters available: " << font_desc->letter_providers.get_num_items();
		throw CL_Error(err);
	}
	else if (len < font_desc->letter_providers.get_num_items())
	{
		cout << "ClanLib Font Warning: Font letters do not match!" << endl;
		cout << "ClanLib Font Warning: Letters expected:  " << len << endl;
		cout << "ClanLib Font Warning: Letters available: "
			<< font_desc->letter_providers.get_num_items() << endl;
	}
	
	if (len > 0)
	{
		font_desc->letter_providers[0]->lock();
	}
	for(int i=0; i<len; i++)
	{
		CL_SurfaceProvider *provider = font_desc->letter_providers[i];
		cl_assert(provider != NULL);

		provider->lock();

		int width = provider->get_width();
		int height = provider->get_height();
		int no_sprs = provider->get_no_sprs();
		int bytes_per_pixel = provider->get_bytes_pr_pixel();
		int transcol = provider->get_transcol();
		int format = provider->get_pixel_format();
		int translate_x = provider->get_translate_x();
		int translate_y = provider->get_translate_y();
		int pitch = provider->get_pitch();

		CL_Palette *pal = provider->get_palette();
		unsigned char *data = (unsigned char *) provider->get_surface_data();
		cl_assert(data != NULL);

		output->write_int32(width);
		output->write_int32(height);
		output->write_int32(no_sprs);
		output->write_int32(transcol);
		
		char palette_bool = (pal != NULL) ? 1 : 0;
		output->write_char8(palette_bool);
		
		if (palette_bool) output->write(pal->palette, 256*3);
		
		output->write_int32(format);

		if (transcol == -1)
		{
			int bytes_pr_line = width*bytes_per_pixel;
			for (int y=0; y<height*no_sprs; y++)
			{
				int img_y = y+translate_y;

				output->write(
					data+translate_x*bytes_per_pixel+img_y*pitch,
					bytes_pr_line);
			}
		}
		else
		{
			for (int y=0; y<height*no_sprs; y++)
			{
				int img_y = y+translate_y;

				output->write(
					data+translate_x*bytes_per_pixel+img_y*pitch,
					bytes_per_pixel*width);
			}
		}

		provider->unlock();
	}
	if (len > 0)
	{
		font_desc->letter_providers[0]->unlock();
	}
}

bool CL_Font_Resource::load()
{
	if (font_desc != NULL) return true; // already loaded.

	if (from_datafile) load_from_datafile();
	else load_from_file();

	if (font_desc == NULL) return false;
	return true;
}

void CL_Font_Resource::load_from_datafile()
{
	CL_InputSourceProvider *provider = parent->get_resource_provider();

	CL_String fullname = prefix;
	if (fullname != "") fullname << "/";
	fullname << get_name();

	font_desc = new CL_Font_Description(
		provider->open_source(fullname));
}

void CL_Font_Resource::load_from_file()
{
	CL_String ext = location.right(4);
	ext.to_lower();
	
	if (ext == ".tga" || options->exists("alphafont"))
	{
		if (options->exists("spacelen") == false)
			throw CL_Error("Font missing spacelen option.");

		if (options->exists("trans_limit") == false)
			throw CL_Error("Font missing trans_limit option.");

		if (options->exists("subtract_width") == false)
			throw CL_Error("Font missing subtract_width option.");

		if (options->exists("letters") == false)
			throw CL_Error("Font missing letters option.");

		int space_len = CL_String(options->get_value("spacelen")).get_as_int();
		float trans_limit = CL_String(options->get_value("trans_limit")).get_as_float();
		int subtract_width = CL_String(options->get_value("subtract_width")).get_as_int();
	
		font_desc = new CL_Font_Description(
			new CL_TargaProvider(location, NULL),
			trans_limit,
			space_len,
			subtract_width,
			options->get_value("letters"));
	}
	else
	{
		if (options->exists("x") == false)
			throw CL_Error("Font missing x coordinate option.");

		if (options->exists("y") == false)
			throw CL_Error("Font missing y coordinate option.");

		if (options->exists("spacelen") == false)
			throw CL_Error("Font missing spacelen option.");

		if (options->exists("letters") == false)
			throw CL_Error("Font missing letters option.");

		int x = CL_String(options->get_value("x")).get_as_int();
		int y = CL_String(options->get_value("y")).get_as_int();
		int space_len = CL_String(options->get_value("spacelen")).get_as_int();
		int subtract_width = 0;
		if (options->exists("subtract_width"))
			subtract_width = CL_String(
				options->get_value("subtract_width")).get_as_int();

		int num_tcols = 1;
		int *tcols = NULL;
		if (num_tcols > 0)
		{
			tcols = new int[num_tcols];
			tcols[0] = 0;
		}

		font_desc = new CL_Font_Description(
			new CL_PCXProvider(location, NULL),
			x, y,
			tcols, num_tcols,
			space_len,
			subtract_width,
			options->get_value("letters"));
	}
}
