/*
	$Id: resourcetype_palette.cpp,v 1.7 1999/06/03 22:57:10 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Palette resource type

*/

#include "../../clanlib.h"
#include "../SurfaceProviders/provider_pcx.h"
#include "resourcetype_palette.h"

/******************************************************************************
						class CL_Res_Palette
******************************************************************************/

CL_Res_Palette::CL_Res_Palette() : CL_ResourceType("palette")
{
}

CL_Resource *CL_Res_Palette::try_create_from_file(CL_String name, CL_String prefix, CL_String res_id, CL_ResourceOptions *options, CL_ResourceManager *parent)
{
	CL_String ext = res_id.right(4);
	ext.to_lower();

	bool is_palette_type = false;
	if (options->exists("type")) 
	{
		if (options->get_value("type") != "palette") return NULL;
		is_palette_type = true;
	}
	
	if (is_palette_type ||
		options->exists("tga") || ext == ".tga" ||
		options->exists("pcx") || ext == ".pcx")
	{
		return new CL_PaletteFileResource(name, prefix, res_id, options, parent);
	}

	return NULL;
}

CL_Resource *CL_Res_Palette::try_create_from_datafile(CL_String name, CL_String prefix, CL_String res_id, CL_ResourceOptions *options, CL_ResourceManager *parent)
{
	CL_String ext = res_id.right(4);
	ext.to_lower();

	bool is_palette_type = false;
	if (options->exists("type")) 
	{
		if (options->get_value("type") != "palette") return NULL;
		is_palette_type = true;
	}
	
	if (is_palette_type ||
		options->exists("tga") || ext == ".tga" ||
		options->exists("pcx") || ext == ".pcx")
	{
		return new CL_PaletteDatafileResource(name, prefix, res_id, options, parent);
	}

	return NULL;
}

CL_Palette *CL_Res_Palette::load(CL_String res_id, CL_ResourceManager &manager)
{
	CL_Resource *res = manager.get_resource(res_id);
	cl_assert(res != NULL);
//	cl_assert(res->get_type() == "palette");
	if (res->get_type() != "palette")
		throw CL_Error("resource is not a palette!!!");

	return ((CL_WritablePaletteResource *) res)->create_palette();
}

void CL_Res_Palette::set_system_palette(CL_String res_id, CL_ResourceManager &manager)
{
	CL_Palette *pal = load(res_id, manager);
	CL_Display::set_palette(pal);
	delete pal;
}

/******************************************************************************
						class CL_WritablePaletteResource
******************************************************************************/

CL_WritablePaletteResource::CL_WritablePaletteResource() : CL_Resource("palette") 
{
}

void CL_WritablePaletteResource::write(CL_OutputSource *output)
{
	CL_Palette *pal = create_palette();
	cl_assert(pal != NULL);

	unsigned char *pal_data = pal->palette;

	output->write_int32(pal->num_colors);
	output->write(pal_data, pal->num_colors*3);

	delete pal;
}

bool CL_WritablePaletteResource::load()
{
	if (loaded) return true;

	loaded_palette = create_palette();
	if (loaded_palette == NULL) return false;

	loaded = true;

	return true;
}

/******************************************************************************
						class CL_PaletteFileResource
******************************************************************************/

CL_PaletteFileResource::CL_PaletteFileResource(CL_String _palette_id, CL_String prefix, CL_String _palette_location, CL_ResourceOptions *_options, CL_ResourceManager *_parent)
{
	name = _palette_id;
	if (prefix == "") full_name = name; else full_name = prefix + "/" + name;
	palette_location = _palette_location;
	options = _options;
	parent = _parent;
}

CL_Palette *CL_PaletteFileResource::create_palette()
{
	CL_String ext = palette_location.right(4);
	ext.to_lower();

	CL_SurfaceProvider *p = new CL_PCXProvider(
		palette_location,
		parent->get_resource_provider());

	cl_assert(p != NULL);

	p->lock();

	cl_assert(p->get_palette() != NULL);
	CL_Palette *ret = new CL_Palette(
		p->get_palette()->palette,
		p->get_palette()->num_colors);

	p->unlock();
	delete p;

	return ret;
}

/******************************************************************************
						class CL_PaletteDatafileResource
******************************************************************************/

CL_PaletteDatafileResource::CL_PaletteDatafileResource(
	CL_String _surface_id,
	CL_String prefix,
	CL_String _palette_location,
	CL_ResourceOptions *_options,
	CL_ResourceManager *_parent)
{
	name = _surface_id;
	if (prefix == "") full_name = name; else full_name = prefix + "/" + name;
	palette_location = _palette_location;
	options = _options;
	parent = _parent;
}

CL_Palette *CL_PaletteDatafileResource::create_palette()
{
	if (parent->get_resource_provider() == NULL) return NULL;

	CL_InputSource *input = parent->get_resource_provider()->open_source(full_name);
	if (input == NULL) return NULL;

	int num_colors = input->read_int32();
	CL_Palette *pal = new CL_Palette;
	pal->palette = new unsigned char[num_colors*3];
	pal->num_colors = num_colors;
	input->read(pal->palette, num_colors*3);

	delete input;

	return pal;
}

