/*
	$Id: resourcetype_sample.cpp,v 1.3 1999/01/20 02:17:28 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Sample resource type

*/

#include "clanlib.h"
#include "../StaticSoundProviders/wave_sample.h"
#include "../StreamSoundProviders/streamed_wave_sample.h"
#include "resourcetype_sample.h"

/******************************************************************************
						class CL_Res_Sample
******************************************************************************/

CL_SoundBuffer *CL_Res_Sample::load(CL_String res_id, CL_ResourceManager &manager)
{
	CL_Resource *res = manager.get_resource(res_id);
	if (res == NULL) return NULL;

	if (res->get_type() == "sample")
	{
		return ((CL_WritableSampleResource *) res)->create_sample();
	}
	return NULL;
}

CL_Res_Sample::CL_Res_Sample() : CL_ResourceType("sample")
{
}

CL_Resource *CL_Res_Sample::try_create_from_file(CL_String name, CL_String prefix, CL_String res_id, CL_ResourceOptions *options, CL_ResourceManager *parent)
{
	CL_String ext = res_id.right(4);
	ext.to_lower();

	if ((options->exists("type") && options->get_value("type") == "sample") ||
		(options->exists("wav") || ext == ".wav"))
	{
		return new CL_SampleFileResource(name, prefix, res_id, options, parent);
	}

	return NULL;
}

CL_Resource *CL_Res_Sample::try_create_from_datafile(CL_String name, CL_String prefix, CL_String res_id, CL_ResourceOptions *options, CL_ResourceManager *parent)
{
	CL_String ext = res_id.right(4);
	ext.to_lower();

	if ((options->exists("type") && options->get_value("type") == "sample") ||
		(options->exists("wav") || ext == ".wav"))
	{
		return new CL_SampleDatafileResource(name, prefix, res_id, options, parent);
	}

	return NULL;
}

/******************************************************************************
						class CL_WritableSampleResource
******************************************************************************/

CL_WritableSampleResource::CL_WritableSampleResource() : CL_Resource("sample") 
{ 
	loaded = false;
	loaded_sample = NULL;
}

void CL_WritableSampleResource::write(CL_OutputSource *output)
{
	CL_InputSourceProvider *provider = CL_InputSourceProvider::create_file_provider(".");
	cl_assert(provider != NULL);

	CL_InputSource *wavefile = provider->open_source(sample_location);
	cl_assert(wavefile != NULL);

	int size_left = wavefile->size();
	while (size_left > 0)
	{
		unsigned char buffer[32768];
		int read = wavefile->read(buffer, 32768);
		size_left -= read;
		output->write(buffer, read);
	}

	delete wavefile;
	delete provider;
}

bool CL_WritableSampleResource::load()
{
	if (loaded) return true;

	if (options->exists("stream"))
	{
		CL_StreamSoundProvider *provider = create_stream_provider();
		loaded_sample = CL_SoundBuffer::create(provider, true);
		if (loaded_sample == NULL) 
		{
			delete provider;
			return false;
		}
	}
	else
	{
		CL_StaticSoundProvider *provider = create_static_provider();
		loaded_sample = CL_SoundBuffer::create(provider, true);
		if (loaded_sample == NULL) 
		{
			delete provider;
			return false;
		}
	}

	loaded = true;
	return true;
}

CL_SoundBuffer *CL_WritableSampleResource::create_sample()
{
	if (!loaded)
	{
		if (!load()) return NULL;
	}

	return loaded_sample;
}

/******************************************************************************
						class CL_SampleFileResource
******************************************************************************/

CL_SampleFileResource::CL_SampleFileResource(
			CL_String _surface_id, CL_String prefix, 
			CL_String _sample_location, 
			CL_ResourceOptions *_options, 
			CL_ResourceManager *_parent)
{
	name = _surface_id;
	if (prefix == "") full_name = name; else full_name = prefix + "/" + name;
	sample_location = _sample_location;
	options = _options;
	parent = _parent;
}

CL_SampleFileResource::~CL_SampleFileResource()
{
	delete options;
}

CL_StaticSoundProvider *CL_SampleFileResource::create_static_provider()
{
	return new CL_Sample(sample_location, parent->get_resource_provider());
}

CL_StreamSoundProvider *CL_SampleFileResource::create_stream_provider()
{
	return new CL_Streamed_WaveSample(sample_location, parent->get_resource_provider(), options->exists("loop"));
}

/******************************************************************************
						class CL_SampleDatafileResource
******************************************************************************/

CL_SampleDatafileResource::CL_SampleDatafileResource(
			CL_String _surface_id, CL_String prefix, 
			CL_String _sample_location, CL_ResourceOptions *_options, 
			CL_ResourceManager *_parent)
{
	name = _surface_id;
	if (prefix == "") full_name = name; else full_name = prefix + "/" + name;
	sample_location = _sample_location;
	options = _options;
	parent = _parent;
}

CL_SampleDatafileResource::~CL_SampleDatafileResource()
{
	delete options;
}

CL_StaticSoundProvider *CL_SampleDatafileResource::create_static_provider()
{
	return new CL_Sample(full_name, parent->get_resource_provider());
}

CL_StreamSoundProvider *CL_SampleDatafileResource::create_stream_provider()
{
	return new CL_Streamed_WaveSample(full_name, parent->get_resource_provider(), options->exists("loop"));
}
