/*
	$Id: resourcetype_surface.cpp,v 1.5 1999/06/03 22:02:20 starch Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Surface resource type

*/

#include "../SurfaceProviders/sprite.h"
#include "../SurfaceProviders/provider_targa.h"
#include "../SurfaceProviders/provider_pcx.h"
#include "../SurfaceProviders/sprite_subarray_provider.h"
#include "../SurfaceProviders/sprite_subsection_provider.h"
#include "../SurfaceProviders/provider_png.h"
#include "resourcetype_surface.h"

/******************************************************************************
						class CL_ResourceType_Surface
******************************************************************************/

CL_Surface *CL_Res_Surface::load(CL_String res_id, CL_ResourceManager &manager)
{
	CL_Resource *res = manager.get_resource(res_id);
	if (res == NULL) return NULL;

	if (res->get_type() == "surface")
	{
		return ((CL_WritableSurfaceResource *) res)->create_surface();
	}
	return NULL;
}

CL_Res_Surface::CL_Res_Surface() : CL_ResourceType("surface")
{
}

CL_Resource *CL_Res_Surface::try_create_from_file(CL_String name, CL_String prefix, CL_String res_id, CL_ResourceOptions *options, CL_ResourceManager *parent)
{
	CL_String ext = res_id.right(4);
	ext.to_lower();

	bool is_surface_type = false;
	if (options->exists("type")) 
	{
		if (options->get_value("type") != "surface") return NULL;
		is_surface_type = true;
	}
	
	if (is_surface_type ||
		options->exists("tga") || ext == ".tga" ||
		options->exists("pcx") || ext == ".pcx" ||
		options->exists("png") || ext == ".png")
	{
		return new CL_SurfaceFileResource(name, prefix, res_id, options, parent);
	}

	return NULL;
}

CL_Resource *CL_Res_Surface::try_create_from_datafile(CL_String name, CL_String prefix, CL_String res_id, CL_ResourceOptions *options, CL_ResourceManager *parent)
{
	CL_String ext = res_id.right(4);
	ext.to_lower();

	bool is_surface_type = false;
	if (options->exists("type")) 
	{
		if (options->get_value("type") != "surface") return NULL;
		is_surface_type = true;
	}
	
	if (is_surface_type ||
		options->exists("tga") || ext == ".tga" ||
		options->exists("pcx") || ext == ".pcx" ||
		options->exists("png") || ext == ".png")
	{
		return new CL_SurfaceDatafileResource(name, prefix, res_id, options, parent);
	}

	return NULL;
}

/******************************************************************************
						class CL_WritableSurfaceResource
******************************************************************************/

CL_WritableSurfaceResource::CL_WritableSurfaceResource() : CL_Resource("surface") 
{ 
	loaded = false;
	loaded_surface = NULL;
}

void CL_WritableSurfaceResource::write(CL_OutputSource *output)
{
	CL_SurfaceProvider *provider = create_provider();
	cl_assert(provider != NULL);

	provider->lock();

	int width = provider->get_width();
	int height = provider->get_height();
	int no_sprs = provider->get_no_sprs();
	int bytes_per_pixel = provider->get_bytes_pr_pixel();
	int transcol = provider->get_transcol();
	int format = provider->get_pixel_format();
	int translate_x = provider->get_translate_x();
	int translate_y = provider->get_translate_y();
	int pitch = provider->get_pitch();

	CL_Palette *pal = provider->get_palette();
	unsigned char *data = (unsigned char *) provider->get_surface_data();
	cl_assert(data != NULL);

	output->write_int32(width);
	output->write_int32(height);
	output->write_int32(no_sprs);
	output->write_int32(transcol);
	
	char palette_bool = (pal != NULL) ? 1 : 0;
	output->write_char8(palette_bool);
	
	if (palette_bool) output->write(pal->palette, 256*3);
	
	output->write_int32(format);

	if (transcol == -1)
	{
		int bytes_pr_line = width*bytes_per_pixel;
		for (int y=0; y<height*no_sprs; y++)
		{
			int img_y = y+translate_y;

			output->write(
				data+translate_x*bytes_per_pixel+img_y*pitch,
				bytes_pr_line);
		}
	}
	else
	{
		for (int y=0; y<height*no_sprs; y++)
		{
			int img_y = y+translate_y;

			output->write(data+translate_x*bytes_per_pixel+img_y*pitch, bytes_per_pixel*width);
		}
	}

	provider->unlock();

	delete provider;
}

bool CL_WritableSurfaceResource::load()
{
	if (loaded) return true;

	CL_SurfaceProvider *provider = create_provider();
	if (provider == NULL) return false;

	loaded_surface = CL_Surface::create(provider, true);
	if (loaded_surface == NULL) 
	{
		delete provider;
		return false;
	}

	loaded = true;
	return true;
}

CL_Surface *CL_WritableSurfaceResource::create_surface()
{
	if (!loaded)
	{
		if (!load()) return NULL;
	}

	return loaded_surface;
}

/******************************************************************************
						class CL_SurfaceFileResource
******************************************************************************/

CL_SurfaceFileResource::CL_SurfaceFileResource(
	CL_String _surface_id, 
	CL_String prefix, 
	CL_String _surface_location, 
	CL_ResourceOptions *_options, 
	CL_ResourceManager *_parent)
{
	name = _surface_id;
	if (prefix == "") full_name = name; else full_name = prefix + "/" + name;
	surface_location = _surface_location;
	options = _options;
	parent = _parent;
}

CL_SurfaceFileResource::~CL_SurfaceFileResource()
{
	delete options;
}

CL_SurfaceProvider *CL_SurfaceFileResource::create_provider()
{
	CL_SurfaceProvider *ret_provider = NULL;

	CL_String ext = surface_location.right(4);
	ext.to_lower();

	if (options->exists("tga") ||  ext == ".tga")
	{
		ret_provider = new CL_TargaProvider(surface_location, NULL);
	}
	else if (options->exists("pcx") || ext == ".pcx")
	{
		ret_provider = new CL_PCXProvider(surface_location, NULL);
	}
#ifdef USE_LIBPNG
	else if (options->exists("png") || ext == ".png")
	{
		ret_provider = new CL_PNGProvider(surface_location, NULL);
	}
#endif
	else
	{
		return NULL;
	}

	if (options->exists("x") && options->exists("y") && 
		options->exists("width") && options->exists("height"))
	{
		if (options->exists("array"))
		{
			int sep = options->get_value("array").find('x');
			if (sep == -1) return NULL;

			ret_provider = new CL_SpriteSubarrayProvider(
				ret_provider,
				options->get_value("x").get_as_int(),
				options->get_value("y").get_as_int(),
				options->get_value("width").get_as_int(),
				options->get_value("height").get_as_int(),
				options->get_value("array").mid(0, sep).get_as_int(),
				options->get_value("array").mid(sep+1).get_as_int());
		}
		else
		{
			ret_provider = new CL_SpriteSubsectionProvider(
				ret_provider,
				options->get_value("x").get_as_int(),
				options->get_value("y").get_as_int(),
				options->get_value("width").get_as_int(),
				options->get_value("height").get_as_int());
		}
	}

	// Explicitly specified transparency colors are _only_ supported in 8 bpp
	//
	if (ret_provider->get_bytes_pr_pixel() == 1)
	{
		CL_ResourceOption *tcol_option = options->get_option("tcol");
		if (tcol_option != NULL)
		{
			if (tcol_option->multi_valued())
			{
				CL_List<CL_String> *tcol_list = tcol_option->get_all_values();
				unsigned char *trans_table = new unsigned char[tcol_list->get_num_items()];

				int pos = 0;
				CL_Iterator<CL_String> counter(tcol_list);
				while (counter.next() != NULL) 
				{
					int tval = counter()->get_as_int();
					if (tval >= 0 && tval < 256) 
					{
						trans_table[pos++] = tval;
					}
				}

				ret_provider = new CL_MaskTranscolProvider(
						ret_provider, 
						true, 
						trans_table, 
						tcol_list->get_num_items());
				delete[] trans_table;
			}
			else
			{
				CL_String tcol = tcol_option->get_value();
				int tval = tcol.get_as_int();
				if (tval >= 0 && tval < 256) 
				{
					unsigned char v = tval;
					ret_provider = new CL_MaskTranscolProvider(
							ret_provider, 
							true, 
							&v, 
							1);
				}
			}
		}
	}

	return ret_provider;
}

/******************************************************************************
						class CL_SurfaceDatafileResource
******************************************************************************/

CL_SurfaceDatafileResource::CL_SurfaceDatafileResource(
	CL_String _surface_id, 
	CL_String prefix, 
	CL_String _surface_location, 
	CL_ResourceOptions *_options, 
	CL_ResourceManager *_parent)
{
	name = _surface_id;
	if (prefix == "") full_name = name; else full_name = prefix + "/" + name;
	surface_location = _surface_location;
	options = _options;
	parent = _parent;
}

CL_SurfaceDatafileResource::~CL_SurfaceDatafileResource()
{
	delete options;
}

CL_SurfaceProvider *CL_SurfaceDatafileResource::create_provider()
{
	return new CL_SpriteProvider(full_name, parent->get_resource_provider());
}

/******************************************************************************
						class CL_MaskTranscolProvider
******************************************************************************/

CL_MaskTranscolProvider::CL_MaskTranscolProvider(
	CL_SurfaceProvider *_parent, 
	bool _delete_provider, 
	const unsigned char * _mask_indices,
	int _num_mask_indices)
{
	parent = _parent;
	cl_assert(parent != NULL);
	cl_assert(parent->get_pixel_format() == PAL8);

	delete_provider = _delete_provider;
	num_mask_indices = _num_mask_indices;
	mask_indices = new unsigned char[num_mask_indices];
	memcpy(mask_indices, _mask_indices, num_mask_indices);
}

CL_MaskTranscolProvider::~CL_MaskTranscolProvider()
{
	if (delete_provider) delete parent;
	delete[] mask_indices;
}

int CL_MaskTranscolProvider::get_pitch() const
{
	return parent->get_pitch();
}

int CL_MaskTranscolProvider::get_translate_x() const
{
	return parent->get_translate_x();
}

int CL_MaskTranscolProvider::get_translate_y() const
{
	return parent->get_translate_y();
}

int CL_MaskTranscolProvider::get_width() const 
{
	return parent->get_width();
}

int CL_MaskTranscolProvider::get_height() const
{
	return parent->get_height();
}

int CL_MaskTranscolProvider::get_no_sprs() const
{
	return parent->get_no_sprs();
}

EPixelFormat CL_MaskTranscolProvider::get_pixel_format() const
{
	return parent->get_pixel_format();
}

CL_Palette *CL_MaskTranscolProvider::get_palette() const
{
	return parent->get_palette();
}

int CL_MaskTranscolProvider::get_transcol() const
{
	return mask_indices[0];
}

void *CL_MaskTranscolProvider::get_surface_data() const
{
	return parent->get_surface_data();
}

void CL_MaskTranscolProvider::perform_lock()
{
	parent->lock();

	int i;
	bool trans_table[256];
	for (i=0;i<256;i++) trans_table[i] = false;
	if (parent->get_transcol() != -1) trans_table[parent->get_transcol()] = true;
	for (i=0;i<num_mask_indices;i++) trans_table[mask_indices[i]] = true;

	unsigned char *pdata = (unsigned char *) parent->get_surface_data();
	for (int s=0;s<parent->get_no_sprs();s++)
	{
		for (int y=0;y<parent->get_height();y++)
		{
			for (int x=0;x<parent->get_width();x++)
			{
				if (trans_table[*pdata]) *pdata = mask_indices[0];
				pdata++;
			}
			pdata += parent->get_pitch()-parent->get_width();
		}
	}
}

void CL_MaskTranscolProvider::perform_unlock()
{
	parent->unlock();
}
