/*

	------------------------------------------------------------------------
	PNG-SurfaceProvider Class
	written by Karsten-Olaf Laux

	in order to works with
	ClanLib, the platform independent game SDK.

	This file is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		PNG (.png) surface provider.
*/

#ifdef USE_LIBPNG

//! l2_component="SurfaceProviders"

#ifndef header_PNGprovider
#define header_PNGprovider

#include "../../clanlib.h"
#include "generic_surfaceprovider.h"
#include <png.h>

class CL_PNGProvider : public CL_SurfaceProvider_Generic
//: Surface provider that can load PNG (.png) files.

{
	CL_String filename;
	bool locked;

	unsigned char *image;

	unsigned short map_length;
	unsigned char *color_map;

	unsigned char datatype;

	int pitch, height, no_sprs;
	int width;

	int bpp;

	bool transparent, ignore_alphachannel, use_alphapixels;

	int trans_num;
	int trans_col;
	unsigned char trans_redcol, trans_greencol, trans_bluecol;

	int pos;

	void read_data();

	void read_header();



	CL_InputSourceProvider *provider;
	CL_InputSource *input_source;

	//PNGlib stuff:
	png_structp png_ptr;
	png_infop info_ptr;
	png_infop end_info;
public:

	CL_InputSource* get_input_source() {return input_source; };
	//:returns Pointer to CL_InputSource
	//:(used by libpng-callback CL_PNGProvider::pngread_file() )


	static CL_Surface *create(CL_String handle, 
				  CL_InputSourceProvider *provider=NULL, 
				  bool transparent=true,
				  bool ignore_alphachannel=false);

	//: Loads the PNG file 'handle' from the inputsource provider 'provider. 
	//: Creates a CL_Surface using the PNG image and returns it.

	CL_PNGProvider(CL_String name, 
			 CL_InputSourceProvider *provider = NULL,
			 bool transparent=true,
			 bool ignore_alphachannel=false);
	//: Constructs a surface provider that can read PNG files.
	//!param: name - Name of the PNG file to load.
	//!param: provider - Input source provider that delivers the PNG file.
	//!param: transparent - True if a transparency color should be used.
	//!param: ignore_alphachannel - True if the alpha component should be ignored.
	//!param: trans_red - Red component of the transparency color.
	//!param: trans_green - Green component of the transparency color.
	//!param: trans_blue - blue component of the transparency color.

	virtual ~CL_PNGProvider();

	virtual int get_pitch() const { return pitch; }
	//: Returns the pitch of the image (bytes per line).

	virtual int get_width() const { return width; }
	//: Returns the width of the image.

	virtual int get_height() const { return height; }
	//: Returns the height of the image.
	
	virtual int get_no_sprs() const { return no_sprs; }
	//: Returns the number of subsprites in the image.

	virtual EPixelFormat get_pixel_format() const { return RGBA8888; }
	//: Returns the pixelformat used by the image.

	virtual CL_Palette *get_palette() const { return NULL; }
	//: Returns the palette used by the image. NULL if system palette.

	virtual int get_transcol() const;
	//: Returns the transparency color used, -1 if none.

	virtual void *get_surface_data() const;
	//: Returns the image data. Provider must be locked before pointer is valid.

	virtual void perform_lock();
	//: Locks the surface provider.

	virtual void perform_unlock();
	//: Unlocks the surface provider.
	
	static void pngread_file(png_structp png_ptr,
				 png_bytep data, 
				 png_size_t length)
	//: Callback used by libpng to retrieve the filedata. 
	//: (calls get_input_source()->read_uchar8())
	  {
	    //since this method is static, we need to know who we are ...
	    CL_PNGProvider *instance =  (CL_PNGProvider *)png_get_io_ptr(png_ptr);
	    //no error-checking here ....
	    cerr << ".";
	    png_size_t i;
	    for(i=0; i< length; i++)
	      data[i]=instance->get_input_source()->read_uchar8();
	  }

};

#endif

#endif /* USE_LIBPNG */
