/* $Id: visual.c,v 1.29 1999/03/05 18:34:13 marcus Exp $
******************************************************************************

   SVGAlib target: initialization

   Copyright (C) 1998 Steve Cheng   [steve@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <ggi/internal/ggi-dl.h>
#include <ggi/display/svgalib.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

static int usagecounter = 0;

/* SVGAlib's signal handlers */
static int vga_signals[] = {
	SIGHUP, SIGINT, SIGQUIT, SIGILL,
	SIGTRAP, SIGIOT, SIGBUS, SIGFPE,
	SIGSEGV, SIGPIPE, SIGALRM, SIGTERM,
	SIGXCPU, SIGXFSZ, SIGVTALRM,
	SIGPROF, SIGPWR};
static struct sigaction old_signals[sizeof(vga_signals)];


void _GGI_svga_freedbs(ggi_visual *vis) {
	int i;

	for (i=LIBGGI_APPLIST(vis)->num-1; i >= 0; i--) {
		_ggi_db_free(LIBGGI_APPBUFS(vis)[i]);
		_ggi_db_del_buffer(LIBGGI_APPLIST(vis), i);
	}
}


static void _GGIexitfunc(void)
{
	vga_setmode(TEXT);
}

static void _GGIchecksvgamodes(ggi_visual *vis)
{
	int modes=0;
	int i;

	SVGA_PRIV(vis)->availmodes =
		_ggi_malloc(vga_lastmodenumber()*sizeof(ggi_modelistmode));

	for (i=1; i <= vga_lastmodenumber(); i++) {
		if (vga_hasmode(i)) {
			vga_modeinfo *modeinfo;
			ggi_modelistmode *sgmode;
			int bpp, size;
			modeinfo = vga_getmodeinfo(i);
			switch(modeinfo->colors) {
			case 2:
				continue; /* 1 bpp doesn't work */
				bpp = size = 1;
				break;
			case 16:
				continue; /* 4 bpp doesn't work */
				bpp = size = 4;
				break;
			case 256:
				bpp = size = 8;
				break;
			case 1<<15:
				bpp = size = 15;
				break;
			case 1<<16:
				bpp = size = 16;
				break;
			case 1<<24:
				bpp = 24;
				if (modeinfo->bytesperpixel==3) {
					size = 24;
				} else {
					size = 32;
				}
				break;
			default:
				continue;
			}
			sgmode = SVGA_PRIV(vis)->availmodes + modes;
			modes++;
			sgmode->x = modeinfo->width;
			sgmode->y = modeinfo->height;
			sgmode->bpp = bpp;
			sgmode->gt = GT_CONSTRUCT(bpp, (bpp <= 8) ?
			      GT_PALETTE : GT_TRUECOLOR, size);
			if ((modeinfo->flags & IS_MODEX))
				sgmode->ismodex = 1;
			else
				sgmode->ismodex = 0;
		}
	}

	if (modes==0) {
		free(SVGA_PRIV(vis)->availmodes);
		SVGA_PRIV(vis)->availmodes=NULL;
	} else {
		SVGA_PRIV(vis)->availmodes =
			_ggi_realloc(SVGA_PRIV(vis)->availmodes,
				     (modes+1)*sizeof(ggi_modelistmode));
		SVGA_PRIV(vis)->availmodes[modes].bpp=0;
	}
}

/* This is exported from svgalib, we use it to check if
   we're called from GGI's svgalib wrapper. */
#define GSW_MAGIC	(-4711)
extern int __svgalib_tty_fd;

int GGIdlinit(ggi_visual *vis,const char *args,void *argptr)
{
	gii_input *inp;
	int i;
	
	if (__svgalib_tty_fd == GSW_MAGIC) {
		ggiPanic("SVGAlib target called from the SVGAlib wrapper!"
			 " Terminating.\n");
	}
	
	ggLock(_ggi_global_lock); /* Entering protected section */
	if (usagecounter > 0) {
		fprintf(stderr, "display-svga: You can only open this target "
			"once in an application.\n");
		ggUnlock(_ggi_global_lock);
		return GGI_DL_ERROR;
	}
	usagecounter++;
	ggUnlock(_ggi_global_lock); /* Exiting protected section */

	LIBGGI_PRIVATE(vis)=_ggi_malloc(sizeof(struct svga_priv));

	LIBGGI_GC(vis) = _ggi_malloc(sizeof(ggi_gc));
	
	/* Save original signal handlers because SVGAlib will set its own */
	for(i = 0; i<sizeof(vga_signals); i++)
		sigaction(vga_signals[i], NULL, old_signals+i);

	if(!_ggiDebugState)
		vga_disabledriverreport();

	vga_safety_fork(_GGIexitfunc);
	if(vga_init()) {
		fprintf(stderr, "display-SVGAlib: vga_init() failed\n");
		usagecounter--;
		return GGI_DL_ERROR;
	}
    
	_GGIchecksvgamodes(vis);

	/* Open keyboard and mouse input */
	vis->input = giiOpen("input-linux-kbd", NULL);
	if (vis->input == NULL) {
		fprintf(stderr, "display-SVGAlib: Couldn't open kbd.\n");
		free(LIBGGI_GC(vis));
		free(LIBGGI_PRIVATE(vis));
		usagecounter--;
		return GGI_DL_ERROR;
	}
	if ((inp = giiOpen("linux-mouse:auto", &args, NULL)) != NULL) {
		vis->input = giiJoinInputs(vis->input, inp);
	}

	/* Has mode management */
	vis->opdisplay->flush=GGI_svga_flush;
	vis->opdisplay->getmode=GGI_svga_getmode;
	vis->opdisplay->setmode=GGI_svga_setmode;
	vis->opdisplay->getapi=GGI_svga_getapi;
	vis->opdisplay->checkmode=GGI_svga_checkmode;
	vis->opdisplay->setflags=GGI_svga_setflags;

	return GGI_DL_OPDISPLAY;
}

int GGIdlcleanup(ggi_visual *vis)
{
	int i;

	_GGIexitfunc();

	_GGI_svga_freedbs(vis);

	giiClose(vis->input);
	vis->input = NULL;

	/* Get rid of SVGAlib's signal handlers */
	for(i = 0; i<sizeof(vga_signals); i++)
		sigaction(vga_signals[i], old_signals+i, NULL);

#ifndef ALWAYS_SEGFAULT_ON_EXIT
	if(!_ggiDebugState)
		signal(SIGSEGV,_exit);
#endif

	fflush(stderr);
	fflush(stdout);

	if (SVGA_PRIV(vis)->availmodes != NULL)
		free(SVGA_PRIV(vis)->availmodes);
	if (LIBGGI_PRIVATE(vis) != NULL)
		free(LIBGGI_PRIVATE(vis)); 

	if(vis->palette)
		free(vis->palette);
	free(LIBGGI_GC(vis));
	
	usagecounter--;

	return 0;
}

#include <ggi/internal/ggidlinit.h>
