#ifndef __INCLUDED_JS_H__
#define __INCLUDED_JS_H__ 1

#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
#  include <windows.h>
#  if defined( __CYGWIN32__ ) || defined( __CYGWIN__ )
#    define NEAR /* */
#    define FAR  /* */
#  endif
#  include <mmsystem.h>
#  include <string.h>
#else

#  include <unistd.h>
#  include <fcntl.h>

#  ifdef __linux__
#    include <linux/joystick.h>
#  else
#    ifndef JS_DATA_TYPE

      /*
        Not Windoze and no joystick driver...

        Well - we'll put these values in and that should
        allow the code to at least compile. The JS open
        routine should error out and shut off all the code
        downstream anyway
      */

       struct JS_DATA_TYPE
       {
         int buttons ;
         int x ;
         int y ;
       } ;

#      define JS_RETURN (sizeof(struct JS_DATA_TYPE))
#    endif
#  endif
#endif

#define JS_TRUE  1
#define JS_FALSE 0

#ifdef WIN32
#  define _JS_MAX_AXES 6
#else

/*
  This is all set up for the older Linux and BSD drivers
  which restrict you to two axes. When the new Linux driver
  becomes more prevelant, we should change this.
*/

#  define _JS_MAX_AXES 2
#endif

class jsJoystick
{
#ifdef WIN32
  JOYINFOEX    js       ;
  UINT         js_id    ;
#else
  JS_DATA_TYPE js       ;
  char         fname [ 128 ] ;
  int          fd       ;
#endif

  int          error    ;
  int          num_axes ;

  float dead_band [ _JS_MAX_AXES ] ;
  float center    [ _JS_MAX_AXES ] ;
  float max       [ _JS_MAX_AXES ] ;
  float min       [ _JS_MAX_AXES ] ;

  void open ()
  {
#ifdef WIN32
    JOYCAPS jsCaps ;

    js . dwFlags = JOY_RETURNALL ;
    js . dwSize  = sizeof ( js ) ;

    memset ( &jsCaps, 0, sizeof(jsCaps) ) ;

    error = ( joyGetDevCaps( js_id, &jsCaps, sizeof(jsCaps) )
                 != JOYERR_NOERROR ) ;

    num_axes = ( jsCaps.wMaxAxes < _JS_MAX_AXES ) ?
                     jsCaps.wMaxAxes : _JS_MAX_AXES ;

    /* WARNING - Fall through case clauses!! */

    switch ( num_axes )
    {
      case 6 : min[5] = (float)jsCaps.wVmin ; max[5] = (float)jsCaps.wVmax ;
      case 5 : min[4] = (float)jsCaps.wUmin ; max[4] = (float)jsCaps.wUmax ;
      case 4 : min[3] = (float)jsCaps.wRmin ; max[3] = (float)jsCaps.wRmax ;
      case 3 : min[2] = (float)jsCaps.wZmin ; max[2] = (float)jsCaps.wZmax ;
      case 2 : min[1] = (float)jsCaps.wYmin ; max[1] = (float)jsCaps.wYmax ;
      case 1 : min[0] = (float)jsCaps.wXmin ; max[0] = (float)jsCaps.wXmax ;
               break ;
      default :
               setError () ;
               break ;
    }

    for ( int i = 0 ; i < num_axes ; i++ )
    {
      center [ i ] = ( max[i] + min[i] ) / 2.0f ;
      dead_band [ i ] = 0.0f ;
    }

#else
    num_axes = _JS_MAX_AXES ;

    fd = ::open ( fname, O_RDONLY ) ;

    error = ( fd < 0 ) ;

    if ( error )
      return ;

    int counter = 0 ;

    /*
      The Linux driver seems to return 512 for all axes
      when no stick is present - but there is a chance
      that could happen by accident - so it's gotta happen
      on both axes for at least 100 attempts.
    */

    do
    { 
      rawRead ( NULL, center ) ;
      counter++ ;
    } while ( ! error &&
                counter < 100 &&
                center[0] == 512.0f &&
                center[1] == 512.0f ) ;
   
    if ( counter >= 100 )
      setError() ;

    for ( int i = 0 ; i < _JS_MAX_AXES ; i++ )
    {
      max [ i ] = center [ i ] * 2.0f ;
      min [ i ] = 0.0f ;
      dead_band [ i ] = 0.0f ;
    }
#endif
  }

  void close ()
  {
#ifndef WIN32
    if ( ! error )
      ::close ( fd ) ;
#endif
  }

  float fudge_axis ( float value, int axis )
  {
    if ( value < center[axis] )
    {
      float xx = ( value - center[ axis ] ) /
		 ( center [ axis ] - min [ axis ] ) ;

      xx = ( xx > -dead_band [ axis ] ) ? 0.0f :
                  ( ( xx + dead_band [ axis ] ) / ( 1.0f - dead_band [ axis ] ) ) ;

      return ( xx < -1.0f ) ? -1.0f : xx ;
    }
    else
    {
      float xx = ( value - center [ axis ] ) /
		 ( max [ axis ] - center [ axis ] ) ;

      xx = ( xx < dead_band [ axis ] ) ? 0.0f :
                  ( ( xx - dead_band [ axis ] ) / ( 1.0f - dead_band [ axis ] ) ) ;

      return ( xx > 1.0f ) ? 1.0f : xx ;
    }
  }

public:

  jsJoystick ( int id = 0 )
  {
#ifdef WIN32
    switch ( id )
    {
      case 0  : js_id = JOYSTICKID1 ; open () ; break ;
      case 1  : js_id = JOYSTICKID2 ; open () ; break;
      default : num_axes = 0 ; setError () ; break ;
    }
#else
    sprintf ( fname, "/dev/js%d", id ) ;
    open () ;
#endif
  }

  ~jsJoystick ()
  {
    close () ;
  }

  int  getNumAxes () { return num_axes ; }
  int  notWorking () { return error ;    }
  void setError   () { error = JS_TRUE ; }

  float getDeadBand ( int axis )           { return dead_band [ axis ] ; }
  void  setDeadBand ( int axis, float db ) { dead_band [ axis ] = db   ; }

  void setMinRange ( float *axes ) { memcpy ( min   , axes, num_axes * sizeof(float) ) ; }
  void setMaxRange ( float *axes ) { memcpy ( max   , axes, num_axes * sizeof(float) ) ; }
  void setCenter   ( float *axes ) { memcpy ( center, axes, num_axes * sizeof(float) ) ; }

  void getMinRange ( float *axes ) { memcpy ( axes, min   , num_axes * sizeof(float) ) ; }
  void getMaxRange ( float *axes ) { memcpy ( axes, max   , num_axes * sizeof(float) ) ; }
  void getCenter   ( float *axes ) { memcpy ( axes, center, num_axes * sizeof(float) ) ; }

  void read ( int *buttons, float *axes )
  {
    if ( error )
    {
      if ( buttons )
        *buttons = 0 ;

      if ( axes )
        for ( int i = 0 ; i < num_axes ; i++ )
          axes[i] = 0.0f ;
    }

    float raw_axes [ _JS_MAX_AXES ] ;

    rawRead ( buttons, raw_axes ) ;

    if ( axes )
      for ( int i = 0 ; i < num_axes ; i++ )
        axes[i] = ( i < num_axes ) ? fudge_axis ( raw_axes[i], i ) : 0.0f ; 
  }

  void rawRead ( int *buttons, float *axes )
  {
    if ( error )
    {
      if ( buttons )
        *buttons = 0 ;

      if ( axes )
        for ( int i = 0 ; i < num_axes ; i++ )
          axes[i] = 1500.0f ;

      return ;
    }

#ifdef WIN32
    MMRESULT status = joyGetPosEx ( js_id, &js ) ;

    if ( status != JOYERR_NOERROR )
    {
      setError() ;
      return ;
    }

    if ( buttons )
      *buttons = (int) js . dwButtons ;

    if ( axes )
    {
      /* WARNING - Fall through case clauses!! */

      switch ( num_axes )
      {
	case 6: axes[5] = (float) js . dwVpos ;
	case 5: axes[4] = (float) js . dwUpos ;
	case 4: axes[3] = (float) js . dwRpos ;
	case 3: axes[2] = (float) js . dwZpos ;
	case 2: axes[1] = (float) js . dwYpos ;
	case 1: axes[0] = (float) js . dwXpos ;
      }
    }
#else
    int status = ::read ( fd, &js, JS_RETURN ) ;

    if ( status != JS_RETURN )
    {
      perror ( fname ) ;
      setError () ;
      return ;
    }

    if ( buttons )
      *buttons = js.buttons ;

    if ( axes )
    {
      axes[0] = (float) js.x ;
      axes[1] = (float) js.y ;
    }
#endif
  }
} ;

#endif

