////////////////////////////////////////////////////////
//
// drawStereo
//
// mdanks@cybermed.ucsd.edu
//
// Implementation file
//
//    Copyright (c) 1998 Mark Danks.
//    For information on usage and redistribution, and for a DISCLAIMER OF ALL
//    WARRANTIES, see the file, "LICENSE.TERMS" in this distribution.
//
/////////////////////////////////////////////////////////

#include "WindowMan.h"

#include <iostream.h>
#include <GL/glu.h>

#include "EventHandlerWin.h"

BOOL bOverlaySetupPixelFormat(HDC hdc, const WindowHints &hints)
{
    // setup the flags
    DWORD newFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
    // double buffered?
    int buffer = 0;
    if (hints.buffer == 2)
        buffer = PFD_DOUBLEBUFFER;
    // stereo buffers?
    if (hints.stereo)
        newFlags = newFlags | PFD_STEREO;     // probably not supported
    // overlay plane?
    if (hints.overlay)
        newFlags = newFlags | PFD_SWAP_LAYER_BUFFERS;

    PIXELFORMATDESCRIPTOR pfd;
    LAYERPLANEDESCRIPTOR  lpd;		// layer plane descriptor

    // get the maximum number of pixel formats
    int maxpf = DescribePixelFormat(hdc, 0, 0, NULL);
    
    // find an overlay layer descriptor
    int pf;
    int bDone = 0;

    for (pf = 0; pf < maxpf && !bDone; pf++)
    {
        DescribePixelFormat(hdc, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

        // the bReserved field of the PIXELFORMATDESCRIPTOR contains the
	    //  number of overlay/underlay planes
        if (pfd.bReserved > 0)
        {
	        // aha! This format has overlays/underlays
            wglDescribeLayerPlane(hdc, pf, 1, sizeof(LAYERPLANEDESCRIPTOR), &lpd);

            // check all of the flags for the overlay
            int allOkay = 1;
	        if ( !(lpd.dwFlags & LPD_SUPPORT_OPENGL) ||
                 !(lpd.dwFlags & LPD_TRANSPARENT) ||
                 (lpd.crTransparent != 0))
                allOkay = 0;
            if ( (hints.overlayBuffer == 1) &&
                 (lpd.dwFlags & LPD_DOUBLEBUFFER))
                allOkay = 0;
            else if ( (hints.overlayBuffer == 2) &&
                      !(lpd.dwFlags & LPD_DOUBLEBUFFER))
                allOkay = 0;
            if ( hints.overlayStereo &&
                 !(lpd.dwFlags & LPD_STEREO))
                allOkay = 0;
            
            // does the main buffer match what we want?
            if ( !(pfd.dwFlags & PFD_DRAW_TO_WINDOW) &&
                 !(pfd.dwFlags & PFD_SUPPORT_OPENGL))
                allOkay = 0;
            if ( (hints.buffer == 1) &&
                 (pfd.dwFlags & PFD_DOUBLEBUFFER))
                allOkay = 0;
            else if ( (hints.buffer == 2) &&
                      !(pfd.dwFlags & PFD_DOUBLEBUFFER))
                allOkay = 0;
            if ( hints.stereo &&
                 !(pfd.dwFlags & PFD_STEREO))
                allOkay = 0;

/* For some reason, this never succeeds...
            if ( hints.overlay &&
                 !(pfd.dwFlags & PFD_SWAP_LAYER_BUFFERS))
                allOkay = 0;
*/
            if (pfd.iPixelType != PFD_TYPE_RGBA &&
                pfd.cColorBits < 24 &&
                pfd.cDepthBits < 24 &&
                pfd.cRedBits < 8 &&
                pfd.cBlueBits < 8 &&
                pfd.cGreenBits < 8)
            {
                allOkay = 0;
            }

            // we suceeded!
            if (allOkay)
                bDone = 1;
	    }
    }
    
    if (!bDone)
    {
        // couldn't find any overlay/underlay planes
//        cerr << "Hardware does not support overlay planes" << endl;
        return(0);
    }
    // get the pixel format back to where it is supposed to be (extra
    //      increment in the for loop)
    pf--;

    // now get the "normal" pixel format descriptor for the layer
    DescribePixelFormat(hdc, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

    // set the pixel format
    if(SetPixelFormat(hdc, pf, &pfd) == FALSE)
    {
	    cerr << "SetPixelFormat() failed:  Cannot set format specified." << endl;
        return (0);
    }
    return(1);
}

BOOL bSetupPixelFormat(HDC hdc, const WindowHints &hints)
{
    PIXELFORMATDESCRIPTOR pfd;
	
	// clean out the descriptor
    memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
    
    pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
	pfd.cRedBits = 8;
	pfd.cBlueBits = 8;
	pfd.cGreenBits = 8;
    pfd.cDepthBits = 24;
    pfd.cAccumBits = 0;
    pfd.cStencilBits = 0;

    // Set conditional flags
    //
    // double buffered?
    if (hints.buffer == 2)
        pfd.dwFlags = pfd.dwFlags | PFD_DOUBLEBUFFER;
    // stereo buffers?
    if (hints.stereo)
        pfd.dwFlags = pfd.dwFlags | PFD_STEREO;     // probably not supported

    int pixelformat;
    if ( (pixelformat = ChoosePixelFormat(hdc, &pfd)) == 0 )
    {
        cerr << "ChoosePixelFormat failed" << endl;
        WindowMan::processWinErrors();
        return(FALSE);
    }
    if (SetPixelFormat(hdc, pixelformat, &pfd) == FALSE)
    {
        cerr << "SetPixelFormat failed" << endl;
        WindowMan::processWinErrors();
        return(FALSE);
    }
    return(TRUE);
}

/////////////////////////////////////////////////////////
//
// WindowMan
//
/////////////////////////////////////////////////////////
// Constructor
//
/////////////////////////////////////////////////////////
WindowMan :: WindowMan()
{ }

/////////////////////////////////////////////////////////
// Destructor
//
/////////////////////////////////////////////////////////
WindowMan :: ~WindowMan()
{
	destroyWindow();
}

/////////////////////////////////////////////////////////
// createWindow
//
/////////////////////////////////////////////////////////
int WindowMan :: createWindow(WindowHints &hints)
{
    static int firstTime = 1;

    // Register the frame class
    HINSTANCE hInstance = GetModuleHandle(NULL);
    if (!hInstance)
    {
        cerr << "Unable to get module instance" << endl;
        return(0);
    }
    if (firstTime)
    {
        WNDCLASS wndclass;
        wndclass.style         = 0;
        wndclass.lpfnWndProc   = MainWndProc;
        wndclass.cbClsExtra    = 0;
        wndclass.cbWndExtra    = 0;
        wndclass.hInstance     = hInstance;
        wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wndclass.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
        wndclass.hbrBackground = NULL;
        wndclass.lpszMenuName  = NULL;
        wndclass.lpszClassName = "drawStereo";

        if (!RegisterClass(&wndclass) )
        {
            cerr << "Unable to register window class" << endl;
            return(0);
        }
        firstTime = 0;
    }
    // Create the window
    m_mainInfo.m_win = CreateWindow("drawStereo",
             "drawStereo",
	         WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
             0, 0,                      // position
             hints.width, hints.height, // size
             NULL, NULL,                // no parent, no menu
             hInstance,                 // instance
             NULL);                     // don't pass anything WM_CREATE
    // Make sure we got a window
    if (!m_mainInfo.m_win)
    {
        cerr << "Unable to create window" << endl;
        return(0);
    }

	// create the device context
    m_mainInfo.m_dc = GetDC(m_mainInfo.m_win);
	if (!m_mainInfo.m_dc)
	{
		cerr << "Unable to get device context" << endl;
		destroyWindow();
		return(0);
	}

    // do we have an overlay plane?
    if (hints.overlay)
    {
        // set the pixel format for the window
        if (!bOverlaySetupPixelFormat(m_mainInfo.m_dc, hints))
        {
			// try a different overlay buffer type
			if (hints.overlayBuffer == 2) hints.overlayBuffer = 1;
			else if (hints.overlayBuffer == 1) hints.overlayBuffer = 2;

			// set the pixel format for the window
			if (!bOverlaySetupPixelFormat(m_mainInfo.m_dc, hints))
			{
				// couldn't find any overlay/underlay planes
				cerr << "Hardware does not support overlay planes" << endl;
				cerr << "Unable to set window pixel format" << endl;
				destroyWindow();
				return(0);
				}
		}
        if (!createOverlay(hints))
            return(0);
    }
    else
    {
        // set the pixel format for the window
        if (!bSetupPixelFormat(m_mainInfo.m_dc, hints))
        {
            cerr << "Unable to set window pixel format" << endl;
		    destroyWindow();
            return(0);
        }
    }

	// create the OpenGL context
    m_mainInfo.m_context = wglCreateContext(m_mainInfo.m_dc);
	if (!m_mainInfo.m_context)
	{
		cerr << "Unable to create OpenGL context" << endl;
		destroyWindow();
		return(0);
	}

	// do we share display lists?
    if (hints.shared)
        wglShareLists(hints.shared, m_mainInfo.m_context);

	// make the context the current rendering context
    if (!wglMakeCurrent(m_mainInfo.m_dc, m_mainInfo.m_context))
	{
		cerr << "Unable to make OpenGL context current" << endl;
		destroyWindow();
		return(0);
	}
    m_mainInfo.m_buffer = hints.buffer;
    m_mainInfo.m_stereo = hints.stereo;

    return(1);
}

/////////////////////////////////////////////////////////
// createOverlay
//
/////////////////////////////////////////////////////////
int WindowMan :: displayWindow()
{
    // show and update main window
    ShowWindow(m_mainInfo.m_win, SW_SHOWNORMAL);
    UpdateWindow(m_mainInfo.m_win);
	return(1);
}

/////////////////////////////////////////////////////////
// createOverlay
//
/////////////////////////////////////////////////////////
int WindowMan :: createOverlay(const WindowHints &hints)
{
    int nEntries = 2;			    // number of entries in palette
    COLORREF crEntries[2] =
    {		                        // entries in custom palette
        0x00000000,			        // black (ref #0 = transparent)
        0x00ff00ff,			        // blue
    };

    crEntries[0] = hints.overlayTransparent;
    crEntries[1] = hints.overlayColor;

    // set up the layer palette
    wglSetLayerPaletteEntries(m_mainInfo.m_dc, 1, 0, nEntries, crEntries);

    // realize the palette
    wglRealizeLayerPalette(m_mainInfo.m_dc, 1, TRUE);

    // get the device context
    m_overInfo.m_dc = GetDC(m_mainInfo.m_win);
    if (!m_overInfo.m_dc)
    {
        cerr << "Unable to get Device context for overlay" << endl;
        return(0);
    }

    // create an OpenGL overlay context
    m_overInfo.m_context = wglCreateLayerContext(m_overInfo.m_dc, 1);
    if (!m_overInfo.m_context)
    {
        cerr << "Unable to create OpenGL context for overlay" << endl;
        return(0);
    }

    m_overInfo.m_buffer = hints.overlayBuffer;
    m_overInfo.m_stereo = hints.overlayStereo;
    return(1);
}

/////////////////////////////////////////////////////////
// destroyWindow
//
/////////////////////////////////////////////////////////
void WindowMan :: destroyWindow()
{
	if (m_mainInfo.m_win)
	{
		if (m_mainInfo.m_dc)
		{
			if (m_mainInfo.m_context)
			{
                wglMakeCurrent(NULL, NULL);
			    wglDeleteContext(m_mainInfo.m_context);
			}
            if (m_overInfo.m_context)
            {
                wglMakeCurrent(NULL, NULL);
                wglDeleteContext(m_overInfo.m_context);
            }
		    ReleaseDC(m_mainInfo.m_win, m_mainInfo.m_dc);
		}
	    DestroyWindow(m_mainInfo.m_win);
	}
    m_mainInfo.setToDefault();
    m_overInfo.setToDefault();
}

/////////////////////////////////////////////////////////
// setTitle
//
/////////////////////////////////////////////////////////
void WindowMan:: setTitle(const char *title)
{
    SetWindowText(m_mainInfo.m_win, title);
}

/////////////////////////////////////////////////////////
// makeMainCurrent
//
/////////////////////////////////////////////////////////
int WindowMan :: makeMainCurrent()
{
	// make the context the current rendering context
    if (!wglMakeCurrent(m_mainInfo.m_dc, m_mainInfo.m_context))
    {
		cout << "Unable to make OpenGL context current" << endl;
        return(0);
    }
    return(1);
}

/////////////////////////////////////////////////////////
// makeOverlayCurrent
//
/////////////////////////////////////////////////////////
int WindowMan :: makeOverlayCurrent()
{
    if (!m_overInfo.m_context)
        return(0);

	// make the context the current rendering context
    if (wglMakeCurrent(m_overInfo.m_dc, m_overInfo.m_context) == FALSE)
    {
		cerr << "Unable to make OpenGL overlay context current" << endl;
        return(0);
    }
    return(1);
}

/////////////////////////////////////////////////////////
// swapMainBuffer
//
/////////////////////////////////////////////////////////
void WindowMan :: swapMainBuffer()
{
    if (!makeMainCurrent())
        return;

    if (m_mainInfo.m_buffer == 2)
    {
        if (m_overInfo.m_context)
            wglSwapLayerBuffers(m_mainInfo.m_dc, WGL_SWAP_MAIN_PLANE);
        else
            SwapBuffers(m_mainInfo.m_dc);
    }

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // force the clear
    if (m_mainInfo.m_buffer == 1)
        glFlush();

    glColor3f(1.0, 1.0, 1.0);
    glLoadIdentity();
}
    
/////////////////////////////////////////////////////////
// swapOverlay
//
/////////////////////////////////////////////////////////
void WindowMan :: swapOverlay()
{
    if (!makeOverlayCurrent())
        return;
    
    if (m_overInfo.m_buffer == 2)
        wglSwapLayerBuffers(m_overInfo.m_dc, WGL_SWAP_OVERLAY1);

    glClear(GL_COLOR_BUFFER_BIT);

    // force the clear
    if (m_overInfo.m_buffer == 1)
        glFlush();

    glIndexi(0);
    glLoadIdentity();
}

/////////////////////////////////////////////////////////
// processGLErrors
//
/////////////////////////////////////////////////////////
void WindowMan :: processGLErrors()
{
    GLenum error;
    while ((error = glGetError()) != GL_NO_ERROR)
    {
        cerr << ">>>>>>>>>" << endl;
        cerr << gluErrorString(error) << endl;
        cerr << "<<<<<<<<<" << endl;
    }
}
/////////////////////////////////////////////////////////
// processWinErrors
//
/////////////////////////////////////////////////////////
void WindowMan :: processWinErrors()
{
    LPVOID lpMsgBuf;

    FormatMessage ( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,
        NULL 
    );

    // Display the string.
    cerr << "Error: " << (char *)lpMsgBuf << endl;

    // Free the buffer.
    LocalFree( lpMsgBuf );
}
