/*-----------------------------------------------------------------

    drawStereo

    create/destroy a window

    Copyright (c) 1998 Mark Danks.
    For information on usage and redistribution, and for a DISCLAIMER OF ALL
    WARRANTIES, see the file, "LICENSE.TERMS" in this distribution.

-----------------------------------------------------------------*/

#ifndef INCLUDE_WINMAN_H_
#define INCLUDE_WINMAN_H_

#include <windows.h>
#include <GL/gl.h>

/*-----------------------------------------------------------------
-------------------------------------------------------------------
CLASS
    WindowHints
    
    Hints for window creation

DESCRIPTION
    
-----------------------------------------------------------------*/
class WindowHints
{
	public:

        //////////
        // Default values
        WindowHints()   : actuallyDisplay(1), buffer(2), stereo(0),
                          width(400), height(400), shared(NULL),
                          overlay(0), overlayBuffer(1), overlayTransparent(0),
                          overlayColor(0x00ff0000), overlayStereo(0)
                        {}

        //////////
        // Should the window be realized
		int         actuallyDisplay;

        //////////
        // Single or double buffered
		int         buffer;

        //////////
        // Is it stereo?
        int         stereo;

        //////////
        // The width/x of the window
		int         width;

        //////////
        // The height/y of the window
		int         height;

        //////////
        // The GLcontext to share rendering with
		HGLRC       shared;

        /////////////////////////////////////////////
        // Overlay inforation

        //////////
        // Is there an overlay plane?
		int         overlay;

        //////////
        // Overlay single or double buffered
        int         overlayBuffer;

        //////////
        // Overlay transparent color
        COLORREF    overlayTransparent;

        //////////
        // Overlay color
        COLORREF    overlayColor;

        //////////
        // Is it stereo?
        int         overlayStereo;
};

/*-----------------------------------------------------------------
-------------------------------------------------------------------
CLASS
    WindowInfo
    

DESCRIPTION
    
-----------------------------------------------------------------*/
class WindowInfo
{
    public:

        //////////
        WindowInfo()        { setToDefault(); }

        //////////
        ~WindowInfo()       {}

        //////////
        // Set all variables to default
        void                setToDefault() {m_buffer=2;m_stereo=0;m_context=NULL;
                                            m_win=NULL;m_dc=NULL;}
        //////////
        // number of buffers
        int                 m_buffer;

        //////////
        // is it stereo?
        int                 m_stereo;

        //////////
        // Windows OpenGL context
		HGLRC               m_context;

        //////////
        // Window handle
		HWND                m_win;

        //////////
        // Device context handle
		HDC                 m_dc;     
};
 
/*-----------------------------------------------------------------
-------------------------------------------------------------------
CLASS
    WindowMan
    
    Creates/Destroys windows.

DESCRIPTION
    
-----------------------------------------------------------------*/
class WindowMan
{
    public:
     
        //////////
        // Constructor
        WindowMan();

        //////////
        // Destructor
        virtual ~WindowMan();

        //////////
        // Create the window(s)
        int                 createWindow(WindowHints &hints);

        //////////
        // Display the window
        int                 displayWindow();

        //////////
        // Destroy the window(s)
        void                destroyWindow();

        //////////
        // Set the title of the window
        void                setTitle(const char *title);

        //////////
        // Set the main plane active
        // [int] == 0 == failure
        int                 makeMainCurrent();

        //////////
        // Set the overlay active
        // [int] == 0 == failure
        int                 makeOverlayCurrent();

        //////////
        // Swap the main buffers
        void                swapMainBuffer();

        //////////
        // Swap the overlay
        void                swapOverlay();

		//////////
		// Get the window pointer
		HWND				getWindow()			{ return(m_mainInfo.m_win); }

		//////////
		// Does it have an overlay?
		int					hasOverlay()		{ if (m_overInfo.m_context) return(1);return(0); }

    private:

        //////////
        // Create the overlay window
        int                 createOverlay(const WindowHints &hints);

        //////////
        // Main info
        WindowInfo          m_mainInfo;

        //////////
        // Overlay info
        WindowInfo          m_overInfo;

    public:

        //////////
        // Process and print out any errors through OpenGL
        static void         processGLErrors();

        //////////
        // Process and print out any errors through Win32
        static void         processWinErrors();
};

#endif  // for header file


