
#ifndef SG_H
#define SG_H  1

#include <limits.h>
#include <math.h>

#ifdef WIN32
#include <float.h>
#endif

#define SGfloat float

#define SG_ZERO  0.0f
#define SG_ONE   1.0f
#define SG_TWO   2.0f
#define SG_180   180.0f
#define SG_MAX   FLT_MAX


#ifndef M_PI
#define SG_PI 3.1415926535
#else
#define SG_PI M_PI
#endif

#define SG_DEGREES_TO_RADIANS  (SG_PI/SG_180)
#define SG_RADIANS_TO_DEGREES  (SG_180/SG_PI)

/*
  These are just convenient redefinitions of standard
  math library functions to stop float/double warnings.
*/
   
inline SGfloat sgSqrt   ( SGfloat x ) { return sqrt ( x ) ; }
inline SGfloat sgSquare ( SGfloat x ) { return x * x ; }
inline SGfloat sgAbs    ( SGfloat a ) { return ( a < SG_ZERO ) ? -a : a ; }

inline int sgCompareFloat ( SGfloat a, SGfloat b, SGfloat tol )
{
  if ( ( a + tol ) < b ) return -1 ;
  if ( ( b + tol ) < a ) return  1 ;
  return 0 ;
}


/*
  Types used in SG.
*/

typedef SGfloat sgVec2 [ 2 ] ;
typedef SGfloat sgVec3 [ 3 ] ;
typedef SGfloat sgVec4 [ 4 ] ;

typedef SGfloat sgMat3  [3][3] ;
typedef SGfloat sgMat4  [4][4] ;

struct sgCoord
{
  sgVec3 xyz ;
  sgVec3 hpr ;
} ;

class sgSphere ;
class sgBox ;
class sgFrustum ;

/*
  Some handy constants
*/

#define SG_OUTSIDE  FALSE
#define SG_INSIDE    TRUE
#define SG_STRADDLE     2

inline SGfloat sgHeightOfPlaneVec2 ( sgVec4 plane, sgVec2 pnt )
{
  if ( plane[2] == SG_ZERO )
    return SG_ZERO ;
  else
    return -( plane[0] * pnt[0] + plane[1] * pnt[1] + plane[3] ) / plane[2] ;
}

/*
  Convert a direction vector into a set of euler angles,
  (with zero roll)
*/
extern void sgHPRfromVec3 ( sgVec3 hpr, sgVec3 src ) ;

extern void sgMakeCoordMat4 ( sgMat4 dst, SGfloat x, SGfloat y, SGfloat z,
                                          SGfloat h, SGfloat p, SGfloat r ) ;

inline void sgMakeCoordMat4( sgMat4 dst, sgVec3 xyz, sgVec3 hpr )
{
  sgMakeCoordMat4 ( dst, xyz[0], xyz[1], xyz[2],
                         hpr[0], hpr[1], hpr[2] ) ;
}

inline void sgMakeCoordMat4( sgMat4 dst, sgCoord *src )
{
  sgMakeCoordMat4 ( dst, src->xyz, src->hpr ) ;
}

extern void sgMakeRotMat4   ( sgMat4 dst, SGfloat angle, sgVec3 axis ) ;

inline void sgMakeRotMat4   ( sgMat4 dst, sgVec3 hpr )
{
  sgMakeCoordMat4 ( dst, SG_ZERO, SG_ZERO, SG_ZERO, hpr[0], hpr[1], hpr[2] ) ;
}

inline void sgMakeRotMat4   ( sgMat4 dst, SGfloat h, SGfloat p, SGfloat r )
{
  sgMakeCoordMat4 ( dst, SG_ZERO, SG_ZERO, SG_ZERO, h, p, r ) ;
}

extern void sgMakeTransMat4 ( sgMat4 dst, sgVec3 xyz ) ;
extern void sgMakeTransMat4 ( sgMat4 dst, SGfloat x, SGfloat y, SGfloat z ) ;


extern void sgSetCoord      ( sgCoord *coord, sgMat4 src ) ;

extern void sgMultMat4      ( sgMat4 dst, sgMat4 a, sgMat4 b ) ;
extern void sgPostMultMat4  ( sgMat4 dst, sgMat4 a ) ;
extern void sgPreMultMat4   ( sgMat4 dst, sgMat4 a ) ;

extern void sgTransposeNegateMat4 ( sgMat4 dst ) ;
extern void sgTransposeNegateMat4 ( sgMat4 dst, sgMat4 src ) ;

extern void sgXformVec3     ( sgVec3 dst, sgVec3 src, sgMat4 mat ) ;
extern void sgXformPnt3     ( sgVec3 dst, sgVec3 src, sgMat4 mat ) ;
extern void sgXformVec4     ( sgVec4 dst, sgVec4 src, sgMat4 mat ) ;
extern void sgXformPnt4     ( sgVec4 dst, sgVec4 src, sgMat4 mat ) ;
extern void sgFullXformPnt3 ( sgVec3 dst, sgVec3 src, sgMat4 mat ) ;
extern void sgFullXformPnt4 ( sgVec4 dst, sgVec4 src, sgMat4 mat ) ;

inline void sgXformVec3     ( sgVec3 dst, sgMat4 mat ) { sgXformVec3 ( dst, dst, mat ) ; }
inline void sgXformPnt3     ( sgVec3 dst, sgMat4 mat ) { sgXformPnt3 ( dst, dst, mat ) ; }
inline void sgXformVec4     ( sgVec4 dst, sgMat4 mat ) { sgXformVec4 ( dst, dst, mat ) ; }
inline void sgXformPnt4     ( sgVec4 dst, sgMat4 mat ) { sgXformPnt4 ( dst, dst, mat ) ; }
inline void sgFullXformPnt3 ( sgVec3 dst, sgMat4 mat ) { sgFullXformPnt3 ( dst, dst, mat ) ; }
inline void sgFullXformPnt4 ( sgVec4 dst, sgMat4 mat ) { sgFullXformPnt4 ( dst, dst, mat ) ; }

/*
  Basic low-level vector functions.
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  For each of Vec2, Vec3 and Vec4, we provide inlines for

  Zero      - set all elements to zero.
  Set       - set each element individually.
  Add       - add vectors element by element.
  Sub       - subtract vectors element by element.
  Scale     - multiply each element of a vector by a variable.
  Negate    - negate each element of a vector.
  Compare   - compare vectors element by element with optional tolerance.
              (return TRUE if vectors are equal - within tolerances)
  Equal     - return TRUE if vectors are exactly equal.
  Length    - compute length of a vector.
  Distance  - compute distance between two points.
  ScalarProduct - scalar (dot) product.
  VectorProduct - vector (cross) product (3-element vectors ONLY!).
  Normalise/Normalize - make vector be one unit long.
*/

inline void sgZeroVec2 ( sgVec2 dst ) { dst[0]=dst[1]=SG_ZERO ; }
inline void sgZeroVec3 ( sgVec3 dst ) { dst[0]=dst[1]=dst[2]=SG_ZERO ; }
inline void sgZeroVec4 ( sgVec4 dst ) { dst[0]=dst[1]=dst[2]=dst[3]=SG_ZERO ; }


inline void sgSetVec2 ( sgVec2 dst, SGfloat x, SGfloat y )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
}

inline void sgSetVec3 ( sgVec3 dst, SGfloat x, SGfloat y, SGfloat z )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
  dst [ 2 ] = z ;
}

inline void sgSetVec4 ( sgVec4 dst, SGfloat x, SGfloat y, SGfloat z, SGfloat w )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
  dst [ 2 ] = z ;
  dst [ 3 ] = w ;
}


inline void sgCopyVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
}

inline void sgCopyVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
  dst [ 2 ] = src [ 2 ] ;
}

inline void sgCopyVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
  dst [ 2 ] = src [ 2 ] ;
  dst [ 3 ] = src [ 3 ] ;
}


inline void sgAddVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
}

inline void sgAddVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
  dst [ 2 ] += src [ 2 ] ;
}

inline void sgAddVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
  dst [ 2 ] += src [ 2 ] ;
  dst [ 3 ] += src [ 3 ] ;
}


inline void sgAddVec2 ( sgVec2 dst, sgVec2 src1, sgVec2 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
}

inline void sgAddVec3 ( sgVec3 dst, sgVec3 src1, sgVec3 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] + src2 [ 2 ] ;
}

inline void sgAddVec4 ( sgVec4 dst, sgVec4 src1, sgVec4 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] + src2 [ 2 ] ;
  dst [ 3 ] = src1 [ 3 ] + src2 [ 3 ] ;
}


inline void sgSubVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
}

inline void sgSubVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
  dst [ 2 ] -= src [ 2 ] ;
}

inline void sgSubVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
  dst [ 2 ] -= src [ 2 ] ;
  dst [ 3 ] -= src [ 3 ] ;
}

inline void sgSubVec2 ( sgVec2 dst, sgVec2 src1, sgVec2 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
}

inline void sgSubVec3 ( sgVec3 dst, sgVec3 src1, sgVec3 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] - src2 [ 2 ] ;
}

inline void sgSubVec4 ( sgVec4 dst, sgVec4 src1, sgVec4 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] - src2 [ 2 ] ;
  dst [ 3 ] = src1 [ 3 ] - src2 [ 3 ] ;
}


inline void sgNegateVec2 ( sgVec2 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
}

inline void sgNegateVec3 ( sgVec3 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
  dst [ 2 ] = -dst [ 2 ] ;
}

inline void sgNegateVec4 ( sgVec4 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
  dst [ 2 ] = -dst [ 2 ] ;
  dst [ 3 ] = -dst [ 3 ] ;
}


inline void sgNegateVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
}

inline void sgNegateVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
  dst [ 2 ] = -src [ 2 ] ;
}

inline void sgNegateVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
  dst [ 2 ] = -src [ 2 ] ;
  dst [ 3 ] = -src [ 3 ] ;
}


inline void sgScaleVec2 ( sgVec2 dst, SGfloat s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
}

inline void sgScaleVec3 ( sgVec3 dst, SGfloat s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
  dst [ 2 ] *= s ;
}

inline void sgScaleVec4 ( sgVec4 dst, SGfloat s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
  dst [ 2 ] *= s ;
  dst [ 3 ] *= s ;
}

inline void sgScaleVec2 ( sgVec2 dst, sgVec2 src, SGfloat s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
}

inline void sgScaleVec3 ( sgVec3 dst, sgVec3 src, SGfloat s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
  dst [ 2 ] = src [ 2 ] * s ;
}

inline void sgScaleVec4 ( sgVec4 dst, sgVec4 src, SGfloat s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
  dst [ 2 ] = src [ 2 ] * s ;
  dst [ 3 ] = src [ 3 ] * s ;
}


inline int sgCompareVec2 ( sgVec2 a, sgVec2 b, SGfloat tol )
{
  int val = 0 ;

  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;

  return 0 ;
}

inline int sgCompareVec3 ( sgVec3 a, sgVec3 b, SGfloat tol )
{
  int val = 0 ;

  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;

  return 0 ;
}

inline int sgCompareVec4 ( sgVec4 a, sgVec4 b, SGfloat tol )
{
  int val = 0 ;

  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[3], b[3], tol ) ) != 0 ) return val ;

  return 0 ;
}


inline int sgEqualVec2 ( sgVec2 a, sgVec2 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] ;
}

inline int sgEqualVec3 ( sgVec3 a, sgVec3 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] &&
         a[2] == b[2] ;
}

inline int sgEqualVec4 ( sgVec4 a, sgVec4 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] &&
         a[2] == b[2] &&
         a[3] == b[3] ;
}


inline SGfloat sgScalarProductVec2 ( sgVec2 a, sgVec2 b )
{
  return a[0]*b[0] + a[1]*b[1] ;
}

inline SGfloat sgScalarProductVec3 ( sgVec3 a, sgVec3 b )
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] ;
}

inline SGfloat sgScalarProductVec4 ( sgVec4 a, sgVec4 b )
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3] ;
}


extern void sgVectorProductVec3 ( sgVec3 dst, sgVec3 a, sgVec3 b ) ;

inline SGfloat sgDistanceVec2 ( sgVec2 a, sgVec2 b )
{
  return sgSqrt ( sgSquare ( a[0]-b[0] ) + sgSquare ( a[1]-b[1] ) ) ;
}

inline SGfloat sgDistanceVec3 ( sgVec3 a, sgVec3 b )
{
  return sgSqrt ( sgSquare ( a[0]-b[0] ) + sgSquare ( a[1]-b[1] ) +
                  sgSquare ( a[2]-b[2] ) ) ;
}

inline SGfloat sgDistanceVec4 ( sgVec4 a, sgVec4 b )
{
  return sgSqrt ( sgSquare ( a[0]-b[0] ) + sgSquare ( a[1]-b[1] ) +
                  sgSquare ( a[2]-b[2] ) + sgSquare ( a[3]-b[3] ) ) ;
}


inline SGfloat sgLengthVec2 ( sgVec2 src )
{
  return sgSqrt ( sgScalarProductVec2 ( src, src ) ) ;
}

inline SGfloat sgLengthVec3 ( sgVec3 src )
{
  return sgSqrt ( sgScalarProductVec3 ( src, src ) ) ;
}

inline SGfloat sgLengthVec4 ( sgVec4 src )
{
  return sgSqrt ( sgScalarProductVec4 ( src, src ) ) ;
}

/* Anglo-US spelling issues.  <sigh> */
#define sgNormalizeVec2 sgNormaliseVec2
#define sgNormalizeVec3 sgNormaliseVec3
#define sgNormalizeVec4 sgNormaliseVec4

inline void sgNormaliseVec2 ( sgVec2 dst )
{
  sgScaleVec2 ( dst, SG_ONE / sgLengthVec2 ( dst ) ) ;
}

inline void sgNormaliseVec3 ( sgVec3 dst )
{
  sgScaleVec3 ( dst, SG_ONE / sgLengthVec3 ( dst ) ) ;
}

inline void sgNormaliseVec4 ( sgVec4 dst )
{
  sgScaleVec4 ( dst, SG_ONE / sgLengthVec4 ( dst ) ) ;
}

inline void sgNormaliseVec2 ( sgVec2 dst, sgVec2 src )
{
  sgScaleVec2 ( dst, src, SG_ONE / sgLengthVec2 ( src ) ) ;
}

inline void sgNormaliseVec3 ( sgVec3 dst, sgVec3 src )
{
  sgScaleVec3 ( dst, src, SG_ONE / sgLengthVec3 ( src ) ) ;
}

inline void sgNormaliseVec4 ( sgVec4 dst, sgVec4 src )
{
  sgScaleVec4 ( dst, src, SG_ONE / sgLengthVec4 ( src ) ) ;
}


inline void sgZeroCoord ( sgCoord *dst )
{
  sgSetVec3 ( dst->xyz, SG_ZERO, SG_ZERO, SG_ZERO ) ;
  sgSetVec3 ( dst->hpr, SG_ZERO, SG_ZERO, SG_ZERO ) ;
}

inline void sgSetCoord ( sgCoord *dst, SGfloat x, SGfloat y, SGfloat z,
                                       SGfloat h, SGfloat p, SGfloat r )
{
  sgSetVec3 ( dst->xyz, x, y, z ) ;
  sgSetVec3 ( dst->hpr, h, p, r ) ;
}

inline void sgSetCoord ( sgCoord *dst, sgVec3 xyz, sgVec3 hpr )
{
  sgCopyVec3 ( dst->xyz, xyz ) ;
  sgCopyVec3 ( dst->hpr, hpr ) ;
}

inline void sgCopyCoord ( sgCoord *dst, sgCoord *src )
{
  sgCopyVec3 ( dst->xyz, src->xyz ) ;
  sgCopyVec3 ( dst->hpr, src->hpr ) ;
}



inline void sgCopyMat4 ( sgMat4 dst, sgMat4 src )
{
  sgCopyVec4 ( dst[ 0 ], src[ 0 ] ) ;
  sgCopyVec4 ( dst[ 1 ], src[ 1 ] ) ;
  sgCopyVec4 ( dst[ 2 ], src[ 2 ] ) ;
  sgCopyVec4 ( dst[ 3 ], src[ 3 ] ) ;
}


inline void sgScaleMat4 ( sgMat4 dst, sgMat4 src, SGfloat scale )
{
  sgScaleVec4 ( dst[0], src[0], scale ) ;
  sgScaleVec4 ( dst[1], src[1], scale ) ;
  sgScaleVec4 ( dst[2], src[2], scale ) ;
  sgScaleVec4 ( dst[3], src[3], scale ) ;
}
 

inline void sgMakeIdentMat4 ( sgMat4 dst )
{
  sgSetVec4 ( dst[0], SG_ONE , SG_ZERO, SG_ZERO, SG_ZERO ) ;
  sgSetVec4 ( dst[1], SG_ZERO, SG_ONE , SG_ZERO, SG_ZERO ) ;
  sgSetVec4 ( dst[2], SG_ZERO, SG_ZERO, SG_ONE , SG_ZERO ) ;
  sgSetVec4 ( dst[3], SG_ZERO, SG_ZERO, SG_ZERO, SG_ONE  ) ;
}

extern int  sgCompare3DSqdDist ( sgVec3 a, sgVec3 b, SGfloat sqd_dist ) ;
extern void sgMakeTransMat4 ( sgMat4 m, SGfloat x, SGfloat y, SGfloat z ) ;
extern void sgMakeTransMat4 ( sgMat4 m, sgVec3 xyz ) ;
extern void sgMakeCoordMat4 ( sgMat4 m, SGfloat x, SGfloat y, SGfloat z,
                                        SGfloat h, SGfloat p, SGfloat r ) ;
extern void sgMakeCoordMat4 ( sgMat4 m, sgCoord *c ) ;

inline SGfloat sgDistToPlaneVec3 ( sgVec4 plane, sgVec3 pnt )
{
  return sgScalarProductVec3 ( plane, pnt ) + plane[3] ;
}
 
inline SGfloat sgHeightAbovePlaneVec3 ( sgVec4 plane, sgVec3 pnt )
{
  return pnt[3] - sgHeightOfPlaneVec2 ( plane, pnt ) ;
}

extern void sgMakeNormal    ( sgVec3 dst, sgVec3 a, sgVec3 b, sgVec3 c ) ;

inline void sgMakePlane ( sgVec4 dst, sgVec3 normal, sgVec3 pnt )
{
  sgCopyVec3 ( dst, normal ) ;
  dst [ 3 ] = - sgScalarProductVec3 ( normal, pnt ) ;
}

inline void sgMakePlane ( sgVec4 dst, sgVec3 a, sgVec3 b, sgVec3 c )
{
  /*
    Ax + By + Cz + D == 0 ;
    D = - ( Ax + By + Cz )
      = - ( A*a[0] + B*a[1] + C*a[2] )
      = - sgScalarProductVec3 ( normal, a ) ;
  */

  sgMakeNormal ( dst, a, b, c ) ;

  dst [ 3 ] = - sgScalarProductVec3 ( dst, a ) ;
}




class sgSphere
{
  sgVec3   center ;
  SGfloat radius ;
public:

  SGfloat *getCenter (void) { return center ; }

  void setCenter ( sgVec3 c )
  {
    sgCopyVec3 ( center, c ) ;
  }

  void setCenter ( SGfloat x, SGfloat y, SGfloat z )
  {
    sgSetVec3 ( center, x, y, z ) ;
  }

  SGfloat getRadius (void) { return radius ; }
  void    setRadius ( SGfloat r ) { radius = r ; }

  int isEmpty (void) { return radius <  SG_ZERO ; }
  void empty  (void) { radius = - SG_ONE ; }

  void orthoXform ( sgMat4 m ) { sgXformPnt3 ( center, center, m ) ; }

  void extend ( sgSphere *s ) ;
  void extend ( sgBox    *b ) ;
  void extend ( sgVec3    v ) ;

  int intersects ( sgSphere *s )
  {
    return sgCompare3DSqdDist ( center, s->getCenter(),
                    sgSquare ( radius + s->getRadius() ) ) <= 0 ;
  }

  int intersects ( sgVec4 plane )
  {
    return sgAbs ( sgDistToPlaneVec3 ( plane, center ) ) <= radius ;
  }

  int intersects ( sgBox *b ) ;
} ;


class sgBox
{
  sgVec3 min ;
  sgVec3 max ;

public:

  SGfloat *getMin (void) { return min ; }
  SGfloat *getMax (void) { return max ; }

  void setMin ( SGfloat x, SGfloat y, SGfloat z )
  { 
    sgSetVec3 ( min, x, y, z ) ;
  }

  void setMax ( SGfloat x, SGfloat y, SGfloat z )
  { 
    sgSetVec3 ( max, x, y, z ) ;
  }

  int isEmpty(void)
  {
    return ( min[0] > max[0] ||
             min[1] > max[1] ||
             min[2] > max[2] ) ;
  }

  void empty (void)
  {
    sgSetVec3 ( min,  SG_MAX,  SG_MAX,  SG_MAX ) ;
    sgSetVec3 ( max, -SG_MAX, -SG_MAX, -SG_MAX ) ;
  }

  void extend  ( sgSphere *s ) ;
  void extend  ( sgBox    *b ) ;
  void extend  ( sgVec3    v ) ;

  int intersects ( sgSphere *s )
  {
    return s -> intersects ( this ) ;
  }

  int intersects ( sgBox *b )
  {
    return min[0] <= b->getMax()[0] && max[0] >= b->getMin()[0] &&
           min[1] <= b->getMax()[1] && max[1] >= b->getMin()[1] &&
           min[2] <= b->getMax()[2] && max[2] >= b->getMin()[2] ;
  }

  int intersects ( sgVec4 plane ) ;
} ;

#define SG_NEAR       0x10
#define SG_FAR        0x20
#define SG_TOP        0x30
#define SG_BOT        0x40
#define SG_LEFT       0x50
#define SG_RIGHT      0x60

class sgFrustum
{
  /* The parameters for a glFrustum or pfMakePerspFrust */
  
  SGfloat left, right, top, bot, nnear, ffar ;

  /* The A,B,C terms of the plane equations of the four sloping planes */

  sgVec3 top_plane, bot_plane, left_plane, right_plane ;

  /* A GL/PF-style perspective matrix for this frustum */

  sgMat4 mat ;

  /* These two are only valid for simple frusta */

  SGfloat hfov ;    /* Horizontal Field of View */
  SGfloat vfov ;    /* Vertical   Field of View */

  void update (void) ;
  int getOutcode ( sgVec4 src ) ;

public:

  sgFrustum (void)
  {
    nnear = SG_ONE ;
    ffar  = 1000000.0f ;
    hfov  = 45.0f ;
    vfov  = 45.0f ;
    update () ;
  }

  void setFrustum ( SGfloat l, SGfloat r, SGfloat b, SGfloat t, SGfloat n, SGfloat f )
  {
    left  = l ; right = r ;
    top   = t ; bot   = b ;
    nnear = n ; ffar  = f ;
    hfov = vfov = SG_ZERO ;
    update () ;
  } 

  SGfloat  getHFOV (void) { return hfov  ; }
  SGfloat  getVFOV (void) { return vfov  ; }
  SGfloat  getNear (void) { return nnear ; }
  SGfloat  getFar  (void) { return ffar  ; }
  SGfloat  getLeft (void) { return left  ; }
  SGfloat  getRight(void) { return right ; }
  SGfloat  getTop  (void) { return top   ; }
  SGfloat  getBot  (void) { return bot   ; }

  void getFOV ( SGfloat *h, SGfloat *v )
  {
    if ( h != (SGfloat *) 0 ) *h = hfov ;
    if ( v != (SGfloat *) 0 ) *v = vfov ;
  }

  void setFOV ( SGfloat h, SGfloat v )
  {
    hfov = ( h <= 0 ) ? ( v * 3.0f / 2.0f ) : h ;
    vfov = ( v <= 0 ) ? ( h * 2.0f / 3.0f ) : v ;
    update () ;
  }

  void getNearFar ( SGfloat *n, SGfloat *f )
  {
    if ( n != (SGfloat *) 0 ) *n = nnear ;
    if ( f != (SGfloat *) 0 ) *f = ffar  ;
  }

  void setNearFar ( SGfloat n, SGfloat f )
  {
    nnear = n ;
    ffar  = f ;
    update () ;
  }

  int  contains ( sgVec3 p ) ;
  int  contains ( sgSphere *s ) ;
} ;


/*
  Quaternion routines are Copyright (C) 1999
  Kevin B. Thompson <kevinbthompson@yahoo.com>
*/

/* Quaternion structure  w = real, (x, y, z) = vector */

struct sgQuat
{
  SGfloat w, x, y, z ;
} ;


inline void sgMakeIdentityQuat ( sgQuat *dst )
{
  dst -> w =
    dst -> x =
      dst -> y = SG_ZERO ;

  dst -> z = SG_ONE ;
}


inline void sgSetQuat ( sgQuat *dst,
                        SGfloat w, SGfloat x, SGfloat y, SGfloat z )
{
  dst->w = w ;
  dst->x = x ;
  dst->y = y ;
  dst->z = z ;
}

inline void sgCopyQuat ( sgQuat *dst, sgQuat *src )
{
  dst->w = src->w ;
  dst->x = src->x ;
  dst->y = src->y ;
  dst->z = src->z ;
}


/* Construct a unit quaternion (length==1) */

inline void sgNormalizeQuat ( sgQuat *dst, sgQuat *src )
{
  SGfloat d = src->w * src->w +
              src->x * src->y +
              src->y * src->y +
              src->z * src->z ;
  
  d = (d > SG_ZERO) ? (SG_ONE / sgSqrt ( d )) : SG_ONE ;

  dst->w = src->w * d ;
  dst->x = src->x * d ;
  dst->y = src->y * d ;
  dst->z = src->z * d ;
}



inline void sgNormalizeQuat ( sgQuat *dst ) { sgNormalizeQuat ( dst, dst ) ; }


inline void sgInvertQuat ( sgQuat *dst, sgQuat *src )
{
  SGfloat d = src->w * src->w +
              src->x * src->y +
              src->y * src->y +
              src->z * src->z ;
  
  d = ( d == SG_ZERO ) ? SG_ONE : ( SG_ONE / d ) ;

  dst->w =  src->w * d ;
  dst->x = -src->x * d ;
  dst->y = -src->y * d ;
  dst->z = -src->z * d ;
}

inline void sgInvertQuat ( sgQuat *dst ) { sgInvertQuat ( dst, dst ) ; }


/* Make an angle and axis of rotation from a Quaternion. */

void sgQuatToAngleAxis ( SGfloat *angle, sgVec3 axis, sgQuat *src ) ;
void sgQuatToAngleAxis ( SGfloat *angle, SGfloat *x, SGfloat *y, SGfloat *z,
                                                      sgQuat *src ) ;

/* Make a quaternion from a given angle and axis of rotation */

void sgMakeQuat ( sgQuat *dst, SGfloat angle, sgVec3 axis ) ;
void sgMakeQuat ( sgQuat *dst, SGfloat angle, SGfloat x, SGfloat y, SGfloat z ) ;

/* Multiply quaternions together (concatenate rotations) */

void sgMultQuat ( sgQuat *dst, sgQuat *a, sgQuat *b ) ;

/* Construct a 4x4 rotation matrix equivelant to a given quaternion */

void sgMakeRotMat4 ( sgMat4 dst, sgQuat *q ) ;


inline void sgPostMultQuat ( sgQuat *dst, sgQuat *q )
{
  sgQuat r ;

  sgCopyQuat ( &r, dst ) ;
  sgMultQuat ( dst, &r, q ) ;
}


inline void sgPreMultQuat ( sgQuat *dst, sgQuat *q )
{
  sgQuat r ;

  sgCopyQuat ( &r, dst ) ;
  sgMultQuat ( dst, q, &r ) ;
}


/* Rotate a quaternion by a given angle and axis (convenience function) */


inline void sgRotQuat ( sgQuat *dst, SGfloat angle, sgVec3 axis )
{
  sgQuat q ;

  sgMakeQuat ( &q, angle, axis ) ;
  sgPostMultQuat ( dst, &q ) ;
  sgNormalizeQuat ( dst ) ;
}


inline void sgRotQuat ( sgQuat *dst, SGfloat angle,
                                     SGfloat x, SGfloat y, SGfloat z )
{
  sgVec3 axis ;

  sgSetVec3 ( axis, x, y, z ) ;
  sgRotQuat ( dst, angle, axis ) ;
}

#define SGDfloat double

#define SGD_ZERO  0.0
#define SGD_ONE   1.0
#define SGD_TWO   2.0
#define SGD_180   180.0
#define SGD_MAX   DBL_MAX

#ifndef M_PI
#define SGD_PI 3.1415926535
#else
#define SGD_PI M_PI
#endif

#define SGD_DEGREES_TO_RADIANS  (SGD_PI/SGD_180)
#define SGD_RADIANS_TO_DEGREES  (SGD_180/SGD_PI)

/*
  These are just convenient redefinitions of standard
  math library functions to stop float/double warnings.
*/
   
inline SGDfloat sgdSqrt   ( SGDfloat x ) { return sqrt ( x ) ; }
inline SGDfloat sgdSquare ( SGDfloat x ) { return x * x ; }
inline SGDfloat sgdAbs    ( SGDfloat a ) { return ( a < SGD_ZERO ) ? -a : a ; }

inline int sgdCompareFloat ( SGDfloat a, SGDfloat b, SGDfloat tol )
{
  if ( ( a + tol ) < b ) return -1 ;
  if ( ( b + tol ) < a ) return  1 ;
  return 0 ;
}


/*
  Types used in SGD.
*/

typedef SGDfloat sgdVec2 [ 2 ] ;
typedef SGDfloat sgdVec3 [ 3 ] ;
typedef SGDfloat sgdVec4 [ 4 ] ;

typedef SGDfloat sgdMat3  [3][3] ;
typedef SGDfloat sgdMat4  [4][4] ;

struct sgdCoord
{
  sgdVec3 xyz ;
  sgdVec3 hpr ;
} ;

class sgdSphere ;
class sgdBox ;
class sgdFrustum ;

/*
  Some handy constants
*/

#define SGD_OUTSIDE  FALSE
#define SGD_INSIDE    TRUE
#define SGD_STRADDLE     2

inline SGDfloat sgdHeightOfPlaneVec2 ( sgdVec4 plane, sgdVec2 pnt )
{
  if ( plane[2] == SGD_ZERO )
    return SGD_ZERO ;
  else
    return -( plane[0] * pnt[0] + plane[1] * pnt[1] + plane[3] ) / plane[2] ;
}

/*
  Convert a direction vector into a set of euler angles,
  (with zero roll)
*/
extern void sgdHPRfromVec3 ( sgdVec3 hpr, sgdVec3 src ) ;

extern void sgdMakeCoordMat4 ( sgdMat4 dst, SGDfloat x, SGDfloat y, SGDfloat z,
                                          SGDfloat h, SGDfloat p, SGDfloat r ) ;

inline void sgdMakeCoordMat4( sgdMat4 dst, sgdVec3 xyz, sgdVec3 hpr )
{
  sgdMakeCoordMat4 ( dst, xyz[0], xyz[1], xyz[2],
                         hpr[0], hpr[1], hpr[2] ) ;
}

inline void sgdMakeCoordMat4( sgdMat4 dst, sgdCoord *src )
{
  sgdMakeCoordMat4 ( dst, src->xyz, src->hpr ) ;
}

extern void sgdMakeRotMat4   ( sgdMat4 dst, SGDfloat angle, sgdVec3 axis ) ;

inline void sgdMakeRotMat4   ( sgdMat4 dst, sgdVec3 hpr )
{
  sgdMakeCoordMat4 ( dst, SGD_ZERO, SGD_ZERO, SGD_ZERO, hpr[0], hpr[1], hpr[2] ) ;
}

inline void sgdMakeRotMat4   ( sgdMat4 dst, SGDfloat h, SGDfloat p, SGDfloat r )
{
  sgdMakeCoordMat4 ( dst, SGD_ZERO, SGD_ZERO, SGD_ZERO, h, p, r ) ;
}

extern void sgdMakeTransMat4 ( sgdMat4 dst, sgdVec3 xyz ) ;
extern void sgdMakeTransMat4 ( sgdMat4 dst, SGDfloat x, SGDfloat y, SGDfloat z ) ;


extern void sgdSetCoord      ( sgdCoord *coord, sgdMat4 src ) ;

extern void sgdMultMat4      ( sgdMat4 dst, sgdMat4 a, sgdMat4 b ) ;
extern void sgdPostMultMat4  ( sgdMat4 dst, sgdMat4 a ) ;
extern void sgdPreMultMat4   ( sgdMat4 dst, sgdMat4 a ) ;

extern void sgdTransposeNegateMat4 ( sgdMat4 dst ) ;
extern void sgdTransposeNegateMat4 ( sgdMat4 dst, sgdMat4 src ) ;

extern void sgdXformVec3     ( sgdVec3 dst, sgdVec3 src, sgdMat4 mat ) ;
extern void sgdXformPnt3     ( sgdVec3 dst, sgdVec3 src, sgdMat4 mat ) ;
extern void sgdXformVec4     ( sgdVec4 dst, sgdVec4 src, sgdMat4 mat ) ;
extern void sgdXformPnt4     ( sgdVec4 dst, sgdVec4 src, sgdMat4 mat ) ;
extern void sgdFullXformPnt3 ( sgdVec3 dst, sgdVec3 src, sgdMat4 mat ) ;
extern void sgdFullXformPnt4 ( sgdVec4 dst, sgdVec4 src, sgdMat4 mat ) ;

inline void sgdXformVec3     ( sgdVec3 dst, sgdMat4 mat ) { sgdXformVec3 ( dst, dst, mat ) ; }
inline void sgdXformPnt3     ( sgdVec3 dst, sgdMat4 mat ) { sgdXformPnt3 ( dst, dst, mat ) ; }
inline void sgdXformVec4     ( sgdVec4 dst, sgdMat4 mat ) { sgdXformVec4 ( dst, dst, mat ) ; }
inline void sgdXformPnt4     ( sgdVec4 dst, sgdMat4 mat ) { sgdXformPnt4 ( dst, dst, mat ) ; }
inline void sgdFullXformPnt3 ( sgdVec3 dst, sgdMat4 mat ) { sgdFullXformPnt3 ( dst, dst, mat ) ; }
inline void sgdFullXformPnt4 ( sgdVec4 dst, sgdMat4 mat ) { sgdFullXformPnt4 ( dst, dst, mat ) ; }

/*
  Basic low-level vector functions.
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  For each of Vec2, Vec3 and Vec4, we provide inlines for

  Zero      - set all elements to zero.
  Set       - set each element individually.
  Add       - add vectors element by element.
  Sub       - subtract vectors element by element.
  Scale     - multiply each element of a vector by a variable.
  Negate    - negate each element of a vector.
  Compare   - compare vectors element by element with optional tolerance.
              (return TRUE if vectors are equal - within tolerances)
  Equal     - return TRUE if vectors are exactly equal.
  Length    - compute length of a vector.
  Distance  - compute distance between two points.
  ScalarProduct - scalar (dot) product.
  VectorProduct - vector (cross) product (3-element vectors ONLY!).
  Normalise/Normalize - make vector be one unit long.
*/

inline void sgdZeroVec2 ( sgdVec2 dst ) { dst[0]=dst[1]=SGD_ZERO ; }
inline void sgdZeroVec3 ( sgdVec3 dst ) { dst[0]=dst[1]=dst[2]=SGD_ZERO ; }
inline void sgdZeroVec4 ( sgdVec4 dst ) { dst[0]=dst[1]=dst[2]=dst[3]=SGD_ZERO ; }


inline void sgdSetVec2 ( sgdVec2 dst, SGDfloat x, SGDfloat y )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
}

inline void sgdSetVec3 ( sgdVec3 dst, SGDfloat x, SGDfloat y, SGDfloat z )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
  dst [ 2 ] = z ;
}

inline void sgdSetVec4 ( sgdVec4 dst, SGDfloat x, SGDfloat y, SGDfloat z, SGDfloat w )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
  dst [ 2 ] = z ;
  dst [ 3 ] = w ;
}


inline void sgdCopyVec2 ( sgdVec2 dst, sgdVec2 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
}

inline void sgdCopyVec3 ( sgdVec3 dst, sgdVec3 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
  dst [ 2 ] = src [ 2 ] ;
}

inline void sgdCopyVec4 ( sgdVec4 dst, sgdVec4 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
  dst [ 2 ] = src [ 2 ] ;
  dst [ 3 ] = src [ 3 ] ;
}


inline void sgdAddVec2 ( sgdVec2 dst, sgdVec2 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
}

inline void sgdAddVec3 ( sgdVec3 dst, sgdVec3 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
  dst [ 2 ] += src [ 2 ] ;
}

inline void sgdAddVec4 ( sgdVec4 dst, sgdVec4 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
  dst [ 2 ] += src [ 2 ] ;
  dst [ 3 ] += src [ 3 ] ;
}


inline void sgdAddVec2 ( sgdVec2 dst, sgdVec2 src1, sgdVec2 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
}

inline void sgdAddVec3 ( sgdVec3 dst, sgdVec3 src1, sgdVec3 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] + src2 [ 2 ] ;
}

inline void sgdAddVec4 ( sgdVec4 dst, sgdVec4 src1, sgdVec4 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] + src2 [ 2 ] ;
  dst [ 3 ] = src1 [ 3 ] + src2 [ 3 ] ;
}


inline void sgdSubVec2 ( sgdVec2 dst, sgdVec2 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
}

inline void sgdSubVec3 ( sgdVec3 dst, sgdVec3 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
  dst [ 2 ] -= src [ 2 ] ;
}

inline void sgdSubVec4 ( sgdVec4 dst, sgdVec4 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
  dst [ 2 ] -= src [ 2 ] ;
  dst [ 3 ] -= src [ 3 ] ;
}

inline void sgdSubVec2 ( sgdVec2 dst, sgdVec2 src1, sgdVec2 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
}

inline void sgdSubVec3 ( sgdVec3 dst, sgdVec3 src1, sgdVec3 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] - src2 [ 2 ] ;
}

inline void sgdSubVec4 ( sgdVec4 dst, sgdVec4 src1, sgdVec4 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] - src2 [ 2 ] ;
  dst [ 3 ] = src1 [ 3 ] - src2 [ 3 ] ;
}


inline void sgdNegateVec2 ( sgdVec2 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
}

inline void sgdNegateVec3 ( sgdVec3 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
  dst [ 2 ] = -dst [ 2 ] ;
}

inline void sgdNegateVec4 ( sgdVec4 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
  dst [ 2 ] = -dst [ 2 ] ;
  dst [ 3 ] = -dst [ 3 ] ;
}


inline void sgdNegateVec2 ( sgdVec2 dst, sgdVec2 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
}

inline void sgdNegateVec3 ( sgdVec3 dst, sgdVec3 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
  dst [ 2 ] = -src [ 2 ] ;
}

inline void sgdNegateVec4 ( sgdVec4 dst, sgdVec4 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
  dst [ 2 ] = -src [ 2 ] ;
  dst [ 3 ] = -src [ 3 ] ;
}


inline void sgdScaleVec2 ( sgdVec2 dst, SGDfloat s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
}

inline void sgdScaleVec3 ( sgdVec3 dst, SGDfloat s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
  dst [ 2 ] *= s ;
}

inline void sgdScaleVec4 ( sgdVec4 dst, SGDfloat s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
  dst [ 2 ] *= s ;
  dst [ 3 ] *= s ;
}

inline void sgdScaleVec2 ( sgdVec2 dst, sgdVec2 src, SGDfloat s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
}

inline void sgdScaleVec3 ( sgdVec3 dst, sgdVec3 src, SGDfloat s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
  dst [ 2 ] = src [ 2 ] * s ;
}

inline void sgdScaleVec4 ( sgdVec4 dst, sgdVec4 src, SGDfloat s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
  dst [ 2 ] = src [ 2 ] * s ;
  dst [ 3 ] = src [ 3 ] * s ;
}


inline int sgdCompareVec2 ( sgdVec2 a, sgdVec2 b, SGDfloat tol )
{
  int val = 0 ;

  if ( ( val = sgdCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgdCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;

  return 0 ;
}

inline int sgdCompareVec3 ( sgdVec3 a, sgdVec3 b, SGDfloat tol )
{
  int val = 0 ;

  if ( ( val = sgdCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgdCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
  if ( ( val = sgdCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;

  return 0 ;
}

inline int sgdCompareVec4 ( sgdVec4 a, sgdVec4 b, SGDfloat tol )
{
  int val = 0 ;

  if ( ( val = sgdCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgdCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
  if ( ( val = sgdCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;
  if ( ( val = sgdCompareFloat( a[3], b[3], tol ) ) != 0 ) return val ;

  return 0 ;
}


inline int sgdEqualVec2 ( sgdVec2 a, sgdVec2 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] ;
}

inline int sgdEqualVec3 ( sgdVec3 a, sgdVec3 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] &&
         a[2] == b[2] ;
}

inline int sgdEqualVec4 ( sgdVec4 a, sgdVec4 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] &&
         a[2] == b[2] &&
         a[3] == b[3] ;
}


inline SGDfloat sgdScalarProductVec2 ( sgdVec2 a, sgdVec2 b )
{
  return a[0]*b[0] + a[1]*b[1] ;
}

inline SGDfloat sgdScalarProductVec3 ( sgdVec3 a, sgdVec3 b )
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] ;
}

inline SGDfloat sgdScalarProductVec4 ( sgdVec4 a, sgdVec4 b )
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3] ;
}


extern void sgdVectorProductVec3 ( sgdVec3 dst, sgdVec3 a, sgdVec3 b ) ;

inline SGDfloat sgdDistanceVec2 ( sgdVec2 a, sgdVec2 b )
{
  return sgdSqrt ( sgdSquare ( a[0]-b[0] ) + sgdSquare ( a[1]-b[1] ) ) ;
}

inline SGDfloat sgdDistanceVec3 ( sgdVec3 a, sgdVec3 b )
{
  return sgdSqrt ( sgdSquare ( a[0]-b[0] ) + sgdSquare ( a[1]-b[1] ) +
                  sgdSquare ( a[2]-b[2] ) ) ;
}

inline SGDfloat sgdDistanceVec4 ( sgdVec4 a, sgdVec4 b )
{
  return sgdSqrt ( sgdSquare ( a[0]-b[0] ) + sgdSquare ( a[1]-b[1] ) +
                  sgdSquare ( a[2]-b[2] ) + sgdSquare ( a[3]-b[3] ) ) ;
}


inline SGDfloat sgdLengthVec2 ( sgdVec2 src )
{
  return sgdSqrt ( sgdScalarProductVec2 ( src, src ) ) ;
}

inline SGDfloat sgdLengthVec3 ( sgdVec3 src )
{
  return sgdSqrt ( sgdScalarProductVec3 ( src, src ) ) ;
}

inline SGDfloat sgdLengthVec4 ( sgdVec4 src )
{
  return sgdSqrt ( sgdScalarProductVec4 ( src, src ) ) ;
}

/* Anglo-US spelling issues.  <sigh> */
#define sgdNormalizeVec2 sgdNormaliseVec2
#define sgdNormalizeVec3 sgdNormaliseVec3
#define sgdNormalizeVec4 sgdNormaliseVec4

inline void sgdNormaliseVec2 ( sgdVec2 dst )
{
  sgdScaleVec2 ( dst, SGD_ONE / sgdLengthVec2 ( dst ) ) ;
}

inline void sgdNormaliseVec3 ( sgdVec3 dst )
{
  sgdScaleVec3 ( dst, SGD_ONE / sgdLengthVec3 ( dst ) ) ;
}

inline void sgdNormaliseVec4 ( sgdVec4 dst )
{
  sgdScaleVec4 ( dst, SGD_ONE / sgdLengthVec4 ( dst ) ) ;
}

inline void sgdNormaliseVec2 ( sgdVec2 dst, sgdVec2 src )
{
  sgdScaleVec2 ( dst, src, SGD_ONE / sgdLengthVec2 ( src ) ) ;
}

inline void sgdNormaliseVec3 ( sgdVec3 dst, sgdVec3 src )
{
  sgdScaleVec3 ( dst, src, SGD_ONE / sgdLengthVec3 ( src ) ) ;
}

inline void sgdNormaliseVec4 ( sgdVec4 dst, sgdVec4 src )
{
  sgdScaleVec4 ( dst, src, SGD_ONE / sgdLengthVec4 ( src ) ) ;
}


inline void sgdZeroCoord ( sgdCoord *dst )
{
  sgdSetVec3 ( dst->xyz, SGD_ZERO, SGD_ZERO, SGD_ZERO ) ;
  sgdSetVec3 ( dst->hpr, SGD_ZERO, SGD_ZERO, SGD_ZERO ) ;
}

inline void sgdSetCoord ( sgdCoord *dst, SGDfloat x, SGDfloat y, SGDfloat z,
                                       SGDfloat h, SGDfloat p, SGDfloat r )
{
  sgdSetVec3 ( dst->xyz, x, y, z ) ;
  sgdSetVec3 ( dst->hpr, h, p, r ) ;
}

inline void sgdSetCoord ( sgdCoord *dst, sgdVec3 xyz, sgdVec3 hpr )
{
  sgdCopyVec3 ( dst->xyz, xyz ) ;
  sgdCopyVec3 ( dst->hpr, hpr ) ;
}

inline void sgdCopyCoord ( sgdCoord *dst, sgdCoord *src )
{
  sgdCopyVec3 ( dst->xyz, src->xyz ) ;
  sgdCopyVec3 ( dst->hpr, src->hpr ) ;
}



inline void sgdCopyMat4 ( sgdMat4 dst, sgdMat4 src )
{
  sgdCopyVec4 ( dst[ 0 ], src[ 0 ] ) ;
  sgdCopyVec4 ( dst[ 1 ], src[ 1 ] ) ;
  sgdCopyVec4 ( dst[ 2 ], src[ 2 ] ) ;
  sgdCopyVec4 ( dst[ 3 ], src[ 3 ] ) ;
}


inline void sgdScaleMat4 ( sgdMat4 dst, sgdMat4 src, SGDfloat scale )
{
  sgdScaleVec4 ( dst[0], src[0], scale ) ;
  sgdScaleVec4 ( dst[1], src[1], scale ) ;
  sgdScaleVec4 ( dst[2], src[2], scale ) ;
  sgdScaleVec4 ( dst[3], src[3], scale ) ;
}
 

inline void sgdMakeIdentMat4 ( sgdMat4 dst )
{
  sgdSetVec4 ( dst[0], SGD_ONE , SGD_ZERO, SGD_ZERO, SGD_ZERO ) ;
  sgdSetVec4 ( dst[1], SGD_ZERO, SGD_ONE , SGD_ZERO, SGD_ZERO ) ;
  sgdSetVec4 ( dst[2], SGD_ZERO, SGD_ZERO, SGD_ONE , SGD_ZERO ) ;
  sgdSetVec4 ( dst[3], SGD_ZERO, SGD_ZERO, SGD_ZERO, SGD_ONE  ) ;
}

extern int  sgdCompare3DSqdDist ( sgdVec3 a, sgdVec3 b, SGDfloat sqd_dist ) ;
extern void sgdMakeTransMat4 ( sgdMat4 m, SGDfloat x, SGDfloat y, SGDfloat z ) ;
extern void sgdMakeTransMat4 ( sgdMat4 m, sgdVec3 xyz ) ;
extern void sgdMakeCoordMat4 ( sgdMat4 m, SGDfloat x, SGDfloat y, SGDfloat z,
                                        SGDfloat h, SGDfloat p, SGDfloat r ) ;
extern void sgdMakeCoordMat4 ( sgdMat4 m, sgdCoord *c ) ;

inline SGDfloat sgdDistToPlaneVec3 ( sgdVec4 plane, sgdVec3 pnt )
{
  return sgdScalarProductVec3 ( plane, pnt ) + plane[3] ;
}
 
inline SGDfloat sgdHeightAbovePlaneVec3 ( sgdVec4 plane, sgdVec3 pnt )
{
  return pnt[3] - sgdHeightOfPlaneVec2 ( plane, pnt ) ;
}

extern void sgdMakeNormal    ( sgdVec3 dst, sgdVec3 a, sgdVec3 b, sgdVec3 c ) ;

inline void sgdMakePlane ( sgdVec4 dst, sgdVec3 normal, sgdVec3 pnt )
{
  sgdCopyVec3 ( dst, normal ) ;
  dst [ 3 ] = - sgdScalarProductVec3 ( normal, pnt ) ;
}

inline void sgdMakePlane ( sgdVec4 dst, sgdVec3 a, sgdVec3 b, sgdVec3 c )
{
  /*
    Ax + By + Cz + D == 0 ;
    D = - ( Ax + By + Cz )
      = - ( A*a[0] + B*a[1] + C*a[2] )
      = - sgdScalarProductVec3 ( normal, a ) ;
  */

  sgdMakeNormal ( dst, a, b, c ) ;

  dst [ 3 ] = - sgdScalarProductVec3 ( dst, a ) ;
}




class sgdSphere
{
  sgdVec3   center ;
  SGDfloat radius ;
public:

  SGDfloat *getCenter (void) { return center ; }

  void setCenter ( sgdVec3 c )
  {
    sgdCopyVec3 ( center, c ) ;
  }

  void setCenter ( SGDfloat x, SGDfloat y, SGDfloat z )
  {
    sgdSetVec3 ( center, x, y, z ) ;
  }

  SGDfloat getRadius (void) { return radius ; }
  void    setRadius ( SGDfloat r ) { radius = r ; }

  int isEmpty (void) { return radius <  SGD_ZERO ; }
  void empty  (void) { radius = - SGD_ONE ; }

  void orthoXform ( sgdMat4 m ) { sgdXformPnt3 ( center, center, m ) ; }

  void extend ( sgdSphere *s ) ;
  void extend ( sgdBox    *b ) ;
  void extend ( sgdVec3    v ) ;

  int intersects ( sgdSphere *s )
  {
    return sgdCompare3DSqdDist ( center, s->getCenter(),
                    sgdSquare ( radius + s->getRadius() ) ) <= 0 ;
  }

  int intersects ( sgdVec4 plane )
  {
    return sgdAbs ( sgdDistToPlaneVec3 ( plane, center ) ) <= radius ;
  }

  int intersects ( sgdBox *b ) ;
} ;


class sgdBox
{
  sgdVec3 min ;
  sgdVec3 max ;

public:

  SGDfloat *getMin (void) { return min ; }
  SGDfloat *getMax (void) { return max ; }

  void setMin ( SGDfloat x, SGDfloat y, SGDfloat z )
  { 
    sgdSetVec3 ( min, x, y, z ) ;
  }

  void setMax ( SGDfloat x, SGDfloat y, SGDfloat z )
  { 
    sgdSetVec3 ( max, x, y, z ) ;
  }

  int isEmpty(void)
  {
    return ( min[0] > max[0] ||
             min[1] > max[1] ||
             min[2] > max[2] ) ;
  }

  void empty (void)
  {
    sgdSetVec3 ( min,  SGD_MAX,  SGD_MAX,  SGD_MAX ) ;
    sgdSetVec3 ( max, -SGD_MAX, -SGD_MAX, -SGD_MAX ) ;
  }

  void extend  ( sgdSphere *s ) ;
  void extend  ( sgdBox    *b ) ;
  void extend  ( sgdVec3    v ) ;

  int intersects ( sgdSphere *s )
  {
    return s -> intersects ( this ) ;
  }

  int intersects ( sgdBox *b )
  {
    return min[0] <= b->getMax()[0] && max[0] >= b->getMin()[0] &&
           min[1] <= b->getMax()[1] && max[1] >= b->getMin()[1] &&
           min[2] <= b->getMax()[2] && max[2] >= b->getMin()[2] ;
  }

  int intersects ( sgdVec4 plane ) ;
} ;

#define SGD_NEAR       0x10
#define SGD_FAR        0x20
#define SGD_TOP        0x30
#define SGD_BOT        0x40
#define SGD_LEFT       0x50
#define SGD_RIGHT      0x60

class sgdFrustum
{
  /* The parameters for a glFrustum or pfMakePerspFrust */
  
  SGDfloat left, right, top, bot, nnear, ffar ;

  /* The A,B,C terms of the plane equations of the four sloping planes */

  sgdVec3 top_plane, bot_plane, left_plane, right_plane ;

  /* A GL/PF-style perspective matrix for this frustum */

  sgdMat4 mat ;

  /* These two are only valid for simple frusta */

  SGDfloat hfov ;    /* Horizontal Field of View */
  SGDfloat vfov ;    /* Vertical   Field of View */

  void update (void) ;
  int getOutcode ( sgdVec4 src ) ;

public:

  sgdFrustum (void)
  {
    nnear = SGD_ONE ;
    ffar  = 1000000.0f ;
    hfov  = 45.0f ;
    vfov  = 45.0f ;
    update () ;
  }

  void setFrustum ( SGDfloat l, SGDfloat r, SGDfloat b, SGDfloat t, SGDfloat n, SGDfloat f )
  {
    left  = l ; right = r ;
    top   = t ; bot   = b ;
    nnear = n ; ffar  = f ;
    hfov = vfov = SGD_ZERO ;
    update () ;
  } 

  SGDfloat  getHFOV (void) { return hfov  ; }
  SGDfloat  getVFOV (void) { return vfov  ; }
  SGDfloat  getNear (void) { return nnear ; }
  SGDfloat  getFar  (void) { return ffar  ; }
  SGDfloat  getLeft (void) { return left  ; }
  SGDfloat  getRight(void) { return right ; }
  SGDfloat  getTop  (void) { return top   ; }
  SGDfloat  getBot  (void) { return bot   ; }

  void getFOV ( SGDfloat *h, SGDfloat *v )
  {
    if ( h != (SGDfloat *) 0 ) *h = hfov ;
    if ( v != (SGDfloat *) 0 ) *v = vfov ;
  }

  void setFOV ( SGDfloat h, SGDfloat v )
  {
    hfov = ( h <= 0 ) ? ( v * 3.0f / 2.0f ) : h ;
    vfov = ( v <= 0 ) ? ( h * 2.0f / 3.0f ) : v ;
    update () ;
  }

  void getNearFar ( SGDfloat *n, SGDfloat *f )
  {
    if ( n != (SGDfloat *) 0 ) *n = nnear ;
    if ( f != (SGDfloat *) 0 ) *f = ffar  ;
  }

  void setNearFar ( SGDfloat n, SGDfloat f )
  {
    nnear = n ;
    ffar  = f ;
    update () ;
  }

  int  contains ( sgdVec3 p ) ;
  int  contains ( sgdSphere *s ) ;
} ;


/*
  Quaternion routines are Copyright (C) 1999
  Kevin B. Thompson <kevinbthompson@yahoo.com>
*/

/* Quaternion structure  w = real, (x, y, z) = vector */

struct sgdQuat
{
  SGDfloat w, x, y, z ;
} ;


inline void sgdMakeIdentityQuat ( sgdQuat *dst )
{
  dst -> w =
    dst -> x =
      dst -> y = SGD_ZERO ;

  dst -> z = SGD_ONE ;
}


inline void sgdSetQuat ( sgdQuat *dst,
                        SGDfloat w, SGDfloat x, SGDfloat y, SGDfloat z )
{
  dst->w = w ;
  dst->x = x ;
  dst->y = y ;
  dst->z = z ;
}

inline void sgdCopyQuat ( sgdQuat *dst, sgdQuat *src )
{
  dst->w = src->w ;
  dst->x = src->x ;
  dst->y = src->y ;
  dst->z = src->z ;
}


/* Construct a unit quaternion (length==1) */

inline void sgdNormalizeQuat ( sgdQuat *dst, sgdQuat *src )
{
  SGDfloat d = src->w * src->w +
              src->x * src->y +
              src->y * src->y +
              src->z * src->z ;
  
  d = (d > SGD_ZERO) ? (SGD_ONE / sgdSqrt ( d )) : SGD_ONE ;

  dst->w = src->w * d ;
  dst->x = src->x * d ;
  dst->y = src->y * d ;
  dst->z = src->z * d ;
}



inline void sgdNormalizeQuat ( sgdQuat *dst ) { sgdNormalizeQuat ( dst, dst ) ; }


inline void sgdInvertQuat ( sgdQuat *dst, sgdQuat *src )
{
  SGDfloat d = src->w * src->w +
              src->x * src->y +
              src->y * src->y +
              src->z * src->z ;
  
  d = ( d == SGD_ZERO ) ? SGD_ONE : ( SGD_ONE / d ) ;

  dst->w =  src->w * d ;
  dst->x = -src->x * d ;
  dst->y = -src->y * d ;
  dst->z = -src->z * d ;
}

inline void sgdInvertQuat ( sgdQuat *dst ) { sgdInvertQuat ( dst, dst ) ; }


/* Make an angle and axis of rotation from a Quaternion. */

void sgdQuatToAngleAxis ( SGDfloat *angle, sgdVec3 axis, sgdQuat *src ) ;
void sgdQuatToAngleAxis ( SGDfloat *angle, SGDfloat *x, SGDfloat *y, SGDfloat *z,
                                                      sgdQuat *src ) ;

/* Make a quaternion from a given angle and axis of rotation */

void sgdMakeQuat ( sgdQuat *dst, SGDfloat angle, sgdVec3 axis ) ;
void sgdMakeQuat ( sgdQuat *dst, SGDfloat angle, SGDfloat x, SGDfloat y, SGDfloat z ) ;

/* Multiply quaternions together (concatenate rotations) */

void sgdMultQuat ( sgdQuat *dst, sgdQuat *a, sgdQuat *b ) ;

/* Construct a 4x4 rotation matrix equivelant to a given quaternion */

void sgdMakeRotMat4 ( sgdMat4 dst, sgdQuat *q ) ;


inline void sgdPostMultQuat ( sgdQuat *dst, sgdQuat *q )
{
  sgdQuat r ;

  sgdCopyQuat ( &r, dst ) ;
  sgdMultQuat ( dst, &r, q ) ;
}


inline void sgdPreMultQuat ( sgdQuat *dst, sgdQuat *q )
{
  sgdQuat r ;

  sgdCopyQuat ( &r, dst ) ;
  sgdMultQuat ( dst, q, &r ) ;
}


/* Rotate a quaternion by a given angle and axis (convenience function) */


inline void sgdRotQuat ( sgdQuat *dst, SGDfloat angle, sgdVec3 axis )
{
  sgdQuat q ;

  sgdMakeQuat ( &q, angle, axis ) ;
  sgdPostMultQuat ( dst, &q ) ;
  sgdNormalizeQuat ( dst ) ;
}


inline void sgdRotQuat ( sgdQuat *dst, SGDfloat angle,
                                     SGDfloat x, SGDfloat y, SGDfloat z )
{
  sgdVec3 axis ;

  sgdSetVec3 ( axis, x, y, z ) ;
  sgdRotQuat ( dst, angle, axis ) ;
}


#endif

