/*
  Copyright (c)1999 ClanSoft.
*/

#ifndef header_hashtable
#define header_hashtable

#include "clanstring.h"

class CL_Hashable
{
public:
	virtual unsigned int get_hashcode()=0;

	virtual bool equals(const CL_String &oh)=0;

	static unsigned int create_hashcode(CL_String &s)
	{
		unsigned int h=0,g;

		for (char *p = s.get_string();*p != '\0';p++)
		{
			h = (h << 4) + (*p);
			g = h&0xf0000000;
			if (g)
			{
				h = h ^ (g >> 24);
				h = h ^ g;
			}
		}

		return h;
	}
};

template<class TYPE = CL_Hashable> class CL_HashTableIterator;

template<class TYPE = CL_Hashable>
class CL_HashTable
{
friend CL_HashTableIterator<TYPE>;
	TYPE *elements;
	unsigned int cur_size;
	unsigned int num_items;

	void rehash()
	{
		unsigned int i;
		unsigned int old_size = cur_size;
		cur_size *= 2;

		TYPE *new_elements = new TYPE[cur_size];
		for (i=0;i<cur_size;i++) new_elements[i] = NULL;

		for (i=0;i<old_size;i++)
		{
			if (elements[i] == NULL) continue;

			unsigned int hashvalue = elements[i]->get_hashcode();
			hashvalue %= cur_size;

			while (new_elements[hashvalue] != NULL)
			{
				hashvalue++;
				if (hashvalue >= cur_size) hashvalue = 0;
			}
			new_elements[hashvalue] = elements[i];
		}

		delete[] elements;
		elements = new_elements;
	}
public:
	CL_HashTable(int initial_size = 4)
	{
		cur_size = initial_size;

		elements = new TYPE[cur_size];
		for (unsigned int i=0;i<cur_size;i++) elements[i] = NULL;

		num_items = 0;
	}

	virtual ~CL_HashTable()
	{
		delete[] elements;
	}

	TYPE lookup(CL_String id)
	{
		if (num_items == 0) return NULL;

		unsigned int hashcode = CL_Hashable::create_hashcode(id);
		hashcode %= cur_size;

		while (elements[hashcode] != NULL)
		{
			if (elements[hashcode]->equals(id)) return elements[hashcode];
			hashcode++;
			if (hashcode >= cur_size) hashcode = 0;
		}

		return NULL;
	}

	TYPE add(TYPE new_elem)
	{
		if ((float(cur_size - num_items)/cur_size) <= 0.15)
		{
			rehash();
		}

		unsigned int hashvalue = new_elem->get_hashcode();
		hashvalue %= cur_size;

		while (elements[hashvalue] != NULL)
		{
			hashvalue++;
			if (hashvalue >= cur_size) hashvalue = 0;
		}
		elements[hashvalue] = new_elem;

		num_items++;
		return new_elem;
	}

	bool del(const CL_String &id)
	{
		if (num_items == 0) return false;

		int hashcode = CL_Hashable::create_hashcode(id);
		hashcode %= cur_size;

		while (elements[hashcode] != NULL)
		{
			if (elements[hashcode]->equals(id))
			{
				elements[hashcode] = NULL;
				num_items--;
				return true;
			}
			hashcode++;
			if (hashcode >= cur_size) hashcode = 0;
		}

		return false;
	}

	bool del(TYPE del_elem)
	{
		if (num_items == 0) return false;

		unsigned int hashcode = del_elem->get_hashcode();
		hashcode %= cur_size;

		while (elements[hashcode] != NULL)
		{
			if (elements[hashcode] == del_elem)
			{
				elements[hashcode] = NULL;
				num_items--;
				return true;
			}
			hashcode++;
			if (hashcode >= cur_size) hashcode = 0;
		}

		return false;
	}

	void empty_table()
	{
		for (int i=0;i<cur_size;i++)
		{
			elements[i] = NULL;
		}
		num_items = 0;
	}

	int get_num_items()
	{
		return num_items;
	}
};

template <class TYPE>
class CL_HashTableIterator
{
	CL_HashTable<TYPE> *list;
	unsigned int current;

public:
	CL_HashTableIterator(CL_HashTable<TYPE> *_list)
	{
		list = _list;
		current = -1;
	}
	
	CL_HashTableIterator(CL_HashTable<TYPE> &_list)
	{
		list = &_list;
		current = -1;
	}

	TYPE first()
	{
		for (int i=0;i<list->cur_size;i++) 
		{
			if (list->elements[i] != NULL) 
			{
				current = i;
				return list->elements[current];
			}
		}

		return NULL;
	}
	
	TYPE last() 
	{
		for (int i=list->cur_size-1;i>=0;i--) 
		{
			if (list->elements[i] != NULL) 
			{
				current = i;
				return list->elements[current];
			}
		}
	}
	
	TYPE prev() 
	{
		if (current == 0) current = list->cur_size;

		for (int i=current-1;i>=0;i--) 
		{
			if (list->elements[i] != NULL) 
			{
				current = i;
				return list->elements[current];
			}
		}

		return NULL;
	}

	TYPE next()
	{
		if (current == list->cur_size) current = -1;

		for (unsigned int i=current+1;i<list->cur_size;i++) 
		{
			if (list->elements[i] != NULL) 
			{
				current = i;
				return list->elements[current];
			}
		}
		return NULL;
	}

	void remove()
	{
		if (current == -1) return;
		list->elements[current] = NULL;;
	}

	TYPE operator()()
	{
		if (current == -1) return NULL;
		return list->elements[current];
	}
};

#endif
