/*
	$Id: vector.h,v 1.12 1999/08/23 21:37:08 sphair Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	1999/06/19 Daniel Vogel
	
	totally replaced old CL_Vector with this code
*/

//! common_component="Math"

#ifndef header_vector
#define header_vector

#include <assert.h>
#include <math.h>
#include <iostream.h>

class CL_Vector
//: Vector class
// This class provides basic functions and operators for working with vectors
{
public:
	
	float x,y,z,w;
	
	CL_Vector(float x = 0.0, float y = 0.0, float z = 0.0, float w = 1.0);
	//: Constructor that initializes a vector

	CL_Vector(const CL_Vector &other);
	//: Copy Constructor
	//!param: other - vector to copy

	float norm() const;				// not using w component
	//: Returns the (euclid) norm of the vector. It only uses the x,y,z ordinates! 
	//!retval: the euclid norm of the vector (in R^3)

	void normalize();				// not using w component
	//: Normalizes the vector (not taking into account the w ordinate!)
	
	float dot(const CL_Vector& v) const;
	//: Returns the dot product of the current vector and v
	//!param: v - other vector 
	//!retval: the dot product of the two vectors
	
	float angle(const CL_Vector& v) const;
	//: Returns the angle between the current vector and v
	//!param: v - other vector 
	//!retval: the angle between the two vectors

	CL_Vector cross(const CL_Vector& v) const;
	//: Returns the cross product of the current vector and v
	//!param: v - other vector 
	//!retval: the cross product of the two vectors
	
	CL_Vector rotate(float angle, const CL_Vector& a) const;
	//: Returns the current vector rotated around vector a and angle angle
	//!param: angle - angle to rotate
	//!param: a - rotation axis
	//!retval: the rotated vector

	CL_Vector operator * (float s);  
	//: Scalar product (vector * scalar)
	//!retval: the scalar product

	friend CL_Vector operator *  (float s, const CL_Vector& v);
	//: Scalar product (scalar * vector)
	//!retval: the scalar product

	void operator += (const CL_Vector& v);
	//: += operator

	void operator -= (const CL_Vector& v);  
	//: -= operator

	void operator *= (float s);  
	//: *= operator (scalar multiplication)

	CL_Vector operator + (const CL_Vector& v);
	//: + operator
	
	CL_Vector operator - (const CL_Vector& v);
	//: - operator
	
	CL_Vector operator - ();
	//: unary - operator

	CL_Vector& operator = (const CL_Vector& v);
	//: assignment operator

	int operator == (CL_Vector& v);
	//: Returns true if current vector equals v
	//!param: v - other vector
	//!retval: true if v equals the current vector, false otherwise
    	
	int operator != (CL_Vector& v);
	//: Returns false if current vector equals v
	//!param: v - other vector
	//!retval: false if v equals the current vector, true otherwise
	
	float& operator [] (int n);
	//: Returns reference to n-th ordinate (0. == x, 1. == y, ...)
	//!param: n - number of ordinate (starting with 0)
	//!retval: reference to the n-th ordinate
	
	friend ostream& operator << (ostream&, CL_Vector& v);
	//: cout's the x,y,z ordinates (meant for debugging)
};



inline CL_Vector::CL_Vector(float x, float y, float z, float w)
{
	this->x = x;
	this->y = y;
	this->z = z;
	this->w = w;
}


inline CL_Vector::CL_Vector(const CL_Vector &other)
{
	x = other.x;
	y = other.y;
	z = other.z;
	w = other.w;
}


inline float CL_Vector::norm() const
{
	return sqrtf(x*x+y*y+z*z);
}


inline void CL_Vector::normalize()
{
	float f = norm();
	if (f!=0)
	{
		x /= f;
		y /= f;
		z /= f;
	}
}


inline float CL_Vector::dot(const CL_Vector& v) const
{
	return x*v.x + y*v.y + z*v.z;  
}


inline float CL_Vector::angle(const CL_Vector& v) const
{
	return acosf(dot(v)/(norm()*v.norm()));  
}


inline CL_Vector CL_Vector::cross(const CL_Vector& v) const
{
	CL_Vector tmp = CL_Vector(y * v.z - z * v.y,
				  z * v.x - x * v.z,
				  x * v.y - y * v.x);
	return tmp;  
}


// quick hack, same as glRotatef(angle, a);
inline CL_Vector CL_Vector::rotate(float angle, const CL_Vector& a) const
{
	CL_Vector tmp = CL_Vector();
	float s = sinf(angle);
	float c = cosf(angle);
	tmp.x = x*(a.x*a.x*(1-c)+c)     + y*(a.x*a.y*(1-c)-a.z*s) + z*(a.x*a.z*(1-c)+a.y*s);
	tmp.y = x*(a.y*a.x*(1-c)+a.z*s) + y*(a.y*a.y*(1-c)+c)     + z*(a.y*a.z*(1-c)-a.x*s);
	tmp.z = x*(a.x*a.z*(1-c)-a.y*s) + y*(a.y*a.z*(1-c)+a.x*s) + z*(a.z*a.z*(1-c)+c);
	return tmp;  
}


inline CL_Vector CL_Vector::operator * (float s)
{
	CL_Vector tmp = CL_Vector(s * x,
				  s * y,
				  s * z,
				  s * w);
	return tmp;
}


inline CL_Vector operator * (float s, const CL_Vector& v)
{
	CL_Vector tmp = CL_Vector(s * v.x,
				  s * v.y,
				  s * v.z,
				  s * v.w);
	return tmp;
}


inline void CL_Vector::operator += (const CL_Vector& v)
{
	x += v.x;
	y += v.y;
	z += v.z;
	w += v.z;
}


inline void CL_Vector::operator -= (const CL_Vector& v)
{
  	x -= v.x;
	y -= v.y;
	z -= v.z;
	w -= v.w;
}


inline void CL_Vector::operator *= (float s)
{
	x *= s;
	y *= s;
	z *= s;
	w *= s;
}


inline CL_Vector CL_Vector::operator + (const CL_Vector& v)
{
	CL_Vector tmp = CL_Vector(x + v.x,
				  y + v.y,
				  z + v.z,
				  w + v.w);
  	return tmp;
}


inline CL_Vector CL_Vector::operator - (const CL_Vector& v)
{
	CL_Vector tmp = CL_Vector(x - v.x,
				  y - v.y,
				  z - v.z,
				  w - v.z);
  	return tmp;
}


inline CL_Vector CL_Vector::operator - ()
{
	CL_Vector tmp = CL_Vector(-x,
				  -y,
				  -z,
				  -w);
  	return tmp;
}


inline CL_Vector& CL_Vector::operator = (const CL_Vector& v)
{ 
	x = v.x;
	y = v.y;
	z = v.z;
	w = v.w;
	return *this;
}
 

inline int CL_Vector::operator == (CL_Vector& v)
{
	return ((x == v.x) && (y == v.y) && (z == v.z) && (w == v.w));
}


inline int CL_Vector::operator != (CL_Vector& v)
{
	return !(operator == (v));
}


inline float & CL_Vector::operator [] (int n)
{
	switch (n)
	{
		case 0:	return x;
		case 1: return y;
		case 2: return z;
		case 3: return w;
	}
	assert(false);
	return x;			// dummy
}


inline ostream& operator << (ostream& os, CL_Vector& v)
{
	os << v.x << " " << v.y << " " << v.z;
	return os;
}

#endif