/*
	$Id: surfaceprovider.h,v 1.12 1999/09/09 08:08:48 starch Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------
*/

//! l1_component="Display 2D"

#ifndef header_surfaceprovider
#define header_surfaceprovider

#include "pixelformat.h"
#include "palette.h"

class CL_SurfaceProvider
//: Provides the image needed by CL_Surface.
// The CL_SurfaceProvider is the interface CL_Surface uses to retrieve the
// image when needed. Because the surface provider is independent of the
// display targets pixel format, it is quite easy to write your own surface
// provider. ClanLib Layer 2 also provides a set of surface providers
// including PCX, PNG and Targa image loaders.
//!also: CL_Surface - Surface class in ClanLib.
//!also: Layer2 SurfaceProviders - A set of useful surface providers.
{
public:
	virtual ~CL_SurfaceProvider() {;}

	virtual int get_pitch() const
	{
		return get_width()*get_bytes_pr_pixel();
	}
	//: Returns the actual pitch of the surface data 
	//: (the width returned by get_width() is not 
	//: necessarily the total width, but may be only a 
	//: part of the surface data contained in the 
	//: provider)
	//!retval: The pitch (bytes per line) of the surface data.

	virtual int get_translate_x() const { return 0; }
	//: Returns the offset of the surface_data returned 
	//: by get_surface_data(). This translation will 
	//: be applied to every putscreen using a given 
	//: surfaceproviders data as input.
	//!retval: The x-offset of the surface data.

	virtual int get_translate_y() const { return 0; }
	//: Returns the offset of the surface_data returned 
	//: by get_surface_data(). This translation will 
	//: be applied to every putscreen using a given 
	//: surfaceproviders data as input.
	//!retval: The y-offset of the surface data.
	
	virtual int get_width() const =0;
	//: Returns the width of the image.
	//!retval: Width of the image.
	
	virtual int get_height() const =0;
	//: Returns the height of the image.
	//!retval: Height of the image.
	
	virtual int get_no_sprs() const =0;
	//: Returns number of subsprites in the image.
	//!retval: Number of subsprites in image.
	
	virtual EPixelFormat get_pixel_format() const =0;
	//: Returns the pixelformat used by the image.
	//!retval: The pixelformat used by image.
	
	virtual CL_Palette *get_palette() const =0;
	//: Returns the palatte used by the image. If NULL is returned, it will use 
	//: the system palette on each display card.
	//!retval: Palette used by image - or NULL if the system palette should be used.
	
	virtual int get_transcol() const =0;
	//: Transparency color in image, -1 if none.
	//!retval: Transparency color in image, -1 if none.
	
	virtual void *get_surface_data() const =0;
	//: Returns a pointer to the start of the image data 
	//: to be copied to a given surface
	//!retval: The pointer to the image data.

	virtual void lock()=0; // must be called before any other funcs
	//: CL_Surface calls lock() before the surface data is accessed. This makes 
	//: it possible to the surfaceprovider to only have the image in memory 
	//: between a lock()/unlock() sequence.

	virtual void unlock()=0;
	//: CL_Surface calls unlock() when finished using the surface data.

	int get_bytes_pr_pixel() const;
	//: Returns the bytes per pixel of the image.
	//!retval: Image bytes per pixel.

	bool pixel_transparent(int x, int y, int spr_no=0) const;
	//: Returns true if the pixel at the specified coordinate ('x', 'y') is
	//: transparent.
	//!retval: True if pixel is transparent.
};

#endif
