/*
	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Layer2 surface layer manager header file.

	Class purpose:
	The CL_LayerManager utility-class provides an easy ordering of surfaces
	for z-buffer'ish showing. Also the class is well suited for parallax-
	scrolling games, with moving layers of surfaces. The purpose of the 
	layermanager is basically to order the different surfaces showed
	using the layermanager into layers, and thereby associating those
	surfaces, which should be showed at the same level, with one another.
*/

//! l2_component="Misc"

#ifndef header_layermanager
#define header_layermanager

#include "../../clanlib.h"

class CL_LayerManager
//: The layer manager class provides an easy ordering of surfaces for
//: z-buffer'ish showing. It is well suited for parallax-scrolling games using
//: moving layers of surfaces.
{
public:
	CL_LayerManager(int num_layers=4);
	//: Constructs a new layer manager with 'num_layers' layers. 
	//!param: num_layers - Number of layers max used with this layer manager.

	virtual ~CL_LayerManager();

	virtual void put_screen(CL_Surface *surf, int x, int y, int frame=0, int layer=-1);
	//: Show a surface at position (x, y) showing frame 'frame', on layer 'layer' 
	//: This put_screen will be remembered by the layermanager, and contained in 
	//: the layer, until a reset is called (either a reset of all layers, or of the 
	//: respective layer). 
	//!param: surf - Surface to be shown.
	//!param: x - X coordinate where the surface is to be shown.
	//!param: y - Y coordinate where the surface is to be shown.
	//!param: frame - Ask Martin Starch about this one!?
	//!param: layer - Z-depth/layer that the surface should be drawn at.
	
	virtual void select_layer(int layer);
	//: Selects a new default layer, on which all subsequent operations will be applied.
	//!param: layer - The new default layer.

	virtual int get_currentlayer();
	//: Returns the currently selected default-layer.
	//!retval: Current select layer.
	
	virtual void move_layer(int x_delta, int y_delta);
	//: Moves the current layer (absolute positioning).
	//!param: x_delta - Number of pixels to move the X coordinate.
	//!param: y_delta - Number of pixels to move the Y coordinate.

	virtual void move_layer(int layer, int x_delta, int y_delta);
	//: Moves layer 'layer' (absolute positioning).
	//!param: x_delta - Number of pixels to move the X coordinate.
	//!param: y_delta - Number of pixels to move the Y coordinate.

	virtual void get_layerpos(int *x_delta, int *y_delta);
	//: Retrieves the current position of the default layer 
	//: Arguments to the function is two pointers to integers, 
	//: which will be filled with the result.
	//!param: x_delta - Pointer to integer filled with current X position.
	//!param: y_delta - Pointer to integer filled with current Y position.

	virtual void get_layerpos(int layer, int *x_delta, int *y_delta);
	//: Retrieves the current position of layer 'layer'. 
	//: Arguments to the function is two pointers to integers, 
	//: which will be filled with the result.
	//!param: layer - Layer which the position should be retrieved from.
	//!param: x_delta - Pointer to integer filled with current X position.
	//!param: y_delta - Pointer to integer filled with current Y position.

	virtual void swap_layers(int layer1, int layer2);
	//: Swaps two layers - subsequently layer 'layer1' will be moved 
	//: to layer position 'layer2' and vice versa.
	//!param: layer1 - The one layer to be swapped.
	//!param: layer2 - The other layer to be swapped.

	virtual void show_layers();
	//: Shows all layers starting from the highest numbered layer, showing 
	//: layer0 last (back-to-front ordering).

	virtual void show_singlelayer(int layer);
	//: Shows layer 'layer'.
	//!param: layer - Layer to be shown.

	virtual void reset_layers();
	//: Resets surface-information in all layers.

	virtual void reset_singlelayer(int layer);
	//: Resets surface-information in layer 'layer'.
	//!param: layer - Layer to be reset.

	virtual void wrap_layer(int layer, int wrap_x, int wrap_y);

private:
	class SingleLayer
	{
		struct DrawItem
		{
			CL_Surface *s;
			int x, y, frame;
		};
		CL_List<DrawItem> draw_items;

		int x, y;
	public:
		SingleLayer();
		virtual ~SingleLayer();

		void move(int x, int y);
		void get_pos(int *x, int *y);
		void add(CL_Surface *s, int x, int y, int frame);
		void show();
		void reset();
		void wrap_layer(int wrap_x, int wrap_y);
	};

	SingleLayer *get_layer(int num);

	int num_layers;

	int cur_layer_num;
	SingleLayer *current_layer;
	
	CL_Array<SingleLayer> layers;
};

#endif
