/*
	$Id: resourcetype_surface.h,v 1.9 1999/08/22 17:18:47 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Surface resourcetype header file

*/

//! l2_component="Resources"

#ifndef header_resourcetype_surface
#define header_resourcetype_surface

#include "../../clanlib.h"
#include "../SurfaceProviders/generic_surfaceprovider.h"
#include "resource_manager.h"

class CL_Res_Surface : public CL_ResourceType
//: The surface resource type.
// This class is used to load surface resources from a resource source
// (datafile or script file).
// <br>
// Use the CL_ResourceManager to get an instance to the resource source, and
// then call one of the static functions here to load the resource:
// <code>
// CL_ResourceManager res_manager("my_datafile.dat", true);
// CL_Surface *my_surface = CL_Res_Surface::load("my_surface", res_manager);
// </code>
//!res_type: surface
//!res_option: tga - Interpret this file as a targa image.
//!res_option: pcx - Interpret this file as a pcx image.
//!res_option: x=<val> - Specify the starting x coordinate in the image file.
//!res_option: y=<val> - Specify the starting y coordinate in the image file.
//!res_option: width=<val> - Specify the number of horizontal pixels to store
//!res_option: from the image.
//!res_option: height=<val> - Specify the number of vertical pixels to store from the image.
//!res_option: array=<xval>x<yval> - Creates a surface containing <xval> * <yval>
//!res_option: sprites. This option requires the existance of x, y, width, height
//!res_option: option values in the resource definition; otherwise the option is
//!res_option: ignored. The sprites are read from left to right, top to bottom,
//!res_option: and are each saved with width, height equal to the width,
//!res_option: height options specified.
//!res_option: tcol=<val> - Specifies a given color as transparency color for this
//!res_option: surface. This option is only meaningful for palettisized input
//!res_option: image formats (PCX).
//!res_option: tcol=(<val>, <val2>, ...) - Same as above, but specifies several transpareny colors.
{
public:
	static CL_Surface *load(CL_String res_id, CL_ResourceManager &manager);

	CL_Res_Surface();

private:
friend CL_ResourceManager;
	virtual CL_Resource *try_create_from_file(
		CL_String name,
		CL_String prefix,
		CL_String res_id,
		CL_ResourceOptions *options,
		CL_ResourceManager *parent);

	virtual CL_Resource *try_create_from_datafile(
		CL_String name,
		CL_String prefix,
		CL_String res_id,
		CL_ResourceOptions *options,
		CL_ResourceManager *parent);
};

class CL_MaskTranscolProvider : public CL_SurfaceProvider_Generic
{
	CL_SurfaceProvider *parent;
	bool delete_provider;
	unsigned char *mask_indices;
	int num_mask_indices;
public:
	CL_MaskTranscolProvider(
			CL_SurfaceProvider *parent, 
			bool delete_provider, 
			const unsigned char *mask_indices,
			int num_mask_indices);
	virtual ~CL_MaskTranscolProvider();

	virtual int get_pitch() const;
	virtual int get_translate_x() const;
	virtual int get_translate_y() const;
	virtual int get_width() const;
	virtual int get_height() const;
	virtual int get_no_sprs() const;
	virtual EPixelFormat get_pixel_format() const;
	virtual CL_Palette *get_palette() const;
	virtual int get_transcol() const;
	
	virtual void *get_surface_data() const;

	virtual void perform_lock();
	virtual void perform_unlock();
};

class CL_WritableSurfaceResource : public CL_Resource
{
protected:
	CL_Surface *loaded_surface;
	bool loaded;
	CL_String surface_location;
	CL_ResourceOptions *options;
	CL_ResourceManager *parent;
public:
	CL_WritableSurfaceResource();
	virtual ~CL_WritableSurfaceResource() { ; }

	virtual void write(CL_OutputSource *output);
	virtual bool load();

	virtual CL_SurfaceProvider *create_provider()=0;
	virtual CL_Surface *create_surface();
};

class CL_SurfaceFileResource : public CL_WritableSurfaceResource
{
public:
	CL_SurfaceFileResource(CL_String surface_id, CL_String prefix, CL_String surface_location, CL_ResourceOptions *options, CL_ResourceManager *parent);
	virtual ~CL_SurfaceFileResource();

	virtual CL_SurfaceProvider *create_provider();
};

class CL_SurfaceDatafileResource : public CL_WritableSurfaceResource
{
public:
	CL_SurfaceDatafileResource(CL_String surface_id, CL_String prefix, CL_String surface_location, CL_ResourceOptions *options, CL_ResourceManager *parent);
	virtual ~CL_SurfaceDatafileResource();

	virtual CL_SurfaceProvider *create_provider();
};

#endif
